/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/common/Container.hh,v 1.4 2001/06/20 10:06:26 alex Exp $
 */

// -*- C++ -*-  Mon Nov 23 1998
// Eugene Skepner <eu@qub.com>
// $Id: Container.hh,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef Container_hh
#define Container_hh

#ifndef Pointer_hh
#include "Pointer.hh"
#endif

#include <set>

//======================================================================

/*
 * The Container template provides a wrapper/extension to a class to add the
 * Shared class functionality.  It does this by building a class that is a
 * descendant of both the base data class and the Shared class.
 */
template <typename C>
class Container : virtual public Shared, public C
{
  public:
     Container() : C() {}
     Container( size_t n ) : C( n ) {}
     Container( const Container<C>& x ) : Container( x ) {}

     Container& operator=( const Container<C>& x ) { C::operator=( x ); return *this; }

     bool operator == ( const Container<C>& toCompare ) const { return (C&)*this ==(C&)toCompare; }
     bool operator <  ( const Container<C>& toCompare ) const { return (C&)*this < (C&)toCompare; }

}; // class Container<C>

//======================================================================

//  template <typename C> void insert( Pointer< Container<C> > to, typename C::const_reference value )
//  {
//       to->push_back( value );
//  }

/*
 * This template provides a simple insert function that adds an item to the back
 * of a STL container.
 */
template <typename C> void insert( C& to, typename C::const_reference value )
{
     to.push_back( value );
}

//======================================================================

//  template <typename Key> void insert( Pointer< Container< set<Key> > > to, typename set<Key>::const_reference value )
//  {
//       to->insert( value );
//  }

/*
 * STL set's (and map's in general) don't provide a push_back function, so use
 * their insert() function to mimic the above insert() function.  Basically,
 * these two functions allow you to always use insert() to add elements to STL
 * lists.
 */
template <typename Key> void insert( set<Key>& to, typename set<Key>::const_reference value )
{
     to.insert( value );
}

/*
 * This function mimics the behavior of the function above while allowing one to use smart pointers
 * when referencing STL set's.
 */
template <typename Key> void insert( Container<set<Key> >& to, typename set<Key>::const_reference value )
{
     to.insert( value );
}

//======================================================================

template <class T> class set_insert_iterator
{
  protected:
     set<T>* container;

  public:
     typedef output_iterator_tag iterator_category;
     typedef void                value_type;
     typedef void                difference_type;
     typedef void                pointer;
     typedef void                reference;

     explicit set_insert_iterator(set<T>& x) : container(&x) {}
     set_insert_iterator<T>&
          operator=(const typename set<T>::value_type& value) {
          container->insert(value);
          return *this;
     }
     set_insert_iterator<T>& operator*() { return *this; }
     set_insert_iterator<T>& operator++() { return *this; }
     set_insert_iterator<T>& operator++(int) { return *this; }
};

template <class T> inline set_insert_iterator<T> inserter(set<T>& x)
{
     return set_insert_iterator<T>( x );
}

//======================================================================

#endif // Container_hh
