/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/common/LanguageInterface.cc,v 1.4 2001/06/20 10:06:27 alex Exp $
 */

// -*- C++ -*-   Thu Nov 26 1998
// Eugene Skepner <eu@qub.com>
// $Id: LanguageInterface.cc,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef LanguageInterface_hh
#include "LanguageInterface.hh"
#endif

#ifndef H_hh
#include "H.hh"
#endif

#include <algorithm>
#include <functional>

//======================================================================

Pointer<LanguageInterface::CallbackData> LanguageInterface::sNoCallbackData = new LanguageInterface::NoCallbackData;

//======================================================================

void LanguageInterface::referenceObject( const Object* object ) const
{
	if ( find( mReferencedObjects.begin(), mReferencedObjects.end(), object ) == mReferencedObjects.end() ) {
		// avoid placing the same object twice
		mReferencedObjects.push_back( object );
		// do not add object to reference objects if it is already in main references list
		for_each( mHObjectsReferences.begin(), mHObjectsReferences.end(), bind2nd( mem_fun1( &HObjectsReferences::referenceObject ), object ) );
	}
}

//======================================================================

void LanguageInterface::dereferenceObject( const Object* object ) const
{
	mReferencedObjects.erase( find( mReferencedObjects.begin(), mReferencedObjects.end(), object ) );
	for_each( mHObjectsReferences.begin(), mHObjectsReferences.end(), bind2nd( mem_fun1( &HObjectsReferences::dereferenceObject ), object ) );
}

//======================================================================

string LanguageInterface::Object::class_documentation() const
{
	const ClassDescription* cd = classDescription();
	string doc = cd->name;
	doc += "\n\n";
	doc += cd->description;
	doc += "\n\nMethods\n\n";

	set<string> described;
	doc += methods_documentation( cd, described );

	string::size_type pos = 0;
	while ( ( pos = doc.find( "Pointer<", pos ) ) != string::npos ) {
		doc.replace( pos, 8, "" );
		if ( ( pos = doc.find( '>', pos ) ) != string::npos )
			doc.replace( pos, 1, "" );
		if ( doc[pos] == ' ' )
			doc.replace( pos, 1, "" );
	}
	pos = 0;
	while ( ( pos = doc.find( "vector<", pos ) ) != string::npos )
		doc.replace( pos, 7, "list<" );
	pos = 0;
	while ( ( pos = doc.find( "set<", pos ) ) != string::npos )
		doc.replace( pos, 4, "list<" );

	return(doc);

} // LanguageInterface::Object::class_documentation

//======================================================================

string LanguageInterface::Object::methods_documentation( const ClassDescription* cd, set<string>& described ) const
{
	string doc;
	for ( const MethodDescription* md = cd->methods; md->name; ++md ) {
		if ( described.find( md->name ) == described.end() ) {
			if ( !md->dynamic() )
				doc += "static ";
			doc += LanguageInterface::method_result_type( md->result, true ) + " " + md->name;
			if ( md->args[0] == MethodDescription::vtVoid ) {
				doc += "()";
			}
			else {
				doc += "(";
				bool optional = false;
				for ( int argno = 0; md->args[argno] != MethodDescription::vtVoid && md->args[argno] != MethodDescription::vtLanguageInterface; ++argno ) {
					if ( argno > 0 )
						doc += ",";
					if ( optional ) {
						doc += "] ";
						optional = false;
					}
					else
						doc += " ";
					if ( md->args[argno].optional() ) {
						doc += "[";
						optional = true;
					}
					doc += method_argument_type( md->args[argno], true );
					if ( md->args[argno].name() ) {
						doc += " ";
						doc += md->args[argno].name();
					}
				}
				if ( optional )
					doc += "]";
				doc += " )";
			}
			doc += "\n";
			doc += md->description;
			doc += "\n\n";
			described.insert( md->name );
		}
	}

	for ( const ClassDescription* const* parent = cd->parents; *parent; ++parent )
		doc += methods_documentation( *parent, described );

	return(doc);

} // LanguageInterface::Object::methods_documentation

//======================================================================

string LanguageInterface::Object::method_documentation( const ClassDescription* cd, const string& aMethod ) const
{
	string doc;
	for ( const MethodDescription* md = cd->methods; md->name; ++md ) {
		if ( md->name == aMethod ) {
			if ( !md->dynamic() )
				doc += "static ";
			doc += LanguageInterface::method_result_type( md->result, true ) + " " + md->name;
			if ( md->args[0] == MethodDescription::vtVoid ) {
				doc += "()";
			}
			else {
				doc += "(";
				bool optional = false;
				for ( int argno = 0; md->args[argno] != MethodDescription::vtVoid && md->args[argno] != MethodDescription::vtLanguageInterface; ++argno ) {
					if ( argno > 0 )
						doc += ",";
					if ( optional ) {
						doc += "] ";
						optional = false;
					}
					else
						doc += " ";
					if ( md->args[argno].optional() ) {
						doc += "[";
						optional = true;
					}
					doc += method_argument_type( md->args[argno], true );
					if ( md->args[argno].name() ) {
						doc += " ";
						doc += md->args[argno].name();
					}
				}
				if ( optional )
					doc += "]";
				doc += " )";
			}
			doc += "\n";
			doc += md->description;
			doc += "\n\n";
			break;
		}
	}

	for ( const ClassDescription* const* parent = cd->parents; doc.empty() && *parent; ++parent )
		doc = method_documentation( *parent, aMethod );

	return(doc);

} // LanguageInterface::Object::method_documentation

//======================================================================

string LanguageInterface::method_result_type( Object::MethodDescription::Value t, bool for_documentation )
{
	switch ( (Object::MethodDescription::ValueType)t ) {
	case Object::MethodDescription::vtVoid:
		return("void");
	case Object::MethodDescription::vtChar:
		return("char");
	case Object::MethodDescription::vtUChar:
		return("unsigned char");
	case Object::MethodDescription::vtShort:
		return("short");
	case Object::MethodDescription::vtUShort:
		return("unsigned short");
	case Object::MethodDescription::vtInt:
		return("int");
	case Object::MethodDescription::vtUInt:
		return("unsigned int");
	case Object::MethodDescription::vtLong:
		return("long");
	case Object::MethodDescription::vtULong:
		return("unsigned long");
	case Object::MethodDescription::vtLongLong:
		return("long long");
	case Object::MethodDescription::vtULongLong:
		return("unsigned long long");
	case Object::MethodDescription::vtBool:
		return("bool");
	case Object::MethodDescription::vtFloat:
		return("float");
	case Object::MethodDescription::vtDouble:
		return("double");
	case Object::MethodDescription::vtLongDouble:
		return("long double");
	case Object::MethodDescription::vtString:
		return("string");
	case Object::MethodDescription::vtObject:
		if ( for_documentation )
			return (const char*)t;
		else
			return(string( "Pointer<" ) + (const char*)t + " >");
	case Object::MethodDescription::vtList:
		if ( for_documentation )
			return (const char*)t;
		else
			return(string( "Pointer<Container<" ) + (const char*)t + " > >");
	case Object::MethodDescription::vtListRef:
		if ( for_documentation )
			return (const char*)t;
		else
			return(string( (const char*)t ) + "&");
	case Object::MethodDescription::vtListPtr:
		if ( for_documentation )
			return (const char*)t;
		else
			return(string( "Pointer<" ) + (const char*)t + " >");
	case Object::MethodDescription::vtLanguageInterface:
		throw Error( string( "vtLanguageInterface cannot be used as method return type for method " ) + (const char*)t );
	case Object::MethodDescription::vtCallbackData:
		if ( for_documentation )
			return("CallbackData");
		else
			return("Pointer<LanguageInterface::CallbackData>");
	}

	return("unknown");

} // LanguageInterface::method_result_type

//======================================================================

string LanguageInterface::method_argument_type( Object::MethodDescription::Value t, bool for_documentation )
{
	switch ( (Object::MethodDescription::ValueType)t ) {
	case Object::MethodDescription::vtVoid:
		return("void");
	case Object::MethodDescription::vtChar:
		return("char");
	case Object::MethodDescription::vtUChar:
		return("unsigned char");
	case Object::MethodDescription::vtShort:
		return("short");
	case Object::MethodDescription::vtUShort:
		return("unsigned short");
	case Object::MethodDescription::vtInt:
		return("int");
	case Object::MethodDescription::vtUInt:
		return("unsigned int");
	case Object::MethodDescription::vtLongLong:
		return("long long");
	case Object::MethodDescription::vtULongLong:
		return("unsigned long long");
	case Object::MethodDescription::vtLong:
		return("long");
	case Object::MethodDescription::vtULong:
		return("unsigned long");
	case Object::MethodDescription::vtBool:
		return("bool");
	case Object::MethodDescription::vtFloat:
		return("float");
	case Object::MethodDescription::vtDouble:
		return("double");
	case Object::MethodDescription::vtLongDouble:
		return("long double");
	case Object::MethodDescription::vtString:
		return("string");
	case Object::MethodDescription::vtObject:
		if ( for_documentation )
			return (const char*)t;
		else
			return(string( "Pointer<" ) + (const char*)t + " >");
	case Object::MethodDescription::vtList:
	case Object::MethodDescription::vtListRef:
		return (const char*)t;
	case Object::MethodDescription::vtListPtr:
		if ( for_documentation )
			return (const char*)t;
		else
			return(string( "Pointer<" ) + (const char*)t + " >");
	case Object::MethodDescription::vtLanguageInterface:
		return("LanguageInterface*");
	case Object::MethodDescription::vtCallbackData:
		if ( for_documentation )
			return("CallbackData");
		else
			return("Pointer<LanguageInterface::CallbackData>");
	}

	return("unknown");

} // LanguageInterface::method_argument_type

//======================================================================
