/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/common/LanguageInterface.hh,v 1.3 2001/06/20 10:06:27 alex Exp $
 */

// -*- C++ -*-  Thu Nov  5 1998
// Eugene Skepner <eu@qub.com>
// $Id: LanguageInterface.hh,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef LanguageInterface_hh
#define LanguageInterface_hh

#ifndef Pointer_hh
#include "Pointer.hh"
#endif

#include <typeinfo>
#include <stdexcept>
#include <map>
#include <cstdio>
#include <set>
#include <list>
#include <algorithm>

//======================================================================

static const char* ObjectStringFormat = "%p_%s";

//======================================================================

class LanguageInterface
{
     friend class H;
     friend class HObjectsReferences;

  public:
     LanguageInterface( bool enableGC ) : mGC( enableGC ) {}
     virtual ~LanguageInterface() {}

     class Error : public runtime_error
     {
       public:
          Error( const string& what_arg ) : runtime_error( what_arg ) {}
     };

     class Object;

     typedef list<const Object*> ReferencedObjects;

     class Object : virtual public Shared
     {
          friend class LanguageInterface;

       public:
          Object()
               {
               }

          virtual ~Object()
               {
                      //cerr << "Object::~Object: " << (void*)this << endl;
               }

          virtual const char* objecttypename() const { return typeid( *this ).name(); }

          struct MethodDescription
          {
               enum ValueType
                    {
                         vtVoid,
                         vtChar, vtUChar,
                         vtShort, vtUShort, vtInt, vtUInt, vtLong, vtULong, vtLongLong, vtULongLong, vtBool,
                         vtFloat, vtDouble, vtLongDouble,
                         vtString, vtObject, vtList, vtListRef, vtListPtr,
                         vtLanguageInterface, vtCallbackData
                    };

               class Value
               {
                 public:
                    Value( ValueType aValueType = vtVoid, const char* aTypeName = 0, const char* aDefaultValue = 0, const char* aName = 0 )
                         : mValueType( aValueType ), mTypeName( aTypeName ), mDefaultValue( aDefaultValue ), mName( aName ) {}

                    operator ValueType () const { return mValueType; }
                    operator const char* () const
                         {
                              if ( mTypeName )
                                   return mTypeName;
                              else if ( mValueType == vtObject )
                                   return "LanguageInterface::Object";
                              else
                                   return 0;
                         }

                    bool optional() const { return mDefaultValue != 0; }
                    const char* defaultValue() const { return mDefaultValue; }
                    const char* name() const { return mName; }

                 private:
                    ValueType mValueType;
                    const char* mTypeName;
                    const char* mDefaultValue;
                    const char* mName; // formal argument name for documentation
               };

               enum Flags
               {
                    fStatic = 0,            // method is static
                    fDynamic = 0x01,        // method is dynamic (i.e. member function)
                    fSafeExecution = 0x02,  // method cannot be called by slave tasks (e.g. by safe slave Tcl interpreter)
                    fGetOrSet = 0x04,       // method is either sets or gets (returns) some value depending on precence of last argument
                    fSetReturnsOld = 0x08,  // if method is of GetOrSet type it returns old value if set is requested
                    fSetReturnsNew = 0x10   // if method is of GetOrSet type it returns new value if set is requested
               };

               MethodDescription() : name( 0 ), name_to_call( 0 ), description( 0 ) {}

               MethodDescription( const char* aName, const char* aNameToCall, unsigned aFlags, const char* aDescription, const Value& aResult, const Value& arg1 = vtVoid, const Value& arg2 = vtVoid, const Value& arg3 = vtVoid, const Value& arg4 = vtVoid, const Value& arg5 = vtVoid, const Value& arg6 = vtVoid, const Value& arg7 = vtVoid, const Value& arg8 = vtVoid, const Value& arg9 = vtVoid, const Value& arg10 = vtVoid )
                    : name( aName ), name_to_call( aNameToCall ), flags( aFlags ), description( aDescription ), result( aResult )
                    {
                         args[0] = arg1; args[1] = arg2; args[2] = arg3; args[3] = arg4; args[4] = arg5;
                         args[5] = arg6; args[6] = arg7; args[7] = arg8; args[8] = arg9; args[9] = arg10;
                         args[10] = vtVoid;
                    }

               void numberOfArgs( int& min_args, int& max_args ) const
                    {
                         min_args = -1;
                         int argno = 0;
                         for ( ; args[argno] != vtVoid; ++argno ) {
                              if ( min_args >= 0 && !args[argno].optional() )
                                   throw Error( string( "Invalid " ) + name + " method description: not optional argument after optional one" );
                              if ( args[argno].optional() && min_args < 0 )
                                   min_args = argno;
                         }
                         max_args = argno;
                         if ( min_args < 0 )
                              min_args = max_args;
                    }

               bool safeExecution() const { return flags & fSafeExecution; }
               bool dynamic() const { return flags & fDynamic; }
               bool getOrSet() const { return flags & fGetOrSet; }
               bool setReturnsOld() const { return ( flags & fGetOrSet ) && ( flags & fSetReturnsOld ); }
               bool setReturnsNew() const { return ( flags & fGetOrSet ) && ( flags & fSetReturnsNew ); }

               const char* name;
               const char* name_to_call;
               unsigned flags;
               const char* description;
               Value result;
               Value args[11];

          }; // struct MethodDescription

          struct CallbackDescription
          {
               CallbackDescription() : type( 0 ) {}
               CallbackDescription( const char* aType, const char* arg1 = 0, const char* arg2 = 0, const char* arg3 = 0, const char* arg4 = 0, const char* arg5 = 0, const char* arg6 = 0, const char* arg7 = 0, const char* arg8 = 0, const char* arg9 = 0, const char* arg10 = 0 )
                    : type( aType )
                    {
                         args[0] = arg1; args[1] = arg2; args[2] = arg3; args[3] = arg4; args[4] = arg5;
                         args[5] = arg6; args[6] = arg7; args[7] = arg8; args[8] = arg9; args[9] = arg10;
                         args[10] = 0;
                    }

               const char* type;
               const char* args[11];

          }; // struct CallbackDescription

          struct ClassDescription
          {
               ClassDescription( const char* aName, const char* aCxxName, const char* aDescription, const MethodDescription* aMethods, const CallbackDescription* aCallbacks, const ClassDescription* aParent1 = 0, const ClassDescription* aParent2 = 0, const ClassDescription* aParent3 = 0, const ClassDescription* aParent4 = 0, const ClassDescription* aParent5 = 0, const ClassDescription* aParent6 = 0 )
                    : name( aName ), cxx_name( aCxxName ), description( aDescription ), methods( aMethods ), callbacks( aCallbacks )
                    {
                         parents[0] = aParent1;
                         parents[1] = aParent2;
                         parents[2] = aParent3;
                         parents[3] = aParent4;
                         parents[4] = aParent5;
                         parents[5] = aParent6;
                         parents[6] = 0;
                    }

               int dynamicMethods() const
                    {
                         int dm = 0;
                         for ( const MethodDescription* m = methods; m->name; ++m )
                              if ( m->dynamic() )
                                   ++dm;
                         return dm;
                    }

               int staticMethods() const
                    {
                         int sm = 0;
                         for ( const MethodDescription* m = methods; m->name; ++m )
                              if ( !m->dynamic() )
                                   ++sm;
                         return sm;
                    }

               bool has_callbacks() const
                    {
                         return callbacks->type != 0;
                    }

               const char* name;
               const char* cxx_name;
               const char* description;
               const MethodDescription* methods;
               const CallbackDescription* callbacks;
               const ClassDescription* parents[7];

          }; // struct ClassDescription

          virtual string class_documentation() const;
          virtual string methods_documentation( const ClassDescription* cd, set<string>& described ) const;
          virtual string method_documentation( const ClassDescription* cd, const string& aMethod ) const;

       private:
          typedef map<const LanguageInterface*,Reference> LanguageReferences;

          mutable LanguageReferences mLanguageRefs;

          virtual Reference::Type languageRef( const LanguageInterface* li, bool gc ) const
               {
                    pair<LanguageReferences::iterator,bool> lri = mLanguageRefs.insert( make_pair( li, Reference( 0 ) ) );
                    if ( gc || lri.first->second.type() == Reference::NotReferenced ) {
                         ref();
                    }
                    return lri.first->second.operator++();
               }

          virtual Reference::Type languageUnref( const LanguageInterface* li, bool gc ) const
               {
                    LanguageReferences::iterator lri = mLanguageRefs.find( li );
                    if ( lri == mLanguageRefs.end() || lri->second.type() == Reference::NotReferenced )
                         throw Error( "Internal: trying to languageUnref not referenced (by the language) object" );
                    Reference::Type rt = lri->second.operator--();
                    if ( gc || rt == Reference::NotReferenced )
                         unref();
                    return rt;
               }

       public:
          virtual const ClassDescription* classDescription() const = 0;

     };

     friend class Object;

     static string object2string( const Object* aObject )
          {
               char s[1024];
               ::sprintf( s, ObjectStringFormat, aObject, aObject ? aObject->objecttypename() : typeid( Object ).name() );
               return s;
          }

       // Returns if automatic garbage collection is enabled
     bool gc() const { return mGC; }

       //const ReferencedObjects& referencedObjects() const { return mReferencedObjects; }
     virtual bool referenced( const Object* object ) const { return find( mReferencedObjects.begin(), mReferencedObjects.end(), object ) != mReferencedObjects.end(); }
     virtual bool referenced( const Object& object ) const { return find( mReferencedObjects.begin(), mReferencedObjects.end(), &object ) != mReferencedObjects.end(); }

     virtual void unregisterObject( const Object* aObject ) = 0;

     static string method_result_type( Object::MethodDescription::Value t, bool for_documentation );
     static string method_argument_type( Object::MethodDescription::Value t, bool for_documentation );

       // --------------------------------------------------

     class SpecificCallbackData;

     class CallbackData : virtual public Shared
     {
       public:
          CallbackData( LanguageInterface* aLanguageInterface ) : mLanguageInterface( aLanguageInterface ) {}

          virtual void activate( SpecificCallbackData& aSpecificCallbackData ) = 0;
          virtual void activate( SpecificCallbackData& aSpecificCallbackData, int arg ) = 0;

       protected:
          LanguageInterface& languageInterface() { return *mLanguageInterface; }

       private:
          LanguageInterface* mLanguageInterface;

     }; // class CallbackData

     class SpecificCallbackData
     {
       public:
          SpecificCallbackData( Pointer<LanguageInterface::CallbackData> aCallbackData ) : mCallbackData( aCallbackData ) {}
          virtual ~SpecificCallbackData() {}

          virtual const char* typeid_name() const = 0;

          Pointer<LanguageInterface::CallbackData> callbackData() const { return mCallbackData; }
          void callbackData( Pointer<LanguageInterface::CallbackData> aCallbackData ) { mCallbackData = aCallbackData; }

       private:
          Pointer<LanguageInterface::CallbackData> mCallbackData;

     }; // class SpecificCallbackData

     class NoCallbackData : public CallbackData
     {
       public:
          NoCallbackData() : CallbackData( 0 ) {}
          virtual void activate( SpecificCallbackData& /*aSpecificCallbackData*/ ) {}
          virtual void activate( SpecificCallbackData& /*aSpecificCallbackData*/, int ) {}

     }; // class NoCallbackData
     static Pointer<CallbackData> sNoCallbackData;

       // --------------------------------------------------

  protected:
     virtual Object::Reference::Type languageRef( const Object* object ) const
          {
               if ( !gc() )
                    referenceObject( object );
               return object->languageRef( this, gc() );
          }

     virtual Object::Reference::Type languageRef( const Object& object ) const
          {
               if ( !gc() )
                    referenceObject( &object );
               return object.languageRef( this, gc() );
          }

     virtual Object::Reference::Type languageUnref( const Object* object ) const
          {
               Object::Reference::Type rt = object->languageUnref( this, gc() );
               if ( !gc() && rt == Object::Reference::NotReferenced )
                    dereferenceObject( object );
               return rt;
          }

     virtual Object::Reference::Type languageUnref( const Object& object ) const
          {
               Object::Reference::Type rt = object.languageUnref( this, gc() );
               if ( !gc() && rt == Object::Reference::NotReferenced )
                    dereferenceObject( &object );
               return rt;
          }

     ReferencedObjects& referencedObjects() { return mReferencedObjects; }

  private:
     bool mGC;
     mutable ReferencedObjects mReferencedObjects;
     mutable list<HObjectsReferences*> mHObjectsReferences;

     virtual void referenceObject( const Object* object ) const;
     virtual void dereferenceObject( const Object* object ) const;

}; // class LanguageInterface

//======================================================================

#define CLASS_METHODS_DESCRIPTIONS_DECLARATION                                                  \
  private:                                                                                      \
     static MethodDescription sMethods[];                                                       \
     static CallbackDescription sCallbacks[];                                                   \
     static ClassDescription sClassDescription;                                                 \
                                                                                                \
  public:                                                                                       \
     virtual const ClassDescription* classDescription() const { return &sClassDescription; }    \
     static const ClassDescription* thisClassDescription() { return &sClassDescription; }       \
     static const char* fileToInclude() { return __FILE__; }

//======================================================================

#endif // LanguageInterface_hh
