/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/common/Pointer.hh,v 1.5 2001/06/20 10:06:27 alex Exp $
 */

// -*- C++ -*-  Tue Mar 24 1998
// Eugene Skepner <eu@qub.com>
// $Id: Pointer.hh,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

/*
 * This module provides a template for smart pointers.  These smart pointers
 * keep a reference count and automatically delete themselves when that
 * reference count drops to zero.  Smart pointers also prevent duplicate
 * smart pointers pointing to the same object.
 */

#ifndef Pointer_hh
#define Pointer_hh

#ifndef Config_hh
#include "Config.hh"
#endif

#include <stdexcept>
#include <typeinfo>

//======================================================================

template <typename S> class Pointer;

/*
 * A Shared object is the base class for a smart pointer types.  It tracks a
 * reference count with each pointer to allow for automatic deletion and
 * prevent dangling pointers.
 */
class Shared
{
  private:
       // Returns 0 if p is allocated using operator new and -1 otherwise
     static int isOnHeap( const Shared* );
     static void onHeap( const void* p );

  public:
     Shared() {}
     Shared( const Shared& ) {}
     /*
      * When assigning one Shared instance to another Shared& variable, don't
      * give the new variable a copy of this one, instead force it to point
      * to the same instance, thus preventing duplicate copies of the same
      * Shared object.
      */
     Shared& operator = ( const Shared& ) { /* mRefs must not be changed */ return *this; }

     virtual ~Shared() {}

     /*
      * Since the overridden assignment operator above ensures no duplicates,
      * equality can be tested with a simple pointer compare.
      */
     bool operator == ( const Shared& toCompare ) const { return this == &toCompare; }
     bool operator != ( const Shared& toCompare ) const { return !operator == ( toCompare ); }

     /*
      * A reference count for a Shared object.
      */
     class Reference
     {
          friend class Shared;

       public:
          enum Type { Invalid, Automatic, Referenced, NotReferenced };

          Reference( int init = -2 ) : mRefs( init ) {}

          Type type() const
               {
                    switch ( mRefs ) {
                      case 0:
                           return NotReferenced;
                      case -1:
                           return Automatic;
                      case -2:
                           return Invalid;
                      default:
                           return Referenced;
                    }
               }

          Type operator ++ () { if ( mRefs >= 0 ) ++mRefs; return type(); }
          Type operator ++ ( int ) { Type result = type(); if ( mRefs >= 0 ) ++mRefs; return result; }
          Type operator -- () { if ( mRefs > 0 ) --mRefs; return type(); }
          Type operator -- ( int ) { Type result = type(); if ( mRefs > 0 ) --mRefs; return result; }

       private:
          int mRefs;

          void set( int aRefs ) { mRefs = aRefs; }
     };

     bool referenced() const
          {
               const Reference::Type t = mRefs.type();
               return t == Reference::Automatic || t == Reference::Referenced;
          }

#undef FRIEND_TEMPLATES
#ifdef FRIEND_TEMPLATES
  private:
     template <typename S> friend class Pointer<S>;
#endif

     static void throwPointerToAutoObjectsProhibited()
          {
               throw runtime_error( _("Smart pointers to automatic objects are prohibited") );
          }

#ifdef FRIEND_TEMPLATES
  protected:
#endif
     Reference::Type ref() const
          {
               if ( mRefs.type() == Reference::Invalid )
                    mRefs.set( isOnHeap( this ) );
               return ++mRefs;
          }

     Reference::Type unref() const
          {
               return --mRefs;
          }

  private:
     mutable Reference mRefs;

  public:
     void operator delete( void* p )
     {
          ::operator delete( p );
     }

     void* operator new( size_t size )
     {
          void* p = ::operator new( size );
          onHeap( p );
          return p;
     }

     void* operator new( size_t /* size */, void* place )
     {
          return place;
     }

}; // class Shared

//======================================================================

/*
 * A smart pointer to a Shared object.  Note that the type S must be a
 * descendant of Shared.  For the most part it simply overrides gobs of
 * operators such as = to ensure that references are always updated.  It
 * also provides some convenience operators to simplify boolean expressions
 * involving smart pointers.
 */
template <typename S> class Pointer
{
  public:
     Pointer( S* s = 0 )
          : m( s )
          {
               if ( m && m->ref() == Shared::Reference::Automatic )
                    S::throwPointerToAutoObjectsProhibited();
          }

     Pointer( const Pointer<S>& p )
          : m( p.m )
          {
               if ( m && m->ref() == Shared::Reference::Automatic )
                    S::throwPointerToAutoObjectsProhibited();
          }

     ~Pointer()
          {
               if ( m && m->referenced() && m->unref() == Shared::Reference::NotReferenced )
                    delete m;
          }

     Pointer<S>& operator = ( const Pointer<S>& p )
          {
               if ( m != p.m ) {
                    if ( m && m->unref() == Shared::Reference::NotReferenced )
                         delete m;
                    m = p.m;
                    if ( m )
                         m->ref();
               }
               return *this;
          }

     Pointer<S>& operator = ( S* s )
          {
               if ( m != s ) {
                    if ( m && m->unref() == Shared::Reference::NotReferenced )
                         delete m;
                    m = s;
                    if ( m )
                         m->ref();
               }
               return *this;
          }

//       template <typename S1> Pointer( const Pointer<S1>& p )
//            : m( dynamic_cast<S*>( const_cast<Pointer<S1>&>( p ).operator S1*() ) )
//            {
//                 if ( p && !m )
//                      throw runtime_error( string( "Invalid downcast from " ) + typeid( S1 ).name() + " to " + typeid( S ).name() );
//                 if ( m && m->ref() == Shared::Reference::Automatic )
//                      S::throwPointerToAutoObjectsProhibited();
//            }

     template <typename S1> Pointer( const Pointer<S1>& p )
          : m( const_cast<Pointer<S1>&>( p ).operator S1*() )
          {
               if ( m && m->ref() == Shared::Reference::Automatic )
                    S::throwPointerToAutoObjectsProhibited();
          }

     template <typename S1> Pointer<S>& operator = ( const Pointer<S1>& p )
          {
               if ( m != (S*)p.operator S1*() ) {
                    if ( m && m->unref() == Shared::Reference::NotReferenced )
                         delete m;
                    m = p.operator S1*();
                    if ( m )
                         m->ref();
               }
               return *this;
          }

     S* operator -> () { return m; }
     const S* operator -> () const { return m; }

     S& operator * () { return *m; }
     const S& operator * () const { return *m; }

     operator S* () { return m; }
     operator const S* () const { return m; }

     operator bool () { return m != 0; }
     operator bool () const { return m != 0; }
     bool operator ! () { return m == 0; }
     bool operator ! () const { return m == 0; }

     bool operator == ( const Pointer<S>& toCompare ) const { return m == toCompare.m; }
     bool operator == ( Pointer<S>& toCompare ) { return m == toCompare.m; }
     bool operator != ( const Pointer<S>& toCompare ) const { return !operator == ( toCompare ); }
     bool operator != ( Pointer<S>& toCompare ) { return !operator == ( toCompare ); }

     friend ostream& operator << ( ostream& os, const Pointer<S> p )
          {
               os << (*p);
               return os;
          }

  private:
     S* m;

}; // class Pointer<S>

//======================================================================

/*
 * A function object that works like so:
 *
 * class Foo;
 * class Bar;
 *
 * Pointer<Foo> first, second;
 *  ...
 * PointersLess<Pointer<Foo>> FooCompare;
 *  ...
 * if(FooCompare(first, second))
 *   printf("The data 'first' points to is less than the data 'second' points "
 *          "to.\n");
 *
 * It's ugly, but 'Less' functions are used often in the standard algorithms
 * for comparisons.
 */
template <class P> class PointersLess
{
  public:
     bool operator() ( const P& a, const P& b ) const
          {
               return *a < *b;
          }

}; // class PointersLess<P>

//======================================================================

#endif // Pointer_hh
