/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/common/ProgressIndicator.hh,v 1.3 2001/06/20 10:06:27 alex Exp $
 */

// -*- C++ -*-  Sun Jan  3 1999
// Eugene Skepner <eu@qub.com>
// $Id: ProgressIndicator.hh,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef ProgressIndicator_hh
#define ProgressIndicator_hh

#ifndef LanguageInterface_hh
#include "LanguageInterface.hh"
#endif

#ifndef Log_hh
#include "Log.hh"
#endif

#include <iostream.h>

#include <stack>

//======================================================================

class ProgressIndicator : virtual public Shared
{
  public:
     enum InfoType
     {
          itFile, itSubpackage, itReference, itAdditional,
          itLast
     };

     class ProgressInfo
     {
       public:
          ProgressInfo( size_t files = 0, size_t subpackages = 0, size_t references = 0, size_t additional = 0 )
               {
                    mInfo[itFile] = files;
                    mInfo[itSubpackage] = subpackages;
                    mInfo[itReference] = references;
                    mInfo[itAdditional] = additional;
                    mComplete = files + subpackages + references + additional;
               }

            // Computes percentage (0..1) of current progress by
            // comparing this and aComplete.
          float progress( const ProgressInfo& aComplete )
               {
                    return float( mComplete ) / float( aComplete.mComplete );
               }

          void progress( InfoType it ) { ++mInfo[it]; ++mComplete; }

       private:
          size_t mInfo[itLast];
          size_t mComplete;
     };

     ProgressIndicator( const string& aName ) : mName( aName ) {}

     const string& name() const { return mName; }

     virtual void initialize( const ProgressInfo& aInfo )
          {
               mComplete = aInfo;
          }

     virtual void progress( InfoType it )
          {
               mCurrent.progress( it );
               progress( mCurrent.progress( mComplete ) );
          }

       // Reports progress to user
     virtual void progress( float /*aProgress*/ )
          {
          }

  private:
     string mName;
     ProgressInfo mComplete;
     ProgressInfo mCurrent;

}; // class ProgressIndicator

//======================================================================

class ProgressIndicatorStack : virtual public LanguageInterface::Object
{
     CLASS_METHODS_DESCRIPTIONS_DECLARATION;

  public:
     ProgressIndicatorStack() {}

     static Pointer<ProgressIndicatorStack> make() { return new ProgressIndicatorStack(); }

     virtual Pointer<ProgressIndicator> push( const string& aName, const ProgressIndicator::ProgressInfo& aInfo )
          {
               Pointer<ProgressIndicator> i = newProgressIndicator( aName );
               i->initialize( aInfo );
               i->progress( 0.0 );
               mStack.push( i );
               return i;
          }

     virtual void pop()
          {
               mStack.pop();
          }

       // Update progress of topmost progress indicator
     virtual void progress( ProgressIndicator::InfoType it )
          {
               mStack.top()->progress( it );
          }

     virtual void progress( float aProgress )
          {
               mStack.top()->progress( aProgress );
          }

  protected:
     virtual Pointer<ProgressIndicator> newProgressIndicator( const string& aName ) { return new ProgressIndicator( aName ); }

  private:
     stack<Pointer<ProgressIndicator> > mStack;

}; // class ProgressIndicatorStack

//======================================================================

class ProgressIndicatorPusher
{
  public:
     ProgressIndicatorPusher( Pointer<ProgressIndicatorStack> aStack, const string& aName, const ProgressIndicator::ProgressInfo& aInfo )
          : mStack( aStack )
          {
               mStack->push( aName, aInfo );
          }

     ~ProgressIndicatorPusher()
          {
               mStack->pop();
          }

     Pointer<ProgressIndicatorStack> stack() { return mStack; }

  private:
     Pointer<ProgressIndicatorStack> mStack;

}; // class ProgressIndicatorPusher

//======================================================================

class ProgressIndicatorStream : public ProgressIndicator
{
  public:
     ProgressIndicatorStream( const string& aName, ostream& aStream, const string& aPrefix )
          : ProgressIndicator( aName ), mStream( aStream ), mPrefix( aPrefix ) {}

       // Reports progress to user
     virtual void progress( float aProgress )
          {
               if ( stream() )
                    stream() << prefix() << name() << ": " << size_t( aProgress * 100.0 ) << '%' << endl;
          }

  protected:
     ostream& stream() { return mStream; }
     const string& prefix() const { return mPrefix; }

  private:
     ostream& mStream;
     string mPrefix;

}; // class ProgressIndicatorStream

//======================================================================

class ProgressIndicatorStreamStack : public ProgressIndicatorStack
{
     CLASS_METHODS_DESCRIPTIONS_DECLARATION;

  public:
     ProgressIndicatorStreamStack( ostream& aStream, const string& aPrefix )
          : mStream( aStream ), mPrefix( aPrefix ) {}

     static Pointer<ProgressIndicatorStack> make( ostream& aStream = clog, const string& aPrefix = "PROGRESS " ) { return new ProgressIndicatorStreamStack( aStream, aPrefix ); }

  protected:
     virtual Pointer<ProgressIndicator> newProgressIndicator( const string& aName ) { return new ProgressIndicatorStream( aName, stream(), prefix() ); }

     ostream& stream() { return mStream; }
     const string& prefix() const { return mPrefix; }

  private:
     ostream& mStream;
     string mPrefix;

}; // class ProgressIndicatorStreamStack

//======================================================================

#endif // ProgressIndicator_hh
