/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/common/String.hh,v 1.3 2001/06/20 10:06:27 alex Exp $
 */

// -*- C++ -*-  Sun Nov 22 1998
// Eugene Skepner <eu@qub.com>
// $Id: String.hh,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

/*
 * This header provides several inline functions to perform conversions between
 * built-in types and the C++ string class.  The two main function sets are the
 * to_string() functions and the from_string() functions.  Some simple +
 * operator overloads are also provided to allow strings to be constructed from
 * variables just as they are in Java.
 */

#ifndef String_hh
#define String_hh

#include <string>
#include <cstdio>
#include <algorithm>

//======================================================================

/* 
 * to_string(variable) - Returns a string representation of the value of
 *   'variable'.  There is also a special case of to_string(int, int)
 *   which returns a string containing the representations of the two integers
 *   seperated by a space.
 */

inline string to_string( const short value )
{
     char s[10];
     ::sprintf( s, "%hd", value );
     return s;
}

inline string to_string( const unsigned short value )
{
     char s[10];
     ::sprintf( s, "%hu", value );
     return s;
}

inline string to_string( const int value )
{
     char s[20];
     ::sprintf( s, "%d", value );
     return s;
}

inline string to_string( const unsigned int value )
{
     char s[20];
     ::sprintf( s, "%u", value );
     return s;
}

inline string to_string( const unsigned int a, const unsigned int b )
{
     char s[40];
     ::sprintf( s, "%u %u", a, b );
     return s;
}

inline string to_string_octal( const unsigned int value )
{
     char s[20];
     ::sprintf( s, "0%o", value );
     return s;
}

inline string to_string( const long value )
{
     char s[40];
     ::sprintf( s, "%ld", value );
     return s;
}

inline string to_string( const unsigned long value )
{
     char s[40];
     ::sprintf( s, "%lu", value );
     return s;
}

inline string to_string( const float value )
{
     char s[40];
     ::sprintf( s, "%g", double( value ) );
     return s;
}

inline string to_string( const double value )
{
     char s[40];
     ::sprintf( s, "%g", value );
     return s;
}

inline string to_string( const void* value )
{
     char s[40];
     ::sprintf( s, "%p", value );
     return s;
}

inline string to_string( const bool value )
{
     return value ? "true" : "false";
}

//======================================================================

/*
 * <string> + variable - Returns a string holding
 *   <string> + to_string(variable).
 */

inline string operator + ( const string& a, short b )
{
     return a + to_string( b );
}

inline string operator + ( const string& a, unsigned short b )
{
     return a + to_string( b );
}

inline string operator + ( const string& a, int b )
{
     return a + to_string( b );
}

inline string operator + ( const string& a, unsigned b )
{
     return a + to_string( b );
}

inline string operator + ( const string& a, long b )
{
     return a + to_string( b );
}

inline string operator + ( const string& a, unsigned long b )
{
     return a + to_string( b );
}

inline string operator + ( const string& a, double b )
{
     return a + to_string( b );
}

//======================================================================

/*
 * from_string(string, type, [char ** end], [base]) - Returns the value
 *   represented in 'string' as a variable of type 'type'.  If the optional
 *   'end' parameter is specified, *end will point to the first invalid
 *   character in 'string'.  The optional 'base' parameter specifies the base
 *   to be used for the conversion and defaults to 10.  Note that for some
 *   types, such as float and double, the base parameter is ignored and
 *   conversion is always done using base 10.
 */

inline int from_string( const string& str, int, const char** end = 0, int base = 10 )
{
     return strtol( str.c_str(), const_cast<char**>( end ), base );
}

inline unsigned from_string( const string& str, unsigned, const char** end = 0, int base = 10 )
{
     return strtoul( str.c_str(), const_cast<char**>( end ), base );
}

inline long from_string( const string& str, long, const char** end = 0, int base = 10 )
{
     return strtol( str.c_str(), const_cast<char**>( end ), base );
}

inline unsigned long from_string( const string& str, unsigned long, const char** end = 0, int base = 10 )
{
     return strtoul( str.c_str(), const_cast<char**>( end ), base );
}

inline float from_string( const string& str, float, const char** end = 0, int /*base*/ = 10 )
{
     return strtod( str.c_str(), const_cast<char**>( end ) );
}

inline double from_string( const string& str, double, const char** end = 0, int /*base*/ = 10 )
{
     return strtod( str.c_str(), const_cast<char**>( end ) );
}

//======================================================================

/*
 * Replaces all instances of the character 'from' in the string 'src' to
 * the character 'to' in a copy of 'src' and returns the changed copy.
 */
inline string replace_chars( const string& src, char from, char to )
{
     string result = src;
     replace( result.begin(), result.end(), from, to );
     return result;
}

//======================================================================

/*
 * Returns a copy of the string 'src' which has been converted to all
 * uppercase.
 */
inline string to_upper( const string& src )
{
     string result( src.length(), '#' );
     transform( src.begin(), src.end(), result.begin(), &toupper );
     return result;
}

//======================================================================

#endif // String_hh
