/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/database/DatabaseEngine.hh,v 1.3 2001/06/20 10:08:11 alex Exp $
 */

// -*- C++ -*-  Tue Aug 11 1998
// $Id: DatabaseEngine.hh,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef DatabaseEngine_hh
#define DatabaseEngine_hh

#ifndef Container_hh
#include "Container.hh"
#endif

#include <ctime>
#include <vector>

//======================================================================

class DatabaseEngine : virtual public Shared
{
  public:
     enum FieldType
     {
          tId, tString, tInteger, tLongLong, tTime
     };

     enum FieldKey
     {
          kNone, kUnique, kDuplicate
     };

       // --------------------------------------------------

     typedef unsigned long Id;
     typedef Container< set<Id> > IdList;

       // --------------------------------------------------

     struct FieldDescription
     {
          typedef FieldType Type;
          typedef FieldKey Key;

          const char* name;
          Type type;
          Key key;
          bool null;            // field can store NULLs

     }; // struct FieldDescription

       // --------------------------------------------------

     struct TableDescription
     {
          const char* name;
          const FieldDescription* fields;

          size_t numberOfFields() const;
          size_t fieldByName( const char* aFieldName ) const;

     }; // struct TableDescription

       // --------------------------------------------------

     struct FieldValue
     {
          FieldValue() : null( false ), present( false ) {}
          FieldValue( Id aId ) : null( false ), present( true ) { value.id = aId; }
          FieldValue( const string& aStr ) : null( false ), present( true ) { str = aStr; }

          union {
               Id id;
               int integer;
               long long longlong;
               time_t time;
          } value;
          string str;
          bool null;
          bool present;
     };

       // --------------------------------------------------

     class Record : virtual public Shared
     {
       public:
          virtual Id id() const = 0;
          virtual void id( Id aId ) const = 0;

            // Returns if value of field is null
          virtual bool null( const char* aFieldName ) const = 0;

            // Check if field of the required type and returns value
            // raises exception if value is null
          virtual Id id( const char* aFieldName ) const = 0;
          virtual const string& str( const char* aFieldName ) const = 0;
          virtual int integer( const char* aFieldName ) const = 0;
          virtual long long longlong( const char* aFieldName ) const = 0;
          virtual time_t time( const char* aFieldName ) const = 0;

            // Sets the value of the field to null if it is allowed
          virtual void setNull( const char* aFieldName ) = 0;

            // Marks field as not present, i.e. next update will not
            // touch this field
          virtual void not_present( const char* aFieldName ) = 0;

            // Check if field of the required type and sets value
          virtual void id( const char* aFieldName, Id aId ) = 0;
          virtual void str( const char* aFieldName, const string& s ) = 0;
          virtual void integer( const char* aFieldName, int i ) = 0;
          virtual void longlong( const char* aFieldName, long long ll ) = 0;
          virtual void time( const char* aFieldName, time_t t ) = 0;

            //virtual bool present() const { returns mPresent; }
            //virtual void present( bool aPresent ) { mPresent = aPresent; }

       protected:
          Record( const TableDescription& aTableDescription ) : mTableDescription( aTableDescription ) {}

          const TableDescription& description() const { return mTableDescription; }

       private:
            //bool mPresent;
          const TableDescription& mTableDescription;

     }; // class Record

     typedef Container<vector<Pointer<Record> > > RecordList;

       // --------------------------------------------------

     class Table
     {
       public:
          virtual ~Table() {}

            // Returns new database implemetation specific record
          virtual Pointer<Record> newRecord() const = 0;

            // Selects all the fields
          virtual Pointer<RecordList> select( const char* aFieldName, const FieldValue& aValue ) = 0;
          virtual Pointer<RecordList> select( const char* aFieldName1, const FieldValue& aValue1, const char* aFieldName2, const FieldValue& aValue2 ) = 0;
          virtual Pointer<RecordList> selectAll() = 0; // select all records
          virtual Pointer<Record> select( Id id ) = 0;
            // Selects specified fields
          virtual Pointer<RecordList> select( const char* aFieldName, const FieldValue& aValue, const vector<string>& aFieldsToSelect ) = 0;
          virtual Pointer<RecordList> select( const char* aFieldName, const FieldValue& aValue, const string& aFieldToSelect ) = 0;
          virtual Pointer<Record> select( Id id, const vector<string>& aFieldsToSelect ) = 0;
          virtual Pointer<Record> select( Id id, const string& aFieldToSelect ) = 0;
            // Selects all ids from the table
          virtual Pointer<IdList> select() = 0;
          virtual size_t number_of_records() = 0;

            // Returns if record with given id is present.
          virtual bool present( Id id ) = 0;

            // Inserts new record, makes sure unique fields will have
            // no duplicates, not null fields receive value
          virtual void insert( const Record& aRecord, bool use_id_of_record = false ) = 0;

            // Updates record
          virtual void update( const Record& aRecord ) = 0;

            // Removes all the records matching
          virtual void remove( const char* aFieldName, const FieldValue& aValue ) = 0;
          virtual void remove( const char* aFieldName1, const FieldValue& aValue1, const char* aFieldName2, const FieldValue& aValue2 ) = 0;
          virtual void remove( Id id ) = 0;
          virtual void remove( const IdList& ids ) = 0;

          virtual void sync() = 0;

          const char* name() const { return mTableDescription.name; }

          virtual Id internalId() = 0;
          virtual void internalId( Id id ) = 0;

       protected:
          Table( const TableDescription& aTableDescription ) : mTableDescription( aTableDescription ) {}

          const TableDescription& description() const { return mTableDescription; }

       private:
          const TableDescription& mTableDescription;

     }; // class Table

       // --------------------------------------------------

     DatabaseEngine( const string& aBaseDir ) : mBaseDir( aBaseDir ), mLockFile( -1 ) {}
     virtual ~DatabaseEngine() {}

     virtual void lock();
     virtual void unlock();

     virtual void create( const TableDescription& aTableDescription ) = 0;
     virtual bool exists( const TableDescription& aTableDescription ) = 0;
     virtual Table* table( const TableDescription& aTableDescription ) = 0;

  protected:
     const string& baseDir() const { return mBaseDir; }

  private:
     string mBaseDir;
     int mLockFile;

}; // class DatabaseEngine

//======================================================================

#endif // DatabaseEngine_hh
