/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/database/DatabaseEngineDb185.cc,v 1.4 2001/06/20 10:08:11 alex Exp $
 */

// -*- C++ -*-   Tue Aug 11 1998
// $Id: DatabaseEngineDb185.cc,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef DatabaseEngineDb185_hh
#include "DatabaseEngineDb185.hh"
#endif

#ifndef Log_hh
#include "Log.hh"
#endif

#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <algorithm>

//======================================================================

DatabaseEngineDb185::DatabaseEngineDb185( const string& aBaseDir, bool /*aCreate*/ )
	: DatabaseEngine( aBaseDir )
{
	mInfoId.flags = 0;
	mInfoId.cachesize = 0;
	mInfoId.maxkeypage = 0;
	mInfoId.minkeypage = 0;
	mInfoId.psize = 0;
	mInfoId.compare = &id_compare;
	mInfoId.prefix = 0;
	mInfoId.lorder = 0;

	mInfoDup.flags = R_DUP;
	mInfoDup.cachesize = 0;
	mInfoDup.maxkeypage = 0;
	mInfoDup.minkeypage = 0;
	mInfoDup.psize = 0;
	mInfoDup.compare = 0;
	mInfoDup.prefix = 0;
	mInfoDup.lorder = 0;

	mInfoDupId.flags = R_DUP;
	mInfoDupId.cachesize = 0;
	mInfoDupId.maxkeypage = 0;
	mInfoDupId.minkeypage = 0;
	mInfoDupId.psize = 0;
	mInfoDupId.compare = &id_compare;
	mInfoDupId.prefix = 0;
	mInfoDupId.lorder = 0;

} // DatabaseEngineDb185::DatabaseEngineDb185

//======================================================================

DatabaseEngineDb185::~DatabaseEngineDb185()
{
} // DatabaseEngineDb185::~DatabaseEngineDb185

//======================================================================

string DatabaseEngineDb185::fieldFilename( const string& aTableName, const string& aFieldName, bool reverse )
{
	return baseDir() + "/" + aTableName + "-" + aFieldName + ( reverse ? "-key" : "" );

} // DatabaseEngineDb185::fieldFilename

//======================================================================

void DatabaseEngineDb185::create( const TableDescription& aTableDescription )
{
	for ( const FieldDescription* field = aTableDescription.fields; field->name; ++field ) {
		DB* table;
		const void* dbinfo = &mInfoId;
		if ( ( table = dbopen( fieldFilename( aTableDescription.name, field->name, false ).c_str(), O_CREAT|O_RDWR, 0644, DB_BTREE, dbinfo ) ) == 0 ) {
			throw runtime_error( string( "Cannot create database table " ) + aTableDescription.name + " field " + field->name + ": " + strerror( errno ) );
		}
		table->close( table );

		if ( field->key == kUnique || field->key == kDuplicate ) {
			DB* table;
			const void* dbinfo = 0;
			if ( field->type == tId && field->key == kDuplicate )
				dbinfo = &mInfoDupId;
			else if ( field->type == tId )
				dbinfo = &mInfoId;
			else if ( field->key == kDuplicate )
				dbinfo = &mInfoDup;
			if ( ( table = dbopen( fieldFilename( aTableDescription.name, field->name, true ).c_str(), O_CREAT|O_RDWR, 0644, DB_BTREE,  dbinfo ) ) == 0 )
				throw runtime_error( string( "Cannot create database table " ) + aTableDescription.name + " field-key " + field->name + ": " + strerror( errno ) );
			table->close( table );
		}
	}

} // DatabaseEngineDb185::create

//======================================================================

bool DatabaseEngineDb185::exists( const TableDescription& aTableDescription )
{
	for ( const FieldDescription* field = aTableDescription.fields; field->name; ++field ) {
		if ( access( ( baseDir() + "/" + aTableDescription.name + "-" + field->name ).c_str(), R_OK|W_OK ) != 0 )
			return(false);
		if ( field->key == kUnique || field->key == kDuplicate )
			if ( access( ( baseDir() + "/" + aTableDescription.name + "-" + field->name + "-key" ).c_str(), R_OK|W_OK ) != 0 )
				return(false);
	}
	return(true);

} // DatabaseEngineDb185::exists

//======================================================================

DatabaseEngine::Table* DatabaseEngineDb185::table( const TableDescription& aTableDescription )
{
	return new Table( aTableDescription, *this );

} // DatabaseEngineDb185::table

//======================================================================

DatabaseEngineDb185::Record::Record( const TableDescription& aTableDescription )
	: DatabaseEngine::Record( aTableDescription ), mId( 0 ),
	mValues( new FieldValue[ aTableDescription.numberOfFields() ] )
{
} // DatabaseEngineDb185::Record::Record

//======================================================================

DatabaseEngineDb185::Record::~Record()
{
	delete [] mValues;

} // DatabaseEngineDb185::Record::~Record

//======================================================================

DatabaseEngine::Id DatabaseEngineDb185::Record::id() const
{
	return(mId);

} // DatabaseEngineDb185::Record::id

//======================================================================

bool DatabaseEngineDb185::Record::null( const char* aFieldName ) const
{
	const FieldValue& fv = mValues[ description().fieldByName( aFieldName ) ];
	if ( !fv.present )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::null: Field " ) + aFieldName + " not present in the record" );
	return(fv.null);

} // DatabaseEngineDb185::Record::null

//======================================================================

DatabaseEngine::Id DatabaseEngineDb185::Record::id( const char* aFieldName ) const
{
	const size_t fieldno = description().fieldByName( aFieldName );
	const FieldValue& fv = mValues[ fieldno ];
	if ( !fv.present )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::id: Field " ) + aFieldName + " not present in the record" );
	if ( description().fields[fieldno].type != tId )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::id: Field " ) + aFieldName + " is not of type id" );
	return(fv.value.id);

} // DatabaseEngineDb185::Record::id

//======================================================================

const string& DatabaseEngineDb185::Record::str( const char* aFieldName ) const
{
	const size_t fieldno = description().fieldByName( aFieldName );
	const FieldValue& fv = mValues[ fieldno ];
	if ( !fv.present )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::str: Field " ) + aFieldName + " not present in the record" );
	if ( description().fields[fieldno].type != tString )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::str: Field " ) + aFieldName + " is not of type String" );
	return(fv.str);

} // DatabaseEngineDb185::Record::str

//======================================================================

int DatabaseEngineDb185::Record::integer( const char* aFieldName ) const
{
	const size_t fieldno = description().fieldByName( aFieldName );
	const FieldValue& fv = mValues[ fieldno ];
	if ( !fv.present )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::integer: Field " ) + aFieldName + " not present in the record" );
	if ( description().fields[fieldno].type != tInteger )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::integer: Field " ) + aFieldName + " is not of type Integer" );
	return(fv.value.integer);

} // DatabaseEngineDb185::Record::integer

//======================================================================

long long DatabaseEngineDb185::Record::longlong( const char* aFieldName ) const
{
	const size_t fieldno = description().fieldByName( aFieldName );
	const FieldValue& fv = mValues[ fieldno ];
	if ( !fv.present )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::longlong: Field " ) + aFieldName + " not present in the record" );
	if ( description().fields[fieldno].type != tLongLong )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::longlong: Field " ) + aFieldName + " is not of type LongLong" );
	return(fv.value.longlong);

} // DatabaseEngineDb185::Record::longlong

//======================================================================

time_t DatabaseEngineDb185::Record::time( const char* aFieldName ) const
{
	const size_t fieldno = description().fieldByName( aFieldName );
	const FieldValue& fv = mValues[ fieldno ];
	if ( !fv.present )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::time: Field " ) + aFieldName + " not present in the record" );
	if ( description().fields[fieldno].type != tTime )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::time: Field " ) + aFieldName + " is not of type Time" );
	return(fv.value.time);

} // DatabaseEngineDb185::Record::time

//======================================================================

void DatabaseEngineDb185::Record::setNull( const char* aFieldName )
{
	FieldValue& fv = mValues[ description().fieldByName( aFieldName ) ];
	fv.present = true;
	fv.null = true;

} // DatabaseEngineDb185::Record::setNull

//======================================================================

void DatabaseEngineDb185::Record::not_present( const char* aFieldName )
{
	FieldValue& fv = mValues[ description().fieldByName( aFieldName ) ];
	fv.present = false;

} // DatabaseEngineDb185::Record::not_present

//======================================================================

void DatabaseEngineDb185::Record::id( const char* aFieldName, Id aId )
{
	const size_t fieldno = description().fieldByName( aFieldName );
	FieldValue& fv = mValues[ fieldno ];
	if ( description().fields[fieldno].type != tId )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::id: Field " ) + aFieldName + " is not of type Id" );
	fv.present = true;
	fv.null = false;
	fv.value.id = aId;

} // DatabaseEngineDb185::Record::id

//======================================================================

void DatabaseEngineDb185::Record::str( const char* aFieldName, const string& s )
{
	const size_t fieldno = description().fieldByName( aFieldName );
	FieldValue& fv = mValues[ fieldno ];
	if ( description().fields[fieldno].type != tString )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::string: Field " ) + aFieldName + " is not of type String" );
	fv.present = true;
	fv.null = false;
	fv.str = s;

} // DatabaseEngineDb185::Record::str

//======================================================================

void DatabaseEngineDb185::Record::integer( const char* aFieldName, int i )
{
	const size_t fieldno = description().fieldByName( aFieldName );
	FieldValue& fv = mValues[ fieldno ];
	if ( description().fields[fieldno].type != tInteger )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::integer: Field " ) + aFieldName + " is not of type Integer" );
	fv.present = true;
	fv.null = false;
	fv.value.integer = i;

} // DatabaseEngineDb185::Record::integer

//======================================================================

void DatabaseEngineDb185::Record::longlong( const char* aFieldName, long long ll )
{
	const size_t fieldno = description().fieldByName( aFieldName );
	FieldValue& fv = mValues[ fieldno ];
	if ( description().fields[fieldno].type != tLongLong )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::longlong: Field " ) + aFieldName + " is not of type LongLong" );
	fv.present = true;
	fv.null = false;
	fv.value.longlong = ll;

} // DatabaseEngineDb185::Record::longlong

//======================================================================

void DatabaseEngineDb185::Record::time( const char* aFieldName, time_t t )
{
	const size_t fieldno = description().fieldByName( aFieldName );
	FieldValue& fv = mValues[ fieldno ];
	if ( description().fields[fieldno].type != tTime )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Record::time: Field " ) + aFieldName + " is not of type Time" );
	fv.present = true;
	fv.null = false;
	fv.value.time = t;

} // DatabaseEngineDb185::Record::time

//======================================================================

void DatabaseEngineDb185::Record::set( const DBT& dbt, const char* aFieldName )
{
	const size_t fieldno = description().fieldByName( aFieldName );
	dbt2field( dbt, description().fields[fieldno], mValues[ fieldno ] );

} // DatabaseEngineDb185::Record::set

//======================================================================

void DatabaseEngineDb185::Record::get( DBT& dbt, const char* aFieldName ) const
{
	const size_t fieldno = description().fieldByName( aFieldName );
	field2dbt( dbt, description().fields[fieldno], mValues[ fieldno ] );

} // DatabaseEngineDb185::Record::get

//======================================================================

const DatabaseEngine::Id DatabaseEngineDb185::Table::sIdId = 1;

//======================================================================

DatabaseEngineDb185::Table::Table( const TableDescription& aTableDescription, DatabaseEngineDb185& aEngine )
	: DatabaseEngine::Table( aTableDescription ), mDirect( 0 ), mReverse( 0 ), mWriteAccess( 0 )
{
	const size_t numFields = description().numberOfFields();

	mDirect = new DB*[numFields];
	mReverse = new DB*[numFields];
	for ( size_t i = 0; i < numFields; ++i ) {
		mDirect[i] = 0;
		mReverse[i] = 0;
	}

	for ( size_t i = 0; i < numFields; ++i ) {
		const string filename = aEngine.fieldFilename( description().name, description().fields[i].name, false );
		const void* dbinfo = 0;
		if ( description().fields[i].type == tId )
			dbinfo = &aEngine.mInfoId;
		if ( ( mDirect[i] = dbopen( filename.c_str(), O_RDWR, 0644, DB_BTREE, dbinfo ) ) == 0 )
			throw runtime_error( string( "Cannot open database file " ) + filename + ": " + strerror( errno ) );

		if ( description().fields[i].key == kUnique || description().fields[i].key == kDuplicate ) {
			const string filename = aEngine.fieldFilename( description().name, description().fields[i].name, true );
			const void* dbinfo = 0;
			if ( description().fields[i].type == tId && description().fields[i].key == kDuplicate )
				dbinfo = &aEngine.mInfoDupId;
			else if ( description().fields[i].type == tId )
				dbinfo = &aEngine.mInfoId;
			else if ( description().fields[i].key == kDuplicate )
				dbinfo = &aEngine.mInfoDup;
			if ( ( mReverse[i] = dbopen( filename.c_str(), O_RDWR, 0644, DB_BTREE, dbinfo ) ) == 0 )
				throw runtime_error( string( "Cannot open database file " ) + filename + ": " + strerror( errno ) );
		}
	}

} // DatabaseEngineDb185::Table::Table

//======================================================================

DatabaseEngineDb185::Table::~Table()
{
	for ( size_t i = 0; description().fields[i].name; ++i ) {
		if ( mDirect[i] )
			mDirect[i]->close( mDirect[i] );
		if ( mReverse[i] )
			mReverse[i]->close( mReverse[i] );
	}
	delete [] mDirect;
	delete [] mReverse;

} // DatabaseEngineDb185::Table::~Table

//======================================================================

Pointer<DatabaseEngine::Record> DatabaseEngineDb185::Table::newRecord() const
{
	return new Record( description() );

} // DatabaseEngineDb185::Table::newRecord

//======================================================================

class records_by_id
{
public:
	records_by_id( DatabaseEngineDb185::Table& aTable ) : mTable( aTable ) {}
	Pointer<DatabaseEngine::Record> operator () ( DatabaseEngine::Id id ) { return mTable.select( id ); }
private:
	DatabaseEngineDb185::Table& mTable;
};

//======================================================================

Pointer<DatabaseEngine::RecordList> DatabaseEngineDb185::Table::select( const char* aFieldName, const FieldValue& aValue )
{
	const size_t fieldno = description().fieldByName( aFieldName );
	if ( description().fields[fieldno].key != kUnique && description().fields[fieldno].key != kDuplicate )
		throw runtime_error( string( "Unable to select from table " ) + description().name + " by field " + aFieldName + ": field is not key" );

	IdList ids;
	idsByField( aFieldName, aValue, ids );

	Pointer<RecordList> vr = new RecordList;
	transform( ids.begin(), ids.end(), back_inserter( *vr ), records_by_id( *this ) );

	return vr;

} // DatabaseEngineDb185::Table::select

//======================================================================

Pointer<DatabaseEngine::RecordList> DatabaseEngineDb185::Table::select( const char* aFieldName1, const FieldValue& aValue1, const char* aFieldName2, const FieldValue& aValue2 )
{
	const size_t fieldno1 = description().fieldByName( aFieldName1 );
	if ( description().fields[fieldno1].key != kUnique && description().fields[fieldno1].key != kDuplicate )
		throw runtime_error( string( "Unable to select from table " ) + description().name + " by field " + aFieldName1 + ": field is not key" );
	const size_t fieldno2 = description().fieldByName( aFieldName2 );
	if ( description().fields[fieldno2].key != kUnique && description().fields[fieldno2].key != kDuplicate )
		throw runtime_error( string( "Unable to select from table " ) + description().name + " by field " + aFieldName2 + ": field is not key" );

	IdList ids;
	idsByField( aFieldName1, aValue1, aFieldName2, aValue2, ids );

	Pointer<RecordList> vr = new RecordList;
	transform( ids.begin(), ids.end(), back_inserter( *vr ), records_by_id( *this ) );

	return vr;

} // DatabaseEngineDb185::Table::select

//======================================================================

Pointer<DatabaseEngine::RecordList> DatabaseEngineDb185::Table::selectAll()
{
	Pointer<IdList> ids = select();
	Pointer<RecordList> vr = new RecordList;
	transform( ids->begin(), ids->end(), back_inserter( *vr ), records_by_id( *this ) );
	return vr;

} // DatabaseEngineDb185::Table::selectAll

//======================================================================

Pointer<DatabaseEngine::Record> DatabaseEngineDb185::Table::select( Id id )
{
	Pointer<Record> rec = new Record( description() );
	rec->id( id );
	const size_t numberOfFields = description().numberOfFields();
	DBT key = { &id, sizeof id };
	for ( size_t fieldno = 0; fieldno < numberOfFields; ++fieldno ) {
		if ( mDirect[fieldno] ) {
			DBT data;
			int err = mDirect[fieldno]->get( mDirect[fieldno], &key, &data, 0 );
			switch ( err ) {
			case 0:
				rec->set( data, description().fields[fieldno].name );
				break;
			case RET_SPECIAL: // not found
				rec->setNull( description().fields[fieldno].name );
				break;
			default:
				throw runtime_error( string( "Unable to select by id from table " ) + description().name + ": " + strerror( errno ) );
			}
		}
	}

	return(rec);

} // DatabaseEngineDb185::Table::select

//======================================================================

bool DatabaseEngineDb185::Table::present( Id id )
{
	bool result = true;
	const size_t numberOfFields = description().numberOfFields();
	DBT key = { &id, sizeof id };
	for ( size_t fieldno = 0; fieldno < numberOfFields && result; ++fieldno ) {
		if ( !description().fields[fieldno].null && mDirect[fieldno] ) {
			DBT data;
			int err = mDirect[fieldno]->get( mDirect[fieldno], &key, &data, 0 );
			switch ( err ) {
			case 0: // present for this field
				break;
			case RET_SPECIAL: // not present for this field
				result = false;
				break;
			default:
				throw runtime_error( string( "Unable to select by id from table " ) + description().name + ": " + strerror( errno ) );
			}
		}
	}
	return(result);

} // DatabaseEngineDb185::Table::present

//======================================================================

void DatabaseEngineDb185::Table::idsByField( const char* aFieldName, const FieldValue& aValue, IdList& ids )
{
	const size_t fieldno = description().fieldByName( aFieldName );
	if ( mReverse[fieldno] == 0 )
		throw runtime_error( string( "Unable to select from table " ) + description().name + " by field " + aFieldName + ": no key tables available" );
	DBT key_field;
	field2dbt( key_field, description().fields[fieldno], aValue );
	const DBT key_field_orig = key_field;
	DBT data_field;
	if ( mReverse[fieldno]->seq( mReverse[fieldno], &key_field, &data_field, R_CURSOR ) == 0 ) {
		if ( equal( key_field, key_field_orig ) ) {
			::insert( ids, dbt2id( data_field ) );
			while ( mReverse[fieldno]->seq( mReverse[fieldno], &key_field, &data_field, R_NEXT ) == 0 && equal( key_field, key_field_orig ) ) {
				::insert( ids, dbt2id( data_field ) );
			}
		}
	}

} // DatabaseEngineDb185::Table::idsByField

//======================================================================

void DatabaseEngineDb185::Table::idsByField( const char* aFieldName1, const FieldValue& aValue1, const char* aFieldName2, const FieldValue& aValue2, IdList& ids )
{
	IdList ids1;
	idsByField( aFieldName1, aValue1, ids1 );
	IdList ids2;
	idsByField( aFieldName2, aValue2, ids2 );

	set_intersection( ids1.begin(), ids1.end(), ids2.begin(), ids2.end(), inserter( ids ) );

} // DatabaseEngineDb185::Table::idsByField

//======================================================================

class records_by_id_fields
{
public:
	records_by_id_fields( DatabaseEngineDb185::Table& aTable, const vector<string>& aFieldsToSelect ) : mTable( aTable ), mFieldsToSelect( aFieldsToSelect ) {}
	Pointer<DatabaseEngine::Record> operator () ( DatabaseEngine::Id id ) { return mTable.select( id, mFieldsToSelect ); }
private:
	DatabaseEngineDb185::Table& mTable;
	const vector<string>& mFieldsToSelect;
};

//======================================================================

Pointer<DatabaseEngine::RecordList> DatabaseEngineDb185::Table::select( const char* aFieldName, const FieldValue& aValue, const vector<string>& aFieldsToSelect )
{
	const size_t fieldno = description().fieldByName( aFieldName );
	if ( description().fields[fieldno].key != kUnique && description().fields[fieldno].key != kDuplicate )
		throw runtime_error( string( "Unable to select from table " ) + description().name + " by field " + aFieldName + ": field is not key" );

	IdList ids;
	idsByField( aFieldName, aValue, ids );

	Pointer<RecordList> vr = new RecordList;
	transform( ids.begin(), ids.end(), back_inserter( *vr ), records_by_id_fields( *this, aFieldsToSelect ) );

	return vr;

} // DatabaseEngineDb185::Table::select

//======================================================================

class records_by_id_field
{
public:
	records_by_id_field( DatabaseEngineDb185::Table& aTable, const string& aFieldToSelect ) : mTable( aTable ), mFieldToSelect( aFieldToSelect ) {}
	Pointer<DatabaseEngine::Record> operator () ( DatabaseEngine::Id id ) { return mTable.select( id, mFieldToSelect ); }
private:
	DatabaseEngineDb185::Table& mTable;
	const string& mFieldToSelect;
};

//======================================================================

Pointer<DatabaseEngine::RecordList> DatabaseEngineDb185::Table::select( const char* aFieldName, const FieldValue& aValue, const string& aFieldToSelect )
{
	const size_t fieldno = description().fieldByName( aFieldName );
	if ( description().fields[fieldno].key != kUnique && description().fields[fieldno].key != kDuplicate )
		throw runtime_error( string( "Unable to select from table " ) + description().name + " by field " + aFieldName + ": field is not key" );

	IdList ids;
	idsByField( aFieldName, aValue, ids );

	Pointer<RecordList> vr = new RecordList;
	transform( ids.begin(), ids.end(), back_inserter( *vr ), records_by_id_field( *this, aFieldToSelect ) );

	return vr;

} // DatabaseEngineDb185::Table::select

//======================================================================

Pointer<DatabaseEngine::Record> DatabaseEngineDb185::Table::select( Id id, const vector<string>& aFieldsToSelect )
{
	Pointer<Record> rec = new Record( description() );
	rec->id( id );
	DBT key;
	id2dbt( key, id );
	for ( vector<string>::const_iterator f = aFieldsToSelect.begin(); f != aFieldsToSelect.end(); ++f ) {
		const size_t fieldno = description().fieldByName( f->c_str() );
		if ( mDirect[fieldno] ) {
			DBT data;
			int err = mDirect[fieldno]->get( mDirect[fieldno], &key, &data, 0 );
			switch ( err ) {
			case 0:
				rec->set( data, f->c_str() );
				break;
			case RET_SPECIAL: // key not found
				rec->setNull( f->c_str() );
				break;
			default:
				throw runtime_error( string( "Unable to select by id from table " ) + description().name + ": " + strerror( errno ) );
			}
		}
	}
	return(rec);

} // DatabaseEngineDb185::Table::select

//======================================================================

Pointer<DatabaseEngine::Record> DatabaseEngineDb185::Table::select( Id id, const string& aFieldToSelect )
{
	Pointer<Record> rec = new Record( description() );
	rec->id( id );
	DBT key;
	id2dbt( key, id );
	const size_t fieldno = description().fieldByName( aFieldToSelect.c_str() );
	if ( mDirect[fieldno] ) {
		DBT data;
		int err = mDirect[fieldno]->get( mDirect[fieldno], &key, &data, 0 );
		switch ( err ) {
		case 0:
			rec->set( data, aFieldToSelect.c_str() );
			break;
		case RET_SPECIAL:
			rec->setNull( aFieldToSelect.c_str() );
			break;
		default:
			throw runtime_error( string( "Unable to select by id from table " ) + description().name + ": " + strerror( errno ) );
		}
	}
	return(rec);

} // DatabaseEngineDb185::Table::select

//======================================================================

Pointer<DatabaseEngine::IdList> DatabaseEngineDb185::Table::select()
{
	if ( mDirect[0] ) {
		Pointer<IdList> ids = new IdList;
		DBT key, data;
		unsigned flags = R_FIRST;
		while ( mDirect[0]->seq( mDirect[0], &key, &data, flags ) == 0 ) {
			flags = R_NEXT;
			Id id = dbt2id( key );
			if ( id != sIdId )
				::insert( *ids, id );
		}

		return(ids);
	}
	else
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Table::select: Unable to select all ids for table " ) + description().name + ": no subtables available" );

} // DatabaseEngineDb185::Table::select

//======================================================================

size_t DatabaseEngineDb185::Table::number_of_records()
{
	if ( mDirect[0] ) {
		DBT key, data;
		unsigned flags = R_FIRST;
		size_t records = 0;
		while ( mDirect[0]->seq( mDirect[0], &key, &data, flags ) == 0 ) {
			flags = R_NEXT;
			Id id = dbt2id( key );
			if ( id != sIdId )
				++records;
		}

		return(records);
	}
	else
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Table::number_of_records: Unable to get number of records for the table " ) + description().name + ": no subtables available" );

} // DatabaseEngineDb185::Table::number_of_records

//======================================================================

void DatabaseEngineDb185::Table::check_record_requirements( const Record& aRecord, bool check_for_presence_of_fields_values ) const
{
	const size_t numberOfFields = description().numberOfFields();

	// Make sure new record will not violate requirements
	for ( size_t fieldno = 0; fieldno < numberOfFields; ++fieldno ) {
		if ( check_for_presence_of_fields_values && ! aRecord.value( fieldno ).present )
			throw runtime_error( string( "Cannot insert into table " ) + description().name + ": value for field " + description().fields[fieldno].name + " is not present" );
		if ( ! description().fields[fieldno].null && aRecord.value( fieldno ).null )
			throw runtime_error( string( "Cannot insert into table " ) + description().name + ": value for not null field " + description().fields[fieldno].name + " is null" );
		if ( description().fields[fieldno].key == kUnique && mReverse[fieldno] != 0 && !aRecord.value( fieldno ).null ) {
			DBT key;
			field2dbt( key, description().fields[fieldno], aRecord.value( fieldno ) );
			DBT data;
			int err = mReverse[fieldno]->get( mReverse[fieldno], &key, &data, 0 );
			switch ( err ) {
			case 0:
				throw runtime_error( string( "Cannot insert into table " ) + description().name + ": value for unique key field " + description().fields[fieldno].name + " already exists" );
			case RET_SPECIAL: // not found
				break;
			default:
				throw runtime_error( string( "Cannot get table " ) + description().name + " field " + description().fields[fieldno].name + ": " + strerror( errno ) );
			}
		}
	}

} // DatabaseEngineDb185::Table::check_record_requirements

//======================================================================

void DatabaseEngineDb185::Table::insert( const DatabaseEngine::Record& aRecord, bool use_id_of_record )
{
	writeAccess();

	const Record& rec = dynamic_cast<const Record&>( aRecord );
	check_record_requirements( rec, true );

	if ( !use_id_of_record )
		rec.id( newId() );
	DBT key_id;
	id2dbt( key_id, rec.id() );
	const size_t numberOfFields = description().numberOfFields();
	for ( size_t fieldno = 0; fieldno < numberOfFields; ++fieldno ) {
		if ( rec.value( fieldno ).present && !rec.value( fieldno ).null && mDirect[fieldno] != 0 ) {
			DBT data;
			field2dbt( data, description().fields[fieldno], rec.value( fieldno ) );
			int err = mDirect[fieldno]->put( mDirect[fieldno], &key_id, &data, 0 );
			if ( err )
				throw runtime_error( string( "Cannot insert into table " ) + description().name + " field " + description().fields[fieldno].name + ": " + strerror( errno ) );
			if ( mReverse[fieldno] != 0 )
				if ( ( err = mReverse[fieldno]->put( mReverse[fieldno], &data, &key_id, 0 ) ) != 0 )
					throw runtime_error( string( "Cannot insert into table " ) + description().name + " field-key " + description().fields[fieldno].name + ": " + strerror( errno ) );
		}
	}

} // DatabaseEngineDb185::Table::insert

//======================================================================

void DatabaseEngineDb185::Table::update( const DatabaseEngine::Record& aRecord )
{
	writeAccess();

	const Record& rec = dynamic_cast<const Record&>( aRecord );
	if ( rec.id() == 0 )
		throw runtime_error( string( "Cannot update table " ) + description().name + ": invalid record id" );
	check_record_requirements( rec, false );

	DBT key_id;
	id2dbt( key_id, rec.id() );
	const size_t numberOfFields = description().numberOfFields();
	for ( size_t fieldno = 0; fieldno < numberOfFields; ++fieldno ) {
		if ( rec.value( fieldno ).present && mDirect[fieldno] != 0 ) {
			DBT data_to_remove;
			if ( mReverse[fieldno] != 0 ) { // old value is needed to update reverse table
				int err = mDirect[fieldno]->get( mDirect[fieldno], &key_id, &data_to_remove, 0 );
				if ( err )
					throw runtime_error( string( "Cannot get table " ) + description().name + " field " + description().fields[fieldno].name + ": " + strerror( errno ) );
			}

			DBT data;
			field2dbt( data, description().fields[fieldno], rec.value( fieldno ) );
			if ( !rec.value( fieldno ).null ) {
				int err = mDirect[fieldno]->put( mDirect[fieldno], &key_id, &data, 0 );
				if ( err )
					throw runtime_error( string( "Cannot update table " ) + description().name + " field " + description().fields[fieldno].name + ": " + strerror( errno ) );
				if ( mReverse[fieldno] != 0 ) {
					removeOldValueFromReverseTable( fieldno, rec.id(), data_to_remove );  // remove old value first
					if ( ( err = mReverse[fieldno]->put( mReverse[fieldno], &data, &key_id, 0 ) ) != 0 )
						throw runtime_error( string( "Cannot update table " ) + description().name + " field-key " + description().fields[fieldno].name + ": " + strerror( errno ) );
				}
			}
			else { // value is null, delete existing records
				int err = mDirect[fieldno]->del( mDirect[fieldno], &key_id, 0 );
				if ( err )
					throw runtime_error( string( "Cannot del table " ) + description().name + " field " + description().fields[fieldno].name + ": " + strerror( errno ) );
				if ( mReverse[fieldno] != 0 )
					removeOldValueFromReverseTable( fieldno, rec.id(), data_to_remove );  // remove old value first
			}
		}
	}

} // DatabaseEngineDb185::Table::update

//======================================================================

void DatabaseEngineDb185::Table::removeOldValueFromReverseTable( size_t fieldno, Id id, DBT& data_to_remove )
{
	if ( mReverse[fieldno] != 0 && data_to_remove.size > 0 ) {
		DBT data_field;
		DBT data_to_remove_orig = data_to_remove;
		int err = mReverse[fieldno]->seq( mReverse[fieldno], &data_to_remove, &data_field, R_CURSOR );
		switch ( err ) {
		case 0:
			if ( equal( data_to_remove, data_to_remove_orig ) ) {
				if ( id == dbt2id( data_field ) ) {
					if ( ( err = mReverse[fieldno]->del( mReverse[fieldno], &data_to_remove, R_CURSOR ) ) != 0 ) {
						throw runtime_error( string( "Cannot del table " ) + description().name + " field " + description().fields[fieldno].name + ": " + strerror( errno ) );
					}
				}
				else {
					while ( mReverse[fieldno]->seq( mReverse[fieldno], &data_to_remove, &data_field, R_NEXT ) == 0 && equal( data_to_remove, data_to_remove_orig ) ) {
						if ( id == dbt2id( data_field ) )
							if ( ( err = mReverse[fieldno]->del( mReverse[fieldno], &data_to_remove, R_CURSOR ) ) != 0 ) {
								throw runtime_error( string( "Cannot del table " ) + description().name + " field " + description().fields[fieldno].name + ": " + strerror( errno ) );
							}
					}
				}
			}
			break;
		case RET_SPECIAL: // not found
			break;
		default:
			throw runtime_error( string( "Cannot get table " ) + description().name + " field " + description().fields[fieldno].name + ": " + strerror( errno ) );
		}
	}

} // DatabaseEngineDb185::Table::removeOldValueFromReverseTable

//======================================================================

void DatabaseEngineDb185::Table::remove( const char* aFieldName, const FieldValue& aValue )
{
	const size_t fno = description().fieldByName( aFieldName );
	if ( description().fields[fno].key != kUnique && description().fields[fno].key != kDuplicate )
		throw runtime_error( string( "Unable to remove from table " ) + description().name + " by field " + aFieldName + ": field is not key" );

	IdList ids;
	idsByField( aFieldName, aValue, ids );
	remove( ids );

} // DatabaseEngineDb185::Table::remove

//======================================================================

void DatabaseEngineDb185::Table::remove( const char* aFieldName1, const FieldValue& aValue1, const char* aFieldName2, const FieldValue& aValue2 )
{
	const size_t fno1 = description().fieldByName( aFieldName1 );
	if ( description().fields[fno1].key != kUnique && description().fields[fno1].key != kDuplicate )
		throw runtime_error( string( "Unable to remove from table " ) + description().name + " by field " + aFieldName1 + ": field is not key" );
	const size_t fno2 = description().fieldByName( aFieldName2 );
	if ( description().fields[fno2].key != kUnique && description().fields[fno2].key != kDuplicate )
		throw runtime_error( string( "Unable to remove from table " ) + description().name + " by field " + aFieldName2 + ": field is not key" );

	IdList ids;
	idsByField( aFieldName1, aValue1, aFieldName2, aValue2, ids );
	remove( ids );

} // DatabaseEngineDb185::Table::remove

//======================================================================

void DatabaseEngineDb185::Table::remove( const IdList& ids )
{
	writeAccess();

	const size_t numberOfFields = description().numberOfFields();
	for ( IdList::const_iterator id = ids.begin(); id != ids.end(); ++id ) {
		DBT key_id;
		id2dbt( key_id, *id );
		for ( size_t fieldno = 0; fieldno < numberOfFields; ++fieldno ) {
			if ( mDirect[fieldno] != 0 ) {
				int err = 0;
				if ( mReverse[fieldno] != 0 ) {
					DBT data;
					if ( mDirect[fieldno]->get( mDirect[fieldno], &key_id, &data, 0 ) == 0 )
						removeOldValueFromReverseTable( fieldno, *id, data );
				}
				if ( ( err = mDirect[fieldno]->del( mDirect[fieldno], &key_id, 0 ) ) != 0 )
					throw runtime_error( string( "Cannot del table " ) + description().name + " field " + description().fields[fieldno].name + ": " + strerror( errno ) );
			}
		}
	}

} // DatabaseEngineDb185::Table::remove

//======================================================================

void DatabaseEngineDb185::Table::remove( Id id )
{
	writeAccess();

	const size_t numberOfFields = description().numberOfFields();
	DBT key_id;
	id2dbt( key_id, id );
	for ( size_t fieldno = 0; fieldno < numberOfFields; ++fieldno ) {
		if ( mDirect[fieldno] != 0 ) {
			int err = 0;
			if ( mReverse[fieldno] != 0 ) {
				DBT data;
				if ( mDirect[fieldno]->get( mDirect[fieldno], &key_id, &data, 0 ) == 0 )
					removeOldValueFromReverseTable( fieldno, id, data );
			}
			if ( ( err = mDirect[fieldno]->del( mDirect[fieldno], &key_id, 0 ) ) != 0 )
				throw runtime_error( string( "Cannot del table " ) + description().name + " field " + description().fields[fieldno].name + ": " + strerror( errno ) );
		}
	}

} // DatabaseEngineDb185::Table::remove

//======================================================================

DatabaseEngine::Id DatabaseEngineDb185::Table::newId()
{
	writeAccess();

	Id id = 0;
	if ( mDirect[0] ) {
		Id id0 = sIdId;
		DBT key;
		id2dbt( key, id0 );
		DBT data;
		int err = mDirect[0]->get( mDirect[0], &key, &data, 0 );
		switch ( err ) {
		case 0:
			id = dbt2id( data ) + 1;
			break;
		case RET_SPECIAL:
			id = 100; // no records in the database
			break;
		default:
			throw runtime_error( string( "Internal: DatabaseEngineDb185::Table::newId: Unable to get id record for table " ) + description().name + ": " + strerror( errno ) );
		}

		DBT new_id_data = { &id, sizeof id };
		if ( mDirect[0]->put( mDirect[0], &key, &new_id_data, 0 ) != 0 )
			throw runtime_error( string( "Internal: DatabaseEngineDb185::Table::newId: Unable to put id record for table " ) + description().name + ": " + strerror( errno ) );
	}
	else
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Table::newId: Unable to generate new id for table " ) + description().name + ": no subtables available" );
	return id;

} // DatabaseEngineDb185::Table::newId

//======================================================================

DatabaseEngine::Id DatabaseEngineDb185::Table::internalId()
{
	Id id = 0;
	if ( mDirect[0] ) {
		Id id0 = sIdId;
		DBT key;
		id2dbt( key, id0 );
		DBT data;
		int err = mDirect[0]->get( mDirect[0], &key, &data, 0 );
		switch ( err ) {
		case 0:
			id = dbt2id( data );
			break;
		case RET_SPECIAL:
			id = 0; // no records in the database
			break;
		default:
			throw runtime_error( string( "Internal: DatabaseEngineDb185::Table::newId: Unable to get id record for table " ) + description().name + ": " + strerror( errno ) );
		}
	}
	else
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Table::newId: Unable to generate new id for table " ) + description().name + ": no subtables available" );
	return id;

} // DatabaseEngineDb185::Table::internalId

//======================================================================

void DatabaseEngineDb185::Table::internalId( DatabaseEngine::Id id )
{
	writeAccess();

	if ( mDirect[0] ) {
		Id id0 = sIdId;
		DBT key;
		id2dbt( key, id0 );
		DBT new_id_data = { &id, sizeof id };
		if ( mDirect[0]->put( mDirect[0], &key, &new_id_data, 0 ) != 0 )
			throw runtime_error( string( "Internal: DatabaseEngineDb185::Table::newId: Unable to put id record for table " ) + description().name + ": " + strerror( errno ) );
	}
	else
		throw runtime_error( string( "Internal: DatabaseEngineDb185::Table::newId: Unable to generate new id for table " ) + description().name + ": no subtables available" );

} // DatabaseEngineDb185::Table::internalId

//======================================================================

void DatabaseEngineDb185::Table::sync()
{
	const size_t numberOfFields = description().numberOfFields();
	for ( size_t fieldno = 0; fieldno < numberOfFields; ++fieldno ) {
		if ( mDirect[fieldno] != 0 )
			mDirect[fieldno]->sync( mDirect[fieldno], 0 );
		if ( mReverse[fieldno] != 0 )
			mReverse[fieldno]->sync( mReverse[fieldno], 0 );
	}

} // DatabaseEngineDb185::Table::sync

//======================================================================

void DatabaseEngineDb185::Table::writeAccess()
{
	++mWriteAccess;
	if ( ( mWriteAccess % 10 ) == 0 )
		sync();

} // DatabaseEngineDb185::Table::writeAccess

//======================================================================

int DatabaseEngineDb185::id_compare( const DBT* a, const DBT* b )
{
	if ( a->data == 0 )
		return -1;
	if ( b->data == 0 )
		return 1;

	const Id ai = *(const Id*)a->data;
	const Id bi = *(const Id*)b->data;

	if ( ai < bi )
		return -1;
	else if ( ai > bi )
		return 1;
	else
		return 0;

} // DatabaseEngineDb185::id_compare

//======================================================================

void DatabaseEngineDb185::id2dbt( DBT& dbt, const Id& id )
{
	// Id must be passed by reference, not by value!!!
	dbt.data = (void*)&id;
	dbt.size = sizeof id;

} // DatabaseEngineDb185::id2dbt

//======================================================================

DatabaseEngine::Id DatabaseEngineDb185::dbt2id( const DBT& dbt )
{
	if ( dbt.size != sizeof( Id ) )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::dbt2id: size of value stored in Dbt does not match Id data type" ) );
	return(*(const Id*)dbt.data);

} // DatabaseEngineDb185::dbt2id

//======================================================================

void DatabaseEngineDb185::field2dbt( DBT& dbt, const FieldDescription& aFieldDescription, const FieldValue& aValue )
{
	if ( !aValue.present )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::field2dbt: value is not present" ) );
	if ( aValue.null )
		throw runtime_error( string( "Internal: DatabaseEngineDb185::field2dbt: value is null" ) );

	switch ( aFieldDescription.type ) {
	case tId:
		id2dbt( dbt, aValue.value.id );
		break;
	case tString:
		dbt.data = (void*)aValue.str.c_str();
		dbt.size = aValue.str.length();
		break;
	case tInteger:
		dbt.data = (void*)&aValue.value.integer;
		dbt.size = sizeof aValue.value.integer;
		break;
	case tLongLong:
		dbt.data = (void*)&aValue.value.longlong;
		dbt.size=  sizeof aValue.value.longlong;
		break;
	case tTime:
		dbt.data = (void*)&aValue.value.time;
		dbt.size = sizeof aValue.value.time;
		break;
	}

} // DatabaseEngineDb185::field2dbt

//======================================================================

void DatabaseEngineDb185::dbt2field( const DBT& dbt, const FieldDescription& aFieldDescription, FieldValue& aValue )
{
	switch ( aFieldDescription.type ) {
	case tId:
		aValue.value.id = dbt2id( dbt );
		break;
	case tString:
		aValue.str.assign( reinterpret_cast<char*>( dbt.data ), dbt.size );
		break;
	case tInteger:
		if ( dbt.size != sizeof( aValue.value.integer ) )
			throw runtime_error( string( "Internal: DatabaseEngineDb185::dbt2field: size of value stored in Dbt does not match data type" ) );
		aValue.value.integer = *(const int*)dbt.data;
		break;
	case tLongLong:
		if ( dbt.size != sizeof( aValue.value.longlong ) )
			throw runtime_error( string( "Internal: DatabaseEngineDb185::dbt2field: size of value stored in Dbt does not match data type" ) );
		aValue.value.longlong = *(const long long*)dbt.data;
		break;
	case tTime:
		if ( dbt.size != sizeof( aValue.value.time ) )
			throw runtime_error( string( "Internal: DatabaseEngineDb185::dbt2field: size of value stored in Dbt does not match data type" ) );
		aValue.value.time = *(const time_t*)dbt.data;
		break;
	}

	aValue.present = true;
	aValue.null = false;

} // DatabaseEngineDb185::dbt2field

//======================================================================

bool DatabaseEngineDb185::equal( const DBT& a, const DBT& b )
{
	return(a.size == b.size && memcmp( a.data, b.data, a.size ) == 0);

} // DatabaseEngineDb185::equal

//======================================================================
