/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/database/DatabaseEngineDb185.hh,v 1.3 2001/06/20 10:08:11 alex Exp $
 */

// -*- C++ -*-  Tue Aug 11 1998
// $Id: DatabaseEngineDb185.hh,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef DatabaseEngineDb185_hh
#define DatabaseEngineDb185_hh

#ifndef DatabaseEngine_hh
#include "DatabaseEngine.hh"
#endif

#include <db.h>

//======================================================================

class DatabaseEngineDb185 : public DatabaseEngine
{
  private:
     class Record;

     class Table : public DatabaseEngine::Table
     {
       public:
          Table( const TableDescription& aTableDescription, DatabaseEngineDb185& aEngine );
          virtual ~Table();

          virtual Pointer<DatabaseEngine::Record> newRecord() const;

            // Selects all the fields
          virtual Pointer<DatabaseEngine::RecordList> select( const char* aFieldName, const FieldValue& aValue );
          virtual Pointer<DatabaseEngine::RecordList> select( const char* aFieldName1, const FieldValue& aValue1, const char* aFieldName2, const FieldValue& aValue2 );
          virtual Pointer<DatabaseEngine::RecordList> selectAll();
          virtual Pointer<DatabaseEngine::Record> select( Id id );
            // Selects specified fields
          virtual Pointer<DatabaseEngine::RecordList> select( const char* aFieldName, const FieldValue& aValue, const vector<string>& aFieldsToSelect );
          virtual Pointer<DatabaseEngine::RecordList> select( const char* aFieldName, const FieldValue& aValue, const string& aFieldToSelect );
          virtual Pointer<DatabaseEngine::Record> select( Id id, const vector<string>& aFieldsToSelect );
          virtual Pointer<DatabaseEngine::Record> select( Id id, const string& aFieldToSelect );
            // Selects all ids from the table
          virtual Pointer<IdList> select();
          virtual size_t number_of_records();

            // Returns if record with given id is present.
          virtual bool present( Id id );

            // Inserts new record, makes sure unique fields will have
            // no duplicates, not null fields receive value
          virtual void insert( const DatabaseEngine::Record& aRecord, bool use_id_of_record = false );

            // Updates record
          virtual void update( const DatabaseEngine::Record& aRecord );

            // Removes all the records matching
          virtual void remove( const char* aFieldName, const FieldValue& aValue );
          virtual void remove( const char* aFieldName1, const FieldValue& aValue1, const char* aFieldName2, const FieldValue& aValue2 );
          virtual void remove( const IdList& ids );
          virtual void remove( Id id );

          virtual void sync();

          virtual DatabaseEngine::Id internalId();
          virtual void internalId( DatabaseEngine::Id id );

       private:
          DB** mDirect;
          DB** mReverse;
          size_t mWriteAccess;

            // Select records ids by aFieldName
          virtual void idsByField( const char* aFieldName, const FieldValue& aValue, IdList& ids );
          virtual void idsByField( const char* aFieldName1, const FieldValue& aValue1, const char* aFieldName2, const FieldValue& aValue2, IdList& ids );

            // Generates unique id for a new record
          virtual Id newId();

          virtual void removeOldValueFromReverseTable( size_t fieldno, Id id, DBT& data_to_remove );

            // Increments number of write accesses and probably calls sync()
          virtual void writeAccess();

          static const Id sIdId;

            // Make sure new record will not violate requirements
          virtual void check_record_requirements( const Record& aRecord, bool check_for_presence_of_fields_values ) const;

     }; // class Table

     friend class Table;

       // --------------------------------------------------

     class Record : public DatabaseEngine::Record
     {
       public:
          Record( const TableDescription& aTableDescription );
          virtual ~Record();

          virtual Id id() const;
          virtual void id( Id aId ) { mId = aId; }

            // Returns if value of field is null
          virtual bool null( const char* aFieldName ) const;

            // Check if field of the required type and returns value
            // raises exception if value is null
          virtual Id id( const char* aFieldName ) const;
          virtual const string& str( const char* aFieldName ) const;
          virtual int integer( const char* aFieldName ) const;
          virtual long long longlong( const char* aFieldName ) const;
          virtual time_t time( const char* aFieldName ) const;

            // Sets the value of the field to null if it is allowed
          virtual void setNull( const char* aFieldName );

            // Marks field as not present, i.e. next update will not
            // touch this field
          virtual void not_present( const char* aFieldName );

            // Check if field of the required type and sets value
          virtual void id( const char* aFieldName, Id aId );
          virtual void str( const char* aFieldName, const string& s );
          virtual void integer( const char* aFieldName, int i );
          virtual void longlong( const char* aFieldName, long long ll );
          virtual void time( const char* aFieldName, time_t t );

       private:
          mutable Id mId;
          FieldValue* mValues;

          friend class Table;

          virtual void id( Id aId ) const { mId = aId; }

            // Sets value of field from dbt
          virtual void set( const DBT& dbt, const char* aFieldName );
            // Sets dbt from value of field
          virtual void get( DBT& dbt, const char* aFieldName ) const;

          const FieldValue& value( size_t fieldno ) const { return mValues[fieldno]; }

     }; // class Record

     friend class Record;

       // --------------------------------------------------

  public:
     DatabaseEngineDb185( const string& aBaseDir, bool aCreate );
     virtual ~DatabaseEngineDb185();

     virtual void create( const TableDescription& aTableDescription );
     virtual bool exists( const TableDescription& aTableDescription );
     virtual DatabaseEngine::Table* table( const TableDescription& aTableDescription );

  protected:

  private:
     BTREEINFO mInfoId, mInfoDup, mInfoDupId;

     virtual string fieldFilename( const string& aTableName, const string& aFieldName, bool reverse );

     static int id_compare( const DBT* a, const DBT* b );
     static void id2dbt( DBT& dbt, const Id& id ); // Id must be passed by reference, not by value!!!
     static Id dbt2id( const DBT& dbt );
     static void field2dbt( DBT& dbt, const FieldDescription& aFieldDescription, const FieldValue& aValue );
     static void dbt2field( const DBT& dbt, const FieldDescription& aFieldDescription, FieldValue& aValue );

     static bool equal( const DBT& a, const DBT& b );

}; // class DatabaseEngineDb185

//======================================================================

#endif // DatabaseEngineDb185_hh
