/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/disk/Disk.cc,v 1.9 2001/06/20 10:07:00 alex Exp $
 */

// -*- C++ -*-   Tue Feb  9 1999
// Eugene Skepner <eu@qub.com>
// $Id: Disk.cc,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef Disk_hh
#include "Disk.hh"
#endif

#include <sys/disklabel.h>

#include <cerrno>
#include <stdexcept>

//======================================================================

Disk::Disk( const string& aDeviceName )
	: mDisk( ::Open_Disk( aDeviceName.c_str() ) ), mModified( false )
{
	if ( mDisk == 0 )
		throw runtime_error( string( "Unable to Open_Disk " ) + aDeviceName + ": " + ::strerror( errno ) );

} // Disk::Disk

//======================================================================

Disk::Disk( disk* aDisk )
	: mDisk( aDisk )
{
} // Disk::Disk

//======================================================================

Disk::~Disk()
{
	if ( mDisk )
		::Free_Disk( mDisk );

} // Disk::~Disk

//======================================================================

Pointer<Disk> Disk::clone()
{
	return new Disk( ::Clone_Disk( mDisk ) );

} // Disk::clone

//======================================================================

void Disk::debug()
{
	::Debug_Disk( mDisk );

} // Disk::debug

//======================================================================

string Disk::name() const
{
	return(mDisk->name);

} // Disk::name

//======================================================================

struct DiskFlagName
{
	const char* name;
	u_long flag;
};

static DiskFlagName diskFlagsNames[] = {
	{ "DISK_ON_TRACK", DISK_ON_TRACK },
	{ "CHUNK_BSD_COMPAT", CHUNK_BSD_COMPAT },
	{ "CHUNK_ALIGN", CHUNK_ALIGN },
	{ "CHUNK_IS_ROOT", CHUNK_IS_ROOT },
	{ "CHUNK_ACTIVE", CHUNK_ACTIVE },
	{ "CHUNK_FORCE_ALL", CHUNK_FORCE_ALL },
	{ 0, 0 }
};

u_long Disk::text2flags( const list<string>& aFlagsNames )
{
	u_long result = 0;
	for ( list<string>::const_iterator fl = aFlagsNames.begin(); fl != aFlagsNames.end(); ++fl ) {
		for ( DiskFlagName* p = diskFlagsNames; p->name; ++p )
			if ( *fl == p->name ) {
				result |= p->flag;
				break;
			}
	}
	return(result);

} // Disk::text2flags

//======================================================================

Pointer<Container<list<string> > > Disk::flags2text( u_long aFlags )
{
	Pointer<Container<list<string> > > result = new Container<list<string> >();
	for ( DiskFlagName* p = diskFlagsNames; p->name; ++p )
		if ( aFlags & p->flag )
			result->push_back( p->name );
	return(result);

} // Disk::flags2text

//======================================================================

u_long Disk::flags() const
{
	return(mDisk->flags);

} // Disk::flags

//======================================================================

u_long Disk::bios_cyl() const
{
	return(mDisk->bios_cyl);

} // Disk::bios_cyl

//======================================================================

u_long Disk::bios_hd() const
{
	return(mDisk->bios_hd);

} // Disk::bios_hd

//======================================================================

u_long Disk::bios_sect() const
{
	return(mDisk->bios_sect);

} // Disk::bios_sectbios_hd

//======================================================================

Pointer<DiskChunk> Disk::chunks()
{
	return new DiskChunk( mDisk->chunks );

} // Disk::chunks

//======================================================================

void Disk::Set_Bios_Geom( u_long cyl, u_long heads, u_long sects )
{
	mModified = true;
	::Set_Bios_Geom( mDisk, cyl, heads, sects );

} // Disk::Set_Bios_Geom

//======================================================================

void Disk::Sanitize_Bios_Geom()
{
	mModified = true;
	::Sanitize_Bios_Geom( mDisk );

} // Disk::Sanitize_Bios_Geom

//======================================================================

int Disk::Delete_Chunk( Pointer<DiskChunk> aChunk )
{
	mModified = true;
	DiskChunk::PartInfo::free( aChunk->mChunk->private_data );
	aChunk->mChunk->private_data = 0;
	int result = ::Delete_Chunk( mDisk, aChunk->mChunk );
	return(result);

} // Disk::Delete_Chunk

//======================================================================

int Disk::Create_Chunk( u_long offset, u_long size, int subtype, u_long flags )
{
	mModified = true;
	return ::Create_Chunk( mDisk, offset, size, type_by_subtype( subtype ), subtype, flags );

} // Disk::Create_Chunk

//======================================================================

Pointer<DiskChunk> Disk::Create_Chunk_DWIM( Pointer<DiskChunk> parent, u_long size, chunk_e type, int subtype, u_long flags )
{
	chunk* result = ::Create_Chunk_DWIM( mDisk, parent->mChunk, size, type, subtype, flags );
	if ( result )
		return new DiskChunk( result );
	else
		return Pointer<DiskChunk>();

} // Disk::Create_Chunk_DWIM

//======================================================================

Pointer<DiskChunk> Disk::Create_Chunk_DWIM( Pointer<DiskChunk> parent, u_long size, const string& type, const list<string>& flags )
{
	return Create_Chunk_DWIM( parent, size, part, type == "fs" ? FS_BSDFFS : FS_SWAP, text2flags( flags ) );

} // Disk::Create_Chunk_DWIM

//======================================================================

//  void Disk::Collapse_Disk()
//  {
//	  ::Collapse_Disk( mDisk );

//  } // Disk::Collapse_Disk

//  //======================================================================

//  int Disk::Collapse_Chunk( struct chunk* aChunk )
//  {
//	  return ::Collapse_Chunk( mDisk, aChunk );

//  } // Disk::Collapse_Chunk

//======================================================================

void Disk::All_FreeBSD( bool force_all )
{
	mModified = true;
	::All_FreeBSD( mDisk, force_all );

} // Disk::All_FreeBSD

//======================================================================

void Disk::undo()
{
	string nam = name();
	if ( mDisk )
		::Free_Disk( mDisk );
	mDisk = ::Open_Disk( nam.c_str() );
	if ( !mDisk )
		throw runtime_error( string( "Can't reopen disk " ) + nam + "! Internal state is probably corrupted" );
	mModified = false;

} // Disk::undo

//======================================================================

string Disk::CheckRules()
{
	return ::CheckRules( mDisk );

} // Disk::CheckRules

//======================================================================

Pointer<Container<list<string> > > Disk::Disk_Names()
{
	char **names = ::Disk_Names();
	Pointer<Container<list<string> > > result = new Container<list<string> >;
	for (char **n = names; *n != NULL; n++) {
	    result->push_back(*n);
	}
	::free(names);
	return(result);

} // Disk::Disk_Names

//======================================================================

void Disk::Set_Boot_Mgr( const u_char *bootmgr )
{
	mModified = true;
	::Set_Boot_Mgr( mDisk, bootmgr, DOSPARTOFF); /* XXX not sure if DOSPARTOFF is correct */

} // Disk::Set_Boot_Mgr

//======================================================================

void Disk::Set_Boot_Blocks( const u_char *_boot1, const u_char *_boot2 )
{
	mModified = true;
	::Set_Boot_Blocks( mDisk, _boot1, _boot2 );

} // Disk::Set_Boot_Blocks

//======================================================================

int Disk::Write_Disk()
{
	int result = ::Write_Disk( mDisk );
	mModified = result;
	return(result);

} // Disk::Write_Disk

//======================================================================

int Disk::Cyl_Aligned( u_long offset )
{
	return ::Cyl_Aligned( mDisk, offset );

} // Disk::Cyl_Aligned

//======================================================================

u_long Disk::Next_Cyl_Aligned( u_long offset )
{
	return ::Next_Cyl_Aligned( mDisk, offset );

} // Disk::Next_Cyl_Aligned

//======================================================================

u_long Disk::Prev_Cyl_Aligned( u_long offset )
{
	return ::Prev_Cyl_Aligned( mDisk, offset );

} // Disk::Prev_Cyl_Aligned

//======================================================================

int Disk::Track_Aligned( u_long offset )
{
	return ::Track_Aligned( mDisk, offset );

} // Disk::Track_Aligned

//======================================================================

u_long Disk::Next_Track_Aligned( u_long offset )
{
	return ::Next_Track_Aligned( mDisk, offset );

} // Disk::Next_Track_Aligned

//======================================================================

u_long Disk::Prev_Track_Aligned( u_long offset )
{
	return ::Prev_Track_Aligned( mDisk, offset );

} // Disk::Prev_Track_Aligned

//======================================================================

int Disk::MakeDev( struct chunk* c, const string& path )
{
	return ::MakeDev( c, path.c_str() );

} // Disk::MakeDev

//======================================================================

int Disk::MakeDevDisk( const string& path )
{
	return ::MakeDevDisk( mDisk, path.c_str() );

} // Disk::MakeDevDisk

//======================================================================

string Disk::ShowChunkFlags( struct chunk* c )
{
	return ::ShowChunkFlags( c );

} // Disk::ShowChunkFlags

//======================================================================

string Disk::ChunkCanBeRoot( struct chunk* c )
{
	char* can = ::ChunkCanBeRoot( c );
	if ( can )
		return(can);
	else
		return string();

} // Disk::ChunkCanBeRoot

//======================================================================

struct SliceTypeName
{
	int type;
	const char* name;
};

static SliceTypeName sliceTypeNames[] = {
	{ 1, "fat (12-bit)" },
	{ 2, "XENIX /" },
	{ 3, "XENIX /usr" },
	{ 4, "fat (16-bit,<=32Mb)" },
	{ 5, "extended DOS" },
	{ 6, "fat (16-bit,>32Mb)" },
	{ 7, "NTFS/HPFS/QNX" },
	{ 8, "AIX bootable" },
	{ 9, "AIX data" },
	{ 10, "OS/2 bootmgr" },
	{ 11, "fat (32-bit)" },
	{ 12, "fat (32-bit,LBA)" },
	{ 14, "fat (16-bit,>32Mb,LBA)" },
	{ 15, "extended DOS, LBA" },
	{ 18, "Compaq Diagnostic" },
	{ 84, "OnTrack diskmgr" },
	{ 100, "Netware 2.x" },
	{ 101, "Netware 3.x" },
	{ 115, "SCO UnixWare" },
	{ 128, "Minix 1.1" },
	{ 129, "Minix 1.5" },
	{ 130, "linux_swap" },
	{ 131, "ext2fs" },
	{ 165, "FreeBSD" },
	{ 166, "OpenBSD FFS" },	/* 0xA6 */
	{ 169, "NetBSD FFS" },	/* 0xA9 */
	{ 182, "OpenBSD" },		/* dedicated */
	{ 183, "bsd/os" },
	{ 184, "bsd/os swap" },
	{ 0, 0 }
};

Pointer<Container<list<string > > > Disk::standard_slice_types()
{
	Pointer<Container<list<string > > > result = new Container<list<string > >;
	for ( SliceTypeName* t = sliceTypeNames; t->name; ++t )
		result->push_back( t->name );
	return(result);

} // Disk::standard_slice_types

//======================================================================

int Disk::standard_slice_type_id( const string& aName )
{
	for ( SliceTypeName* t = sliceTypeNames; t->name; ++t )
		if ( aName == t->name )
			return(t->type);
	return 0;

} // Disk::standard_slice_type_id

//======================================================================

string Disk::slice_type_name( int type, int subtype )
{
	switch (type) {
	case 0:
		return("whole");
	case 1:
		for ( SliceTypeName* t = sliceTypeNames; t->name; ++t )
			if ( subtype == t->type )
				return(t->name);
		return("unknown");
	case 2:
		return("fat");
	case 3:
		switch ( subtype ) {
		case 165:
			for ( SliceTypeName* t = sliceTypeNames; t->name; ++t )
				if ( subtype == t->type )
					return(t->name);
		default:
			return("unknown");
		}
	case 4:
		return("extended");
	case 5:
		return("part");
	case 6:
		return("unused");
	default:
		return("unknown");
	}

} // Disk::slice_type_name

//======================================================================

chunk_e Disk::type_by_subtype( int subtype )
{
	switch ( subtype ) {
	case 165:
		return(freebsd);
	case 6:
		return(fat);
	default:
		return(unknown);
	}

} // Disk::type_by_subtype

//======================================================================

string DiskChunk::mountpoint() const
{
	if ( type() == ::fat || is_fs() ) {
		if ( mChunk->private_data ) {
			return(static_cast<PartInfo*>( mChunk->private_data )->mountpoint);
		}
		else {
			return("<none>");
		}
	}
	else if ( is_swap() )
		return("swap");
	else
		return("<none>");

} // DiskChunk::mountpoint

//======================================================================

bool DiskChunk::is_fs() const
{
	return(type() == ::part && subtype() != FS_SWAP);

} // DiskChunk::is_fs

//======================================================================

bool DiskChunk::is_swap() const
{
	return(type() == ::part && subtype() == FS_SWAP);

} // DiskChunk::is_swap

//======================================================================

void DiskChunk::mountpoint( const string& aMountpoint ) const
{
	if ( type() == ::fat || is_fs() ) {
		if ( mChunk->private_data ) {
			static_cast<PartInfo*>( mChunk->private_data )->mountpoint = aMountpoint;
		}
		else {
			mChunk->private_data = static_cast<void*>( new PartInfo( aMountpoint, 1 ) );
			mChunk->private_free = &PartInfo::free;
			mChunk->private_clone = &PartInfo::clone;
		}
	}

} // DiskChunk::mountpoint

//======================================================================

int DiskChunk::newfs() const
{
	if ( is_fs() && mChunk->private_data )
		return(static_cast<PartInfo*>( mChunk->private_data )->newfs);
	else
		return 2;

} // DiskChunk::newfs

//======================================================================

void DiskChunk::newfs( int aNewfs ) const
{
	if ( is_fs() ) {
		if ( mChunk->private_data ) {
			static_cast<PartInfo*>( mChunk->private_data )->newfs = aNewfs;
		}
		else {
			mChunk->private_data = static_cast<void*>( new PartInfo( "<none>", aNewfs ) );
			mChunk->private_free = &PartInfo::free;
			mChunk->private_clone = &PartInfo::clone;
		}
	}

} // DiskChunk::newfs

//======================================================================

string DiskChunk::newfs_cmd() const
{
	if ( ( type() == ::fat || is_fs() ) && mChunk->private_data ) {
		return(static_cast<PartInfo*>( mChunk->private_data )->newfs_cmd);
	}
	else {
		return string();
	}

} // DiskChunk::newfs_cmd

//======================================================================

void DiskChunk::newfs_cmd( const string& aCmd ) const
{
	if ( type() == ::fat || is_fs() ) {
		if ( !mChunk->private_data ) {
			mChunk->private_data = static_cast<void*>( new PartInfo( "/change_me", 1 ) );
			mChunk->private_free = &PartInfo::free;
			mChunk->private_clone = &PartInfo::clone;
		}
		static_cast<PartInfo*>( mChunk->private_data )->newfs_cmd = aCmd;
	}

} // DiskChunk::newfs_cmd

//======================================================================

const char* chunkTypeText[] = {
	"whole",
	"unknown",
	"fat",
	"freebsd",
	"extended",
	"part",
	"unused",
	NULL
};

string DiskChunk::type_text() const
{
	return(chunkTypeText[ type() ]);

} // DiskChunk::type_text

//======================================================================

DiskChunk::PartInfo::PartInfo( const string& aMountpoint, bool aNewfs )
	: mountpoint( aMountpoint ), newfs( aNewfs ), newfs_cmd( "newfs -b 8192 -f 1024" )
{
	if ( mountpoint.empty() )
		mountpoint = "/change_me";

} // DiskChunk::PartInfo::PartInfo

//======================================================================
