/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/disk/Disk.hh,v 1.4 2001/06/20 10:07:00 alex Exp $
 */

// -*- C++ -*-  Tue Feb  9 1999
// Eugene Skepner <eu@qub.com>
// $Id: Disk.hh,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef Disk_hh
#define Disk_hh

#ifndef LanguageInterface_hh
#include "LanguageInterface.hh"
#endif

#ifndef Container_hh
#include "Container.hh"
#endif

#include <sys/types.h>

extern "C" {
#include <libdisk.h>
}

class DiskChunk;

//======================================================================

class Disk : virtual public LanguageInterface::Object
{
     CLASS_METHODS_DESCRIPTIONS_DECLARATION;

  public:
     Disk( const string& aDeviceName );
     virtual ~Disk();

     static Pointer<Disk> open( const string& aDeviceName ) { return new Disk( aDeviceName ); }

     static Pointer<Container<list<string> > > Disk_Names();

     static u_long text2flags( const list<string>& aFlagsNames );
     static Pointer<Container<list<string> > > flags2text( u_long aFlags );

     virtual Pointer<Disk> clone();
     virtual void debug();

     virtual string name() const;
     virtual u_long flags() const;
     Pointer<Container<list<string> > > flags_names() const { return flags2text( flags() ); }
     virtual u_long bios_cyl() const;
     virtual u_long bios_hd() const;
     virtual u_long bios_sect() const;

     virtual bool modified() const { return mModified; }

       // --------------------------------------------------

     virtual Pointer<DiskChunk> chunks();
     virtual int Delete_Chunk( Pointer<DiskChunk> aChunk );
     virtual int Create_Chunk( u_long offset, u_long size, int subtype, u_long flags );
     int Create_Chunk( u_long offset, u_long size, int subtype, const list<string>& flags )
          { return Create_Chunk( offset, size, subtype, text2flags( flags ) ); }
     virtual Pointer<DiskChunk> Create_Chunk_DWIM( Pointer<DiskChunk> parent, u_long size, chunk_e type, int subtype, u_long flags );
       // type: "fs" or "swap"
     Pointer<DiskChunk> Create_Chunk_DWIM( Pointer<DiskChunk> parent, u_long size, const string& type, const list<string>& flags );

       // --------------------------------------------------

     virtual void All_FreeBSD( bool force_all );
     virtual void Set_Bios_Geom( u_long cyl, u_long heads, u_long sects );
     virtual void undo();
     virtual int Write_Disk();

       // --------------------------------------------------

     static Pointer<Container<list<string > > > standard_slice_types();
     static int standard_slice_type_id( const string& aName );
     static string slice_type_name( int type, int subtype );
     static string slice_type_name( int subtype ) { return slice_type_name( 1, subtype ); }

       // --------------------------------------------------

     virtual void Sanitize_Bios_Geom();
       //virtual void Collapse_Disk();
       //virtual int Collapse_Chunk( struct chunk* aChunk );
     virtual string CheckRules();
     virtual void Set_Boot_Mgr( const u_char* bootmgr );
     virtual void Set_Boot_Blocks( const u_char *_boot1, const u_char *_boot2 );
     virtual int Cyl_Aligned( u_long offset );
     virtual u_long Next_Cyl_Aligned( u_long offset );
     virtual u_long Prev_Cyl_Aligned( u_long offset );
     virtual int Track_Aligned( u_long offset );
     virtual u_long Next_Track_Aligned( u_long offset );
     virtual u_long Prev_Track_Aligned( u_long offset );
     virtual int MakeDevDisk( const string& path );

     static int MakeDev( struct chunk* c, const string& path );
     static string ShowChunkFlags( struct chunk* c );
     static string ChunkCanBeRoot( struct chunk* c );

  private:
     disk* mDisk;
     bool mModified;

     Disk( disk* aDisk );
     static chunk_e type_by_subtype( int subtype );

}; // class Disk

//======================================================================

class DiskChunk : virtual public LanguageInterface::Object
{
     CLASS_METHODS_DESCRIPTIONS_DECLARATION;

  public:
     virtual ~DiskChunk() {}

     virtual long offset() const { return mChunk->offset; }
     virtual u_long size() const { return mChunk->size; }
     virtual u_long end() const { return mChunk->end; }
     virtual string name() const { return mChunk->name; }
     virtual int type() const { return mChunk->type; }
     virtual string type_text() const;
     virtual int subtype() const { return mChunk->subtype; }
     virtual string slice_type_name() const { return Disk::slice_type_name( mChunk->type, mChunk->subtype ); }

     virtual u_long flags() const { return mChunk->flags; }
     virtual void flags( u_long aFlags ) const { mChunk->flags = aFlags; }

     void flags_names( const list<string>& aFlagsNames ) { flags( Disk::text2flags( aFlagsNames ) ); }
     Pointer<Container<list<string> > > flags_names() const { return Disk::flags2text( flags() ); }

     virtual string mountpoint() const;
     virtual void mountpoint( const string& aMountpoint ) const;
     virtual int newfs() const;
     virtual void newfs( int aNewfs ) const;
     virtual string newfs_cmd() const;
     virtual void newfs_cmd( const string& aCmd ) const;

     virtual bool is_fs() const;
     virtual bool is_swap() const;

     virtual Pointer<DiskChunk> part() { return mChunk->part ? new DiskChunk( mChunk->part ) : 0; }
     virtual Pointer<DiskChunk> next() { return mChunk->next ? new DiskChunk( mChunk->next ) : 0; }

     virtual void set_type( int aType ) const
          {
               mChunk->subtype = aType;
               switch ( aType ) {
                 case 165:
                      mChunk->type = freebsd;
                      break;
                 case 6:
                      mChunk->type = fat;
                      break;
                 default:
                      mChunk->type = unknown;
                      break;
               }
          }

  private:
     friend class Disk;
     DiskChunk( chunk* aChunk ) : mChunk( aChunk ) {}

     chunk* mChunk;

     struct PartInfo
     {
          PartInfo( const string& aMountpoint, bool aNewfs );
          static void* clone( void* p ) { return new PartInfo( *static_cast<PartInfo*>( p ) ); }
          static void free( void* p ) { if (p != NULL) delete static_cast<PartInfo*>( p ); }

          string mountpoint;
          int newfs; // 0-no, 1-yes, 2-unknown
          string newfs_cmd;
     };

}; // class DiskChunk

//======================================================================

#endif // Disk_hh
