/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/file/Archive.cc,v 1.4 2001/06/20 10:10:29 alex Exp $
 */

// -*- C++ -*-   Wed Dec 23 1998
// Eugene Skepner <eu@qub.com>
// $Id: Archive.cc,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef Archive_hh
#include "Archive.hh"
#endif

#include <sys/mman.h>
#include <sys/stat.h>

#include "ziplib.icc"

//======================================================================

Archive::Archive()
{
} // Archive::Archive

//======================================================================

Archive::Archive( const string& aUrl, bool aCreate )
	: mStreamFile( aUrl )
{
	if ( aCreate ) {
		if ( mStreamFile.isLocal() )
			mZip.create( mStreamFile.open_local( false ).c_str() );
		else
			throw runtime_error( string( "Unable to create remote archive: " ) + mStreamFile.url() );
	}
	else {
		mZip.open( mStreamFile.open_local( true ).c_str() );
	}

} // Archive::Archive

//======================================================================

Archive::~Archive()
{
	mStreamFile.close();

} // Archive::~Archive

//======================================================================

Pointer<Container<list<string> > > Archive::directory()
{
	const size_t files = numberOfFiles();
	if ( files == 0 )
		return(new Container<list<string> >);

	Pointer<Container<list<string> > > l = new Container<list<string> >;
	for ( size_t i = 0; i < files; ++i )
		l->push_back( mZip.entries()[i].name() );

	return l;

} // Archive::directory

//======================================================================

size_t Archive::numberOfFiles()
{
	return mZip.numberOfEntries();

} // Archive::numberOfFiles

//======================================================================

void Archive::add( const string& pathname, const string& pathname_to_record )
{
	try {
		int fd = ::open( pathname.c_str(), O_RDONLY );
		if ( fd >= 0 ) {
			struct stat stats;
			if ( fstat( fd, &stats ) )
				throw runtime_error( string( "Unable to fstat " ) + pathname + ": " + strerror( errno ) );
			void* data = mmap( 0, stats.st_size, PROT_READ, MAP_FILE | MAP_PRIVATE, fd, 0 );
			if ( data == 0 )
				throw runtime_error( string( "Unable to mmap " ) + pathname + ": " + strerror( errno ) );
			try {
				mZip.add( pathname_to_record.c_str(), (char*)data, stats.st_size, stats.st_mode, stats.st_mtime );
			}
			catch ( ... ) {
				munmap( (char*)data, stats.st_size );
				close( fd );
				throw;
			}
			munmap( (char*)data, stats.st_size );
			close( fd );
		}
		else {
			throw runtime_error( string( "Unable to open " ) + pathname + ": " + strerror( errno ) );
		}
	}
	catch ( runtime_error& err ) {
		throw runtime_error( string( "Cannot add " ) + pathname_to_record + " to stream: " + err.what() );
	}

} // Archive::add

//======================================================================

unsigned long Archive::size( const string& file )
{
	try {
		return mZip.entries()[ mZip.find_entry( file.c_str() ) ].uncompressedSize();
	}
	catch ( Zip::error& err ) {
		throw runtime_error( string( "Cannot find size of " ) + file + ": " + err.what() );
	}

} // Archive::size

//======================================================================

void Archive::extract( const string& file_to_extract, const string& pathname_to_store )
{
	try {
		size_t len = mZip.length( file_to_extract.c_str() );
		int fd = ::open( pathname_to_store.c_str(), O_RDWR|O_CREAT|O_TRUNC, 0600 );
		if ( fd < 0 )
			throw runtime_error( string( "Unable to create " ) + pathname_to_store + ": " + strerror( errno ) );
		if ( ftruncate( fd, len ) ) {
			close( fd );
			throw runtime_error( string( "Unable to change size of " ) + pathname_to_store + ": " + strerror( errno ) );
		}
		void* data = mmap( 0, len, PROT_READ|PROT_WRITE, MAP_FILE | MAP_SHARED, fd, 0 );
		if ( data == 0 ) {
			close( fd );
			throw runtime_error( string( "Unable to mmap " ) + pathname_to_store + ": " + strerror( errno ) );
		}
		try {
			mZip.extract( file_to_extract.c_str(), (char*)data, len );
		}
		catch ( ... ) {
			msync( (char*)data, len, 0 );
			munmap( (char*)data, len );
			close( fd );
			throw;
		}
		msync( (char*)data, len, 0 );
		munmap( (char*)data, len );
		close( fd );
	}
	catch ( runtime_error& err ) {
		throw runtime_error( string( "Cannot extract " ) + file_to_extract + ": " + err.what() );
	}

} // Archive::extract

//======================================================================

void Archive::extract_append( const string& file_to_extract, const string& pathname_to_append, bool create )
{
	try {
		size_t len = 0;
		char* data = mZip.extract( file_to_extract.c_str(), 0, len );
		const int flags = O_RDWR | O_APPEND | ( create ? O_CREAT : 0 );
		int fd = ::open( pathname_to_append.c_str(), flags, 0600 );
		if ( fd < 0 ) {
			delete data;
			throw runtime_error( string( "Unable to open " ) + pathname_to_append + ": " + strerror( errno ) );
		}
		if ( write( fd, data, len ) != int( len ) ) {
			delete data;
			fsync( fd );
			close( fd );
			throw runtime_error( string( "Unable to write " ) + pathname_to_append + ": " + strerror( errno ) );
		}
		fsync( fd );
		close( fd );
		delete data;
	}
	catch ( runtime_error& err ) {
		throw runtime_error( string( "Cannot extract " ) + file_to_extract + ": " + err.what() );
	}

} // Archive::extract_append

//======================================================================

void* Archive::extract( const string& file_to_extract, unsigned long& length )
{
	try {
		return mZip.extract( file_to_extract.c_str(), 0, (unsigned int &)length );
	}
	catch ( Zip::error& err ) {
		throw runtime_error( string( "Cannot extract " ) + file_to_extract + ": " + err.what() );
	}

} // Archive::extract

//======================================================================

Pointer<Archive> Archive::open( const string& aUrl )
{
	return new Archive( aUrl, false );

} // Archive::open

//======================================================================

Pointer<Archive> Archive::create( const string& aUrl )
{
	return new Archive( aUrl, true );

} // Archive::create

//======================================================================
