/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/file/DirectoriesAccess.cc,v 1.4 2001/06/20 10:10:30 alex Exp $
 */

// -*- C++ -*-   Wed Dec 23 1998
// Eugene Skepner <eu@qub.com>
// $Id: DirectoriesAccess.cc,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef DirectoriesAccess_hh
#include "DirectoriesAccess.hh"
#endif

//======================================================================

Pointer<DirectoriesAccess> DirectoriesAccess::sDirectoriesAccess;

//======================================================================

DirectoriesAccess::DirectoriesAccess()
	: mDeny(), mAllow()
{
} // DirectoriesAccess::DirectoriesAccess

//======================================================================

DirectoriesAccess::~DirectoriesAccess()
{
} // DirectoriesAccess::~DirectoriesAccess

//======================================================================

bool DirectoriesAccess::accessAllowed( const string& aDir ) const
{
	if ( aDir[0] != '/' ) {
		cerr << "Warning: Access is not allowed to non-absolute directories specifications: " << aDir << endl;
		return(false);
	}

	string longest_denial;
	for ( List::const_iterator i = mDeny.begin(); i != mDeny.end(); ++i ) {
		const DirComparison dc = compare( aDir, *i );
		if ( dc == dcEqual ) {
			longest_denial = *i;
			break;
		}
		else if ( dc == dcSubdirectory && longest_denial.length() < i->length() )
			longest_denial = *i;
	}

	if ( longest_denial == aDir )
		return(false); // even if it is in allow list as well

	if ( longest_denial.length() > 0 ) {
		for ( List::const_iterator i = mAllow.begin(); i != mAllow.end(); ++i ) {
			if ( compare( *i, longest_denial ) == dcSubdirectory ) {
				const DirComparison dc = compare( aDir, *i );
				if ( dc == dcEqual || dc == dcSubdirectory ) {
					longest_denial = "";
					break;
				}
			}
		}
	}

	if ( longest_denial.length() > 0 )
		cerr << "Warning: Access is not allowed to " << aDir << endl;

	return(longest_denial.length() == 0);

} // DirectoriesAccess::accessAllowed

//======================================================================

void DirectoriesAccess::deny( const List& aDeny )
{
	mDeny = aDeny;
	normalize( mDeny );

} // DirectoriesAccess::deny

//======================================================================

void DirectoriesAccess::allow( const List& aAllow )
{
	mAllow = aAllow;
	normalize( mAllow );

} // DirectoriesAccess::allow

//======================================================================

void DirectoriesAccess::normalize( List& aDirs )
{
	for ( List::iterator i = aDirs.begin(); i != aDirs.end(); ++i ) {
		// remove duplicated slashes
		string::size_type pos;
		while ( ( pos = i->find( "//" ) ) != string::npos )
			i->erase( pos, 1 );

		// remove trailing slash
		if ( i->length() > 1 && *i->rend() == '/' )
			i->erase( i->end() - 1 );
	}

} // DirectoriesAccess::normalize

//======================================================================

DirectoriesAccess::DirComparison DirectoriesAccess::compare( const string& aAsked, const string& aMain ) const
{
	const string::size_type alen = aAsked.length();
	const string::size_type mlen = aMain.length();

	DirComparison dc = dcNoRelated;

	if ( mlen == 1 && aMain[0] == '/' && aAsked[0] == '/' ) { // aMain is root and aAsked is absolute
		dc = dcSubdirectory;
	}
	else if ( mlen < alen && !strncmp( aAsked.c_str(), aMain.c_str(), mlen ) ) {
		if ( aAsked[mlen] == '/' )
			dc = dcSubdirectory;
	}
	else if ( alen == mlen && aAsked == aMain )
		dc = dcEqual;

	return dc;

} // DirectoriesAccess::compare

//======================================================================

Pointer<DirectoriesAccess> DirectoriesAccess::instance()
{
	if ( !sDirectoriesAccess )
		sDirectoriesAccess = new DirectoriesAccess();
	return(sDirectoriesAccess);

} // DirectoriesAccess::instance

//======================================================================
