/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/file/Fetch.cc,v 1.4 2001/06/20 10:10:30 alex Exp $
 */

// -*- C++ -*-   Wed Dec 23 1998
// Eugene Skepner <eu@qub.com>
// $Id: Fetch.cc,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef Fetch_hh
#include "Fetch.hh"
#endif

#include <stdexcept>
#include <cerrno>
#include <unistd.h>

#ifdef FETCH_LIB
extern "C" {
#include <fetch.h>
#include <com_err.h>
}
#endif

//======================================================================

Pointer<Fetch> Fetch::sFetch = 0;

bool Fetch::sComErrHookInstalled = false;
string Fetch::sComErr;

//======================================================================

Fetch::Fetch()
{
#ifdef FETCH_LIB
	if ( !sComErrHookInstalled ) {
		set_com_err_hook( &com_err_hook );
		sComErrHookInstalled = true;
	}
#endif
} // Fetch::Fetch

//======================================================================

Fetch::~Fetch()
{
} // Fetch::~Fetch

//======================================================================

bool Fetch::isRemote( const string& aUrl )
{
	if ( !hasSchemeSpecifier( aUrl ) || !strncmp( aUrl.c_str(), "file:", 5 ) )
		return(false);
	else
		return(true);

} // Fetch::isRemote

//======================================================================

void Fetch::stat( const string& aUrl, Stat& aStat )
{
#ifdef FETCH_LIB
	url_stat st;
	if ( fetchStatURL( const_cast<char*>( addSchemeSpecifier( aUrl ).c_str() ), &st, 0 ) == -1 ) {
		//throw runtime_error( string( "Cannot stat file " ) + aUrl + ": " + sComErr );
		st.size = 0;
		st.mtime = 0;
	}

	aStat.size = st.size;
	aStat.time = st.mtime;
	aStat.type = Stat::Regular;
#else
	throw runtime_error( "Libfetch support disabled" );
#endif
} // Fetch::stat

//======================================================================

Pointer<Container<list<string> > > Fetch::directory( const string& /*aUrl*/ )
{
#ifdef FETCH_LIB
	return 0;
#else
	throw runtime_error( "Libfetch support disabled" );
#endif
} // Fetch::directory

//======================================================================

void Fetch::download( const string& aUrl, const string& aLocalName )
{
#ifdef FETCH_LIB
	FILE* in = 0;
	FILE* out = 0;

	try {
		in = fetchGetURL( const_cast<char*>( addSchemeSpecifier( aUrl ).c_str() ), 0 );
		if ( in == 0 )
			throw runtime_error( string( "Cannot open URL " ) + aUrl + ": " + sComErr );
		out = fopen( aLocalName.c_str(), "w" );
		if ( out == 0 )
			throw runtime_error( string( "Cannot open file " ) + aLocalName + ": " + ::strerror( errno ) );

		int c;
		while ( ( c = getc( in ) ) != EOF )
			putc( c, out );

		fclose( out );
		fclose( in );
	}
	catch ( ... ) {
		if ( in )
			fclose( in );
		if ( out )
			fclose( out );
		throw;
	}
#else
	throw runtime_error( "Libfetch support disabled" );
#endif
} // Fetch::download

//======================================================================

bool Fetch::hasSchemeSpecifier( const string& aUrl )
{
	bool has = false;
	for ( size_t i = 0; i < aUrl.length(); ++i ) {
		if ( aUrl[i] == '/' )
			break;
		if ( aUrl[i] == ':' ) {
			has = true;
			break;
		}
	}
	return(has);

} // Fetch::hasSchemeSpecifier

//======================================================================

string Fetch::addSchemeSpecifier( const string& aUrl )
{
	if ( hasSchemeSpecifier( aUrl ) ) {
		return(aUrl);
	}
	else if ( aUrl[0] == '/' ) {
		return(string( "file:" ) + aUrl);
	}
	else {
		char buf[MAXPATHLEN+1];
		if ( getcwd( buf, MAXPATHLEN + 1 ) == 0 )
			throw runtime_error( string( "Cannot get current working directory: " ) + ::strerror( errno ) );
		return(string( "file:" ) + buf + "/" + aUrl);
	}

} // Fetch::addSchemeSpecifier

//======================================================================

void Fetch::com_err_hook( const char* whoami, long code, const char* format, va_list args )
{
#ifdef FETCH_LIB
	sComErr = "";
	if ( code ) {
		if ( whoami )
			sComErr += string( whoami ) + ": ";

		sComErr += string( error_message( code ) ) + " ";

		if ( format ) {
			char s[2048];
			vsprintf( s, format, args );
			sComErr += s;
		}
	}
#endif
} // Fetch::com_err_hook

//======================================================================

Pointer<Fetch> Fetch::instance()
{
	if ( !sFetch )
		sFetch = new Fetch;
	return(sFetch);

} // Fetch::instance

//======================================================================
