/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/file/File.cc,v 1.6 2001/09/05 22:36:40 alex Exp $
 */

// -*- C++ -*-   Wed Dec 23 1998
// Eugene Skepner <eu@qub.com>
// $Id: File.cc,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef File_hh
#include "File.hh"
#endif

#ifndef Fetch_hh
#include "Fetch.hh"
#endif

#ifndef String_hh
#include "String.hh"
#endif

#ifndef Configuration_hh
#include "Configuration.hh"
#endif

#ifndef RealFile_hh
#include "RealFile.hh"
#endif

#include <unistd.h>
#include <cerrno>
#include <stdexcept>
#include <sys/stat.h>
#include <dirent.h>
#include <fnmatch.h>

//======================================================================

File::File()
	: mRemoveOnClose( false )
{
} // File::File

//======================================================================

File::File( const string& aUrl )
	: mUrl( aUrl ), mRemoveOnClose( false ), mFile(NULL)
{
} // File::File

//======================================================================

File::~File()
{
#ifndef LANGUAGE_INTERFACE
	if (mFile != NULL)
		::fclose(mFile);
	if ( mRemoveOnClose && mLocalName.length() > 0 )
		::unlink( mLocalName.c_str() );
#endif
} // File::~File

//======================================================================

string File::open_local( bool remove_on_close )
{
#ifndef LANGUAGE_INTERFACE
	if ( isLocal() ) {
		mRemoveOnClose = false;
		return localName();
	}
	else {
		mRemoveOnClose = remove_on_close;
		Fetch::Stat st;
		Fetch::instance()->stat( mUrl, st );
		if ( st.type != Fetch::Stat::Regular )
			throw runtime_error( string( "Cannot download non-regular file " ) + mUrl );
		Fetch::instance()->download( mUrl, localName() );
		return localName();
	}
#endif
} // File::open_local

//======================================================================

void File::close()
{
#ifndef LANGUAGE_INTERFACE
	if ( mUrl.length() ) {
		if ( mLocalName.length() > 0 ) {
			if (mFile != NULL) {
				::fclose(mFile);
				mFile = NULL;
			}
			if ( mRemoveOnClose ) {
				::unlink( mLocalName.c_str() );
				mRemoveOnClose = false;
			}
		}
		else {
			throw runtime_error( string( "Unable to close file " ) + mUrl + ": file was not opened locally" );
		}
	}
#endif
} // File::close

//======================================================================

FILE* File::open_for_reading( bool remove_local_copy_on_close )
{
#ifndef LANGUAGE_INTERFACE
	if ( !isLocal() ) {
		Fetch::Stat st;
		Fetch::instance()->stat( mUrl, st );
		if ( st.type != Fetch::Stat::Regular )
			throw runtime_error( string( "Cannot open non-regular file " ) + mUrl + " for reading" );
		Fetch::instance()->download( mUrl, localName() );
		mRemoveOnClose = remove_local_copy_on_close;
	}
	FILE* f = mFile = ::fopen( localName().c_str(), "r" );
	if ( f == 0 )
		throw runtime_error( string( "Unable to open local file " ) + localName() + ": " + ::strerror( errno ) );
	return f;
#endif
} // File::open_for_reading

//======================================================================

FILE* File::create()
{
#ifndef LANGUAGE_INTERFACE
	if ( isLocal() ) {
		FILE* f = mFile = ::fopen( localName().c_str(), "w+" );
		if ( f == 0 )
			throw runtime_error( string( "Unable to create local file " ) + mUrl + ": " + ::strerror( errno ) );
		return f;
	}
	else {
		throw runtime_error( string( "Unable to create remote file " ) + mUrl + ": not yet implemented" );
	}
	return NULL;
#endif
} // File::create

//======================================================================

bool File::isLocal() const
{
#ifndef LANGUAGE_INTERFACE
	return ! Fetch::instance()->isRemote( mUrl );
#endif
} // File::isLocal

//======================================================================

string File::localName() const // returns an empty string if file is remote
{
#ifndef LANGUAGE_INTERFACE
	if ( mLocalName.length() == 0 ) {
		if ( isLocal() ) {
			if ( !strncmp( mUrl.c_str(), "file:", 5 ) ) {
				mLocalName = mUrl.substr( 5 );
			}
			else {
				mLocalName = mUrl;
			}
		}
		else {
			static char templ[ PATH_MAX ];
			strcpy( templ, ( Configuration::instance()->tempDir() + "/libhfile-XXXXXX" ).c_str() );
			const char* s = ::mktemp( templ );
			if ( s == 0 )
				throw runtime_error( string( "Unable to find unique name for temporary file in " ) + Configuration::instance()->tempDir() + ": " + ::strerror( errno ) );
			mLocalName = s;
		}
	}
	return(mLocalName);
#endif
} // File::localName

//======================================================================

void File::patch( const string& file_to_patch, const void* patch_data, size_t patch_length, bool reverse = false )
{
#ifndef LANGUAGE_INTERFACE
	string cmd = Configuration::instance()->path_to_patch() + ( reverse ? " -R" : "" ) + " -t -s " + file_to_patch;
	if ( Configuration::instance()->no_action() ) {
		cerr << "NO ACTION: Execute patch: `" << cmd << "'" << endl;
	}
	else {
		FILE* patch_process = ::popen( cmd.c_str(), "w" );
		if ( patch_process == 0 )
			throw runtime_error( string( "Unable to run \"" ) + cmd + "\": " + ::strerror( errno ) );
		if ( ::fwrite( patch_data, 1, patch_length, patch_process ) != patch_length ) {
			::pclose( patch_process );
			throw runtime_error( string( "Unable to write all data top patch process \"" ) + cmd + "\": " + ::strerror( errno ) );
		}
		const int status = pclose( patch_process );
		if ( status )
			throw runtime_error( string( "Patch command: \"" ) + cmd + "\" returned status: " + to_string( status ) );
	}
#endif
} // File::patch

//======================================================================

int File::exec( const string& aCommand )
{
#ifndef LANGUAGE_INTERFACE
	if ( aCommand.length() > 0 ) {
		return ::system( aCommand.c_str() );
	}
	else
		return 0;
#endif
} // File::exec

//======================================================================

void File::make_path( const string& path )
{
#ifndef LANGUAGE_INTERFACE
	const int mode = 0755;

	string dirpath = path;

	struct stat stats;
	if ( ::stat( dirpath.c_str(), &stats ) ) {
		char* slash = const_cast<char*>( dirpath.c_str() );
		while ( *slash == '/' )
			slash++;
		while ( ( slash = strchr( slash, '/' ) ) ) {
			*slash = '\0';
			if ( ::stat( dirpath.c_str(), &stats ) ) {
				if ( ::mkdir( dirpath.c_str(), mode ) )
					throw runtime_error( string( "cannot create directory \"" ) + dirpath + "\": " + ::strerror( errno ) );
			} else if ( !S_ISDIR( stats.st_mode ) )
				throw runtime_error( string( "\"" ) + dirpath + "\" exists but is not a directory" );

			*slash++ = '/';

			// Avoid unnecessary calls to `stat' when given
			// pathnames containing multiple adjacent slashes.
			while (*slash == '/')
				slash++;
		}

		// We're done making leading directories. Create the final
		// component of the path.

		// The path could end in "/." or contain "/..", so test if
		// we really have to create the directory.

		if ( ::stat( dirpath.c_str(), &stats ) && ::mkdir( dirpath.c_str(), mode ) )
			throw runtime_error( string( "cannot create directory \"" ) + dirpath + "\": " + ::strerror( errno ) );
	}
	else {
		// The entire path already exists

		if ( !S_ISDIR( stats.st_mode ) )
			throw runtime_error( string( "\"" ) + dirpath + "\" exists but is not a directory" );

	}
#endif
} // File::make_path

//======================================================================

Pointer<Container<list<Pointer<RealFile> > > > File::scan( const string& aDir, bool aRecurseIntoSubdirs, bool aIncludeDirs, Pointer<Container<list<string> > > aIgnoreNames )
{
#ifndef LANGUAGE_INTERFACE
	Pointer<Container<list<Pointer<RealFile> > > > files = new Container<list<Pointer<RealFile> > >;
	read_directory( aDir, aRecurseIntoSubdirs, aIncludeDirs, aIgnoreNames, *files );
	return(files);
#endif
} // File::scan

//======================================================================

void File::read_directory( const string& aDir, bool aRecurseIntoSubdirs, bool aIncludeDirs, Pointer<Container<list<string> > > aIgnoreNames, list<Pointer<RealFile> >& files )
{
#ifndef LANGUAGE_INTERFACE
	DIR* dir = opendir( aDir.c_str() );
	if ( dir == 0 )
		throw runtime_error( string( "Unable to open directory " ) + aDir + ": " + ::strerror( errno ) );

	try {
		for ( dirent* d = readdir( dir ); d; d = readdir( dir ) ) {
			if ( strcmp( d->d_name, "." ) && strcmp( d->d_name, ".." ) ) {
				bool add = true;
				if ( aIgnoreNames ) {
					for ( list<string>::const_iterator ignore = aIgnoreNames->begin(); add && ignore != aIgnoreNames->end(); ++ignore )
						add = fnmatch( ignore->c_str(), d->d_name, FNM_PATHNAME | FNM_PERIOD ) != 0;
				}
				if ( add )
					add_file( d->d_name, aDir, aRecurseIntoSubdirs, aIncludeDirs, aIgnoreNames, files );
			}
		}
	}
	catch ( ... ) {
		closedir( dir );
		throw;
	}
	closedir( dir );
#endif
} // File::read_directory

//======================================================================

void File::add_file( const string& aName, const string& aDir, bool aRecurseIntoSubdirs, bool aIncludeDirs, Pointer<Container<list<string> > > aIgnoreNames, list<Pointer<RealFile> >& files )
{
#ifndef LANGUAGE_INTERFACE
	Pointer<RealFile> file = new RealFile( aDir + "/" + aName );
	if ( S_ISDIR( file->mode() ) ) {
		if ( aIncludeDirs )
			files.push_back( file );
		if ( aRecurseIntoSubdirs )
			read_directory( aDir + "/" + aName, aRecurseIntoSubdirs, aIncludeDirs, aIgnoreNames, files );
	}
	else {
		files.push_back( file );
	}
#endif
} // File::add_file

//======================================================================

bool File::exist( const string& aPathname )
{
	struct stat stats;
	return(::lstat( aPathname.c_str(), &stats ) == 0);

} // File::exist

//======================================================================
