/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/file/RealFile.cc,v 1.4 2001/06/20 10:10:31 alex Exp $
 */

// -*- C++ -*-   Mon Dec 21 1998
// Eugene Skepner <eu@qub.com>
// $Id: RealFile.cc,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef RealFile_hh
#include "RealFile.hh"
#endif

#ifndef ChangeDirTemporarily_hh
#include "ChangeDirTemporarily.hh"
#endif

#ifndef File_hh
#include "File.hh"
#endif

#ifndef String_hh
#include "String.hh"
#endif

#include <cerrno>
#include <pwd.h>
#include <grp.h>
#include <utime.h>
#include <sys/param.h>
#include <sys/mount.h>
#include <unistd.h>

//======================================================================

RealFile::RealFile( const string& aPathname )
	: mPathname( aPathname )
{
	if ( lstat( mPathname.c_str(), &mStat ) )
		throw runtime_error( string( "Unable to lstat " ) + mPathname + ": " + ::strerror( errno ) );

} // RealFile::RealFile

//======================================================================

RealFile::RealFile( const string& aPathname, struct stat aStat )
	: mPathname( aPathname ), mStat( aStat )
{
} // RealFile::RealFile

//======================================================================

string RealFile::md5() const
{
	if ( S_ISREG( mStat.st_mode ) )
		return File::md5( pathname() );
	else
		return string();

} // RealFile::md5

//======================================================================

string RealFile::user() const
{
	const passwd* pwd = getpwuid( mStat.st_uid );
	if ( pwd == 0 ) {
		cerr << "Warning: User name for uid " << mStat.st_uid << " not found, using \"nobody\"" << endl;
		return("nobody");
	}
	else {
		return(pwd->pw_name);
	}

} // RealFile::user

//======================================================================

string RealFile::group() const
{
	const struct group* grp = getgrgid( mStat.st_gid );
	if ( grp == 0 ) {
		cerr << "Warning: Group name for gid " << mStat.st_gid << " not found, using \"nobody\"" << endl;
		return("nobody");
	}
	else {
		return(grp->gr_name);
	}

} // RealFile::group

//======================================================================

string RealFile::symlink() const
{
	char buf[1026];
	int len = 0;
	if ( S_ISLNK( mStat.st_mode ) ) {
		if ( ( len = readlink( pathname().c_str(), buf, 1024 ) ) == -1 ) {
			throw runtime_error( string( "Unable to readlink " ) + pathname() + ": " + ::strerror( errno ) );
		}
	}
	buf[len] = 0;
	return(buf);

} // RealFile::symlink

//======================================================================

string RealFile::filesystem() const
{
	string dir = pathname();
	if ( !S_ISDIR( mStat.st_mode ) ) { // FILE is not a directory
		dir.erase( dir.find_last_not_of( '/' ) + 1 ); // strip trailing slashes
		dir = File::dirname( dir.c_str() );
	}

	ChangeDirTemporarily cdt( dir );

	struct stat last_stat;
	if ( stat( ".", &last_stat ) < 0 )
		throw runtime_error( string( "Unable to stat . of " ) + dir + ": " + ::strerror( errno ) );

	for ( ;; ) {
		struct stat st;
		if ( stat( "..", &st ) < 0 )
			break; // looks like root
		if ( st.st_dev != last_stat.st_dev || st.st_ino == last_stat.st_ino )
			break; // cwd is the mount point.
		if ( chdir( ".." ) < 0 )
			break; // consider cwd is the mount point.
		last_stat = st;
	}

	char buf[MAXPATHLEN+1];
	if ( getcwd( buf, MAXPATHLEN + 1 ) == 0 )
		throw runtime_error( string( "Cannot get current working directory: " ) + ::strerror( errno ) );

	return(buf);

} // RealFile::filesystem

//======================================================================

void RealFile::mtime( time_t aTime )
{
	utimbuf timbuf;
	timbuf.actime = timbuf.modtime = aTime;
	if ( utime( pathname().c_str(), &timbuf ) )
		throw runtime_error( "Cannot utime " + pathname() + ": " + ::strerror( errno ) );
	lstat( mPathname.c_str(), &mStat );

} // RealFile::mtime

//======================================================================

void RealFile::user( const string& aUser )
{
	uid_t uid = 65534;
	struct passwd* pwd = getpwnam( aUser.c_str() );
	if ( pwd == 0 ) {
		cerr << "Warning: Uid for user name " << aUser << " not found, using \"nobody\"" << endl;
		pwd = getpwnam( "nobody" );
	}

	if ( pwd != 0 ) {
		uid = pwd->pw_uid;
	}

	if ( mStat.st_uid != uid ) {
		if ( lchown( pathname().c_str(), uid, gid_t( -1 ) ) )
			throw runtime_error( "Cannot chown " + pathname() + " to user " + aUser + ": " + ::strerror( errno ) );
		lstat( mPathname.c_str(), &mStat );
	}

} // RealFile::user

//======================================================================

void RealFile::group( const string& aGroup )
{
	gid_t gid = 65534;
	struct group* grp = getgrnam( aGroup.c_str() );
	if ( grp == 0 ) {
		cerr << "Warning: Group name for group " << aGroup << " not found, using \"nobody\"" << endl;
		grp = getgrnam( "nobody" );
	}
	if ( grp != 0 ) {
		gid = grp->gr_gid;
	}

	if ( mStat.st_gid != gid ) {
		if ( lchown( pathname().c_str(), uid_t( -1 ), gid ) )
			throw runtime_error( "Cannot chown " + pathname() + " to group " + aGroup + ": " + ::strerror( errno ) );
		lstat( mPathname.c_str(), &mStat );
	}

} // RealFile::group

//======================================================================

void RealFile::mode( mode_t aMode )
{
	if ( chmod( pathname().c_str(), aMode ) )
		throw runtime_error( "Cannot chmod " + pathname() + ": " + ::strerror( errno ) );
	lstat( mPathname.c_str(), &mStat );

} // RealFile::mode

//======================================================================

Pointer<Container<vector<long> > > RealFile::device() const
{
	dev_t devt = dev();
	Pointer<Container<vector<long> > > result;
	if ( major( devt ) == 0 ) {
		result = new Container<vector<long> >;
	}
	else {
		result = new Container<vector<long> >( 2 );
		(*result)[0] = major( devt );
		(*result)[1] = minor( devt );
	}
	return(result);

} // RealFile::device

//======================================================================

string RealFile::inode() const
{
	return to_string( mStat.st_ino );

} // RealFile::inode

//======================================================================
