/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/file/ziplib.hh,v 1.3 2001/06/20 10:10:31 alex Exp $
 */

// -*- C++ -*-  Mon Sep  7 1998
// $Id: ziplib.hh,v 1.1.1.1 2000/02/15 16:30:34 alex Exp $

#ifndef ziplib_hh
#define ziplib_hh

#include <sys/types.h>
#include <stdexcept>

//======================================================================

class Zip
{
  public:
     class error : public runtime_error
     {
       public:
          error( const char* a, const char* b = "", const char* c = "", const char* d = "", const char* e = "" );
     };

     class Entry
     {
       public:
          ~Entry() { delete filename_; delete extra_; delete comment_; }

          const char* name() const { return filename_; }
          size_t compressedSize() const { return comp_size; }
          size_t uncompressedSize() const { return uncomp_size; }
          mode_t mode() const { return mode_; }
          void mode( mode_t aMode ) { mode_ = aMode; }
          time_t time() const;
          void time( time_t aTime );

          Entry& operator = ( const Entry& entry );


       private:
          Entry();
          size_t fill( const char* data );
          size_t push( char* data ) const;
          friend class Zip;

          u_char id[4];
          u_int16_t version_made_by;
          u_int16_t version_to_extract;
          u_int16_t gp_flag;
          u_int16_t comp_method;
          u_int32_t m_time;
          u_int32_t crc32;
          u_int32_t comp_size;
          u_int32_t uncomp_size;
          u_int16_t filename_len;
          u_int16_t extra_len;
          u_int16_t comment_len;
          u_int16_t start_disk_no;
          u_int16_t int_attrib;
          u_int16_t ext_attrib;
          u_int16_t mode_;
          u_int32_t local_offset;
          char* filename_;
          char* extra_;
          char* comment_;
            //ZipAttrs       *attrs;
     };

     Zip();
     ~Zip();

     void open( const char* filename );
     void create( const char* filename );
     size_t numberOfEntries() const;
     const Entry* entries() const; // last entry has name 0
     size_t find_entry( const char* name );
       // removes file from the archive
     void remove( const char* name );
       // extract file into data (allocated or mmaped by user),
       // if data==0 it is allocated and returned
     char* extract( const char* name, char* data, size_t& len );
       // Returns uncompressed size of file
     size_t length( const char* name );
       // add file into the archive with name
     void add( const char* name, const char* data, size_t len, mode_t aMode, time_t aTime );

  private:
     struct ECHDR {
          ECHDR() : comment( 0 ) {}
          ~ECHDR() { delete comment; }
          void fill( const char* data );
          void push( char* data ) const;

          u_char id[4];
          u_int16_t this_diskno;
          u_int16_t first_diskno;
          u_int16_t this_disk_num_entry;
          u_int16_t num_entry;
          u_int32_t cendir_size;
          u_int32_t cendir_offset;
          u_int16_t comment_len;
          char* comment;
     };

     struct LocHDR {
          LocHDR() : filename( 0 ), extra( 0 ), data( 0 ) {}
          ~LocHDR() { delete filename; delete extra; }
          void fill( const char* data );

          u_char id[4];
          u_int16_t version_to_extract;
          u_int16_t gp_flag;
          u_int16_t comp_method;
          u_int32_t m_time;
          u_int32_t crc32;
          u_int32_t comp_size;
          u_int32_t uncomp_size;
          u_int16_t filename_len;
          u_int16_t extra_len;
          char* filename;
          char* extra;
          char* data;
     };

     int mFd;
     bool mReadOnly;
     size_t mSize;
     char* mMappedFile;
     ECHDR mHeader;
     Entry* mEntries;

     char* find_end_of_central_directory();
     char* inflate( const LocHDR& h, const char* compressed, char* result );
     void close();
     void add_entry( const Entry& entry );
     void remove_entry( size_t entryno );
     void recreate_center_dir( char* data );

}; // class Zip

//======================================================================

#endif // ziplib_hh
