/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/hui/Hui.hh,v 1.17 2001/06/20 10:24:44 alex Exp $
 */

// -*- C++ -*-  Tue Dec  1 1998
// Eugene Skepner <eu@qub.com>
// $Id: Hui.hh,v 1.1.1.1 2000/02/15 16:30:35 alex Exp $

#ifndef Hui_hh
#define Hui_hh

#ifndef LanguageInterface_hh
#include "LanguageInterface.hh"
#endif

#ifndef Container_hh
#include "Container.hh"
#endif

#ifndef Log_hh
#include "Log.hh"
#endif

#include <stdexcept>
#include <limits>
#include <vector>

#define ARRANGE_FREE		0
#define ARRANGE_HORIZONTAL	1
#define ARRANGE_VERTICAL	2

/* ====================================================================== */

class Hui : virtual public LanguageInterface::Object
{
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;

public:
	class Error : public runtime_error
	{
	public:
		Error( const string& what_arg ) : runtime_error( what_arg ) {}
	};

	virtual void run() = 0; /* Run the application */
	virtual void exit(int status) = 0; /* exit the application */
	virtual void processPendingEvents() = 0;

	enum Type {tAuto, tText, tGraphics};
	virtual const char* type() const = 0; /* type of UI */
	static Pointer<Hui> open(const string& aType);
	static Pointer<Hui> open(Type type = tAuto);

	virtual ~Hui();

	/* -------------------------------------------------- */

	typedef short Dimension;
	typedef pair<Dimension,Dimension> DimPair;
	typedef vector<Dimension> DimVector;

	class Position;

	class DimensionPair : public DimPair
	{
	public:
		DimensionPair(Dimension a, Dimension b) {}
		DimensionPair(const DimVector& v) 
		{
			if (v.size() == 2) {
				first = v[0];
				second = v[1];
			}
		}
		virtual ~DimensionPair() {}
		
		operator Pointer< Container< vector<Dimension> > > () const
		{
			Pointer< Container< vector<Dimension> > > r = new Container< vector<Dimension> >(2);
			(*r)[0] = first;
			(*r)[1] = second;
			return r;
		}

	}; /* class DimensionPair */

	class Size : public DimensionPair
	{
	public:
		Size(const Dimension aWidth = 0,
		     const Dimension aHeight = 0)
			: DimensionPair(aWidth, aHeight) {
			first = aWidth;
			second = aHeight;
		}
		Size(const vector<Dimension>& v) 
			: DimensionPair(v){
			/* XXX missing initialization */
		}

		Dimension width() const {
			return first;
		}
		void width(Dimension aWidth) {
			first = aWidth;
		}
		Dimension height() const {
			return second;
		}
		void height(Dimension aHeight) {
			second = aHeight;
		}

		static Dimension maxWidth() {
			return numeric_limits<Dimension>::max();
		}
		static Dimension maxHeight() {
			return numeric_limits<Dimension>::max();
		}

		Size& operator += (const Size& aSize)
		{
			if (aSize.width() != maxWidth() && width() != maxWidth())
				width(width() + aSize.width());
			else
				width(maxWidth());
			
			if (aSize.height() != maxHeight() && height() != maxHeight())
				height(height() + aSize.height());
			else
				height(maxHeight());
			return(*this);
		}
		Size& operator += (const Position& aPosition);
		Size operator + ( const Position& aPosition ) const
		{
			Size result = *this;
			result += aPosition;
			return(result);
		}
		Size& operator -= (const Position& aPosition);
		Size operator - (const Position& aPosition) const
		{
			Size result = *this;
			result -= aPosition;
			return(result);
		}
		Size& operator -= (const Size& aSize)
		{
			if (width() != maxWidth())
				width(width() - aSize.width());
			if (height() != maxHeight())
				height(height() - aSize.height());
			return(*this);
		}
		Size operator - (const Size& aSize) const
		{
			Size result = *this;
			result -= aSize;
			return(result);
		}
		Size operator + (const Size& aSize) const
		{
			Size result = *this;
			result += aSize;
			return(result);
		}
		
	}; /* class Size */

	virtual Size size() const = 0;

	class Position : public DimensionPair
	{
	public:
		Position(Dimension aX = 0,
			 Dimension aY = 0) 
			: DimensionPair(aX, aY) {
			x(aX);
			y(aY);
		}
		Position(const vector<Dimension>& v) 
			: DimensionPair(v) {
			/* XXX probably missing initialization, too */
		}

		Dimension x() const {
			return first;
		}
		void x(Dimension aX) {
			first = aX;
		}
		Dimension y() const {
			return second;
		}
		void y(Dimension aY) {
			second = aY;
		}
		Position operator + (const Position& p) const {
			return(Position(x() + p.x(),
					y() + p.y()));
		}

	}; /* class Position */

	/* -------------------------------------------------- */

	class Widget;
	class GroupWidget;

	class WidgetCallbackData : public LanguageInterface::SpecificCallbackData
	{
	public:
		WidgetCallbackData(Hui& aHui,
				   Widget& aWidget,
				   Pointer<LanguageInterface::CallbackData> aCallbackData)
			: LanguageInterface::SpecificCallbackData(aCallbackData),
			  mHui(aHui),
			  mWidget(aWidget) {
		}

		virtual const char* typeid_name() const {
			return(typeid(Hui).name());
		}

		virtual const string& activate(bool call_callback_completed);

		template <typename Arg> void activate_with_arg(Arg arg)
		{
			try {
				if (callbackData())
					callbackData()->activate(*this, arg);
			}
			catch (exception& err) {
				LOG(ERROR, "Hui::WidgetCallbackData::activate_with_arg: " << err.what());
			}
		}
		
		string& result() {
			return mResult;
		}
		const string& result() const {
			return mResult;
		}

		Pointer<Hui> hui() {
			return &mHui;
		}
		Pointer<Widget> widget() {
			return &mWidget;
		}

	private:
		Hui& mHui;
		Widget& mWidget;
		string mResult;
		
	}; /* class WidgetCallbackData */

	/* -------------------------------------------------- */

	class Widget : virtual public LanguageInterface::Object
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;
		
	public:
		virtual ~Widget() {}

		virtual string name(const string& aName);
		virtual const string& name() const {
			return mName;
		}
		virtual Pointer<Hui::Widget> byName(const string& aName) const
		{
			return(name() == aName ?
			       const_cast<Widget*>(this) 
			       : 0);
		}

		virtual void show() = 0;
		virtual void hide() = 0;
		virtual bool hidden() const = 0;

		virtual void disable() = 0;
		virtual void enable() = 0;
		virtual bool disabled() const = 0;

		virtual void enter();
		virtual void leave();
		virtual void onEnter(Pointer<LanguageInterface::CallbackData> aCallbackData) {
			mOnEnter.callbackData(aCallbackData);
		}
		virtual Pointer<LanguageInterface::CallbackData> onEnter() const {
			return(mOnEnter.callbackData());
		}
		virtual void onLeave(Pointer<LanguageInterface::CallbackData> aCallbackData ) {
			mOnLeave.callbackData(aCallbackData);
		}
		virtual Pointer<LanguageInterface::CallbackData> onLeave() const {
			return(mOnLeave.callbackData());
		}

		virtual void move(const Position& aPosition) = 0;
		virtual Position position() const = 0;
		virtual void resize(const Size& aSize) = 0;
		virtual Size size() const = 0;

		virtual void callback_completed(const WidgetCallbackData& /*aCallbackData*/) {
			/* XXX empty? what is this for? */
		}

	protected:
		Widget(Hui& aHui, const string& aName);
		
	private:
		string mName;
		WidgetCallbackData mOnEnter;
		WidgetCallbackData mOnLeave;
		
	}; /* class Widget */
       
	/* -------------------------------------------------- */

	class LabelWidget;
	class ButtonWidget;
	class InputLineWidget;
	class CheckBoxesWidget;
	class RadioMenuWidget;
	class ListBoxWidget;
	class TextEditWidget;
	class ProgressIndicatorWidget;
	class ContainerWidget;

	/* -------------------------------------------------- */

	class EmbeddedWidget : virtual public Widget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;
		
	protected:
		EmbeddedWidget(Hui& aHui, const string& aName)
			: Widget( aHui, aName ) {
		}

	}; /* class EmbeddedWidget */

	/* -------------------------------------------------- */

	class GroupWidget : virtual public Widget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;
	     
	public:
		
		virtual Pointer<Hui::LabelWidget> label(const string& aName,
							const string& aText,
							const Hui::Position& aPosition) = 0;
		virtual Pointer<Hui::ButtonWidget> button(const string& aName,
							  const string& aText,
							  Pointer<LanguageInterface::CallbackData> aCallbackData,
							  const string& aButtonType,
							  const Hui::Position& aPosition,
							  const string& aComment) = 0;
		virtual Pointer<Hui::InputLineWidget> inputLine(const string& aName,
								const string& aInitialText,
								int aMaxLength,
								const string& aValidator,
								const string& aTitle,
								const Hui::Position& aPosition,
								const string& aComment) = 0;
		virtual Pointer<Hui::CheckBoxesWidget> checkBoxes(const string& aName,
								  const vector<string>& aTexts,
								  const vector<bool>& aInitialValues,
								  const string& aTitle,
								  const Hui::Position& aPosition,
								  const vector<string>& aComments) = 0;
		virtual Pointer<Hui::RadioMenuWidget> radioMenu(const string& aName,
								const vector<string>& aTexts,
								const int aInitialValue,
								const string& aTitle,
								const Hui::Position& aPosition,
								const vector<string>& aComments) = 0;
		virtual Pointer<Hui::ListBoxWidget> listBox(const string& aName,
							    const vector<string>& aTexts,
							    const int aInitiallySelected,
							    const string& aTitle,
							    const Hui::Position& aPosition,
							    const string& aComment) = 0;
		virtual Pointer<Hui::TextEditWidget> textEdit(const string& aName,
							      const string& aInitialText,
							      bool aAllowEdit,
							      const string& aTitle,
							      const Hui::Position& aPosition,
							      const string& aComment) = 0;
		virtual Pointer<Hui::ProgressIndicatorWidget> progressIndicator(const string& aName,
										float aMin,
										float aMax,
										float aInitial,
										const string& aTitle,
										const Hui::Position& aPosition,
										const string& aComment) = 0;
		virtual Pointer<Hui::ContainerWidget> container(const string& aName,
								const Hui::Position& aPosition,
								const string& aAlignment,
								const string& aComment) = 0;

		virtual void arrange() = 0;
		virtual void arrange(const string& aAlignmen, bool arrange_new) = 0;

		virtual Pointer<Hui::Widget> byName(const string& aName) const = 0;

		virtual Pointer<Hui::Widget> current() const = 0;
		virtual void current(Pointer<Hui::EmbeddedWidget> aWidget) = 0;
	     
		virtual void erase() = 0;
	     
	protected:
		GroupWidget(Hui& aHui, const string& aName)
			: Widget(aHui, aName) {
		}
		Hui& hui() {
		}
	     
	}; /* class GroupWidget */
	
	/* -------------------------------------------------- */
	
    class ContainerWidget : virtual public GroupWidget
    {
	
	CLASS_METHODS_DESCRIPTIONS_DECLARATION;
	
    public:
	
    protected:
	ContainerWidget(Hui& aHui, const string& aName)
	    : Widget(aHui, aName),
	      GroupWidget(aHui, aName)
	    {
	    }
	
    }; /* class ContainerWidget */
    
    /* -------------------------------------------------- */

    class ToplevelWidget : virtual public GroupWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;
		
	public:
		virtual void title(const string& aTitle) = 0;
		virtual string title() const = 0;
		
	protected:
		ToplevelWidget( Hui& aHui, const string& aName )
			: Widget(aHui, aName),
			  GroupWidget(aHui, aName) {
		}
		
	}; /* class ToplevelWidget */

	/* -------------------------------------------------- */

	class DialogWidget : virtual public ToplevelWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;
		
		virtual const string& result() const = 0;
		virtual const string& exec() = 0;
		
	protected:
		DialogWidget(Hui& aHui, const string& aName)
			: Widget(aHui, aName),
			  GroupWidget(aHui, aName),
			  ToplevelWidget(aHui, aName) {
		}

	}; /* class DialogWidget */

	/* -------------------------------------------------- */

	class WindowWidget : virtual public ToplevelWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;
		
	protected:
		WindowWidget(Hui& aHui, const string& aName)
			: Widget(aHui, aName),
			  GroupWidget(aHui, aName),
			  ToplevelWidget(aHui, aName) {
		}

	}; /* class WindowWidget */

	/* -------------------------------------------------- */

	class LabelWidget : virtual public EmbeddedWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	public:
		virtual void text(const string& aText) = 0;
		virtual string text() const = 0;
		
	protected:
		LabelWidget(Hui& aHui, const string& aName)
			: Widget(aHui, aName),
			  EmbeddedWidget(aHui, aName) {
		}

	}; /* class LabelWidget */

	// --------------------------------------------------

	class ButtonWidget : virtual public EmbeddedWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	public:
		virtual void text(const string& aText) = 0;
		virtual string text() const = 0;
		
		virtual void click() {
			mOnClick.activate(true);
		}
		virtual void onClick(Pointer<LanguageInterface::CallbackData> aCallbackData) {
			mOnClick.callbackData(aCallbackData);
		}
		virtual Pointer<LanguageInterface::CallbackData> onClick() const {
			return(mOnClick.callbackData());
		}

	protected:
		ButtonWidget(Hui& aHui,
			     const string& aName,
			     Pointer<LanguageInterface::CallbackData> aCallbackData)
			: Widget(aHui, aName),
			  EmbeddedWidget(aHui, aName),
			  mOnClick(aHui, *this, aCallbackData) {
		}

	private:
		WidgetCallbackData mOnClick;

	}; /* class ButtonWidget */

	/* -------------------------------------------------- */

	class InputLineWidget : virtual public EmbeddedWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	public:
		virtual void title(const string& aTitle) = 0;
		virtual string title() const = 0;
		virtual void text(const string& aText) = 0;
		virtual string text() const = 0;

	protected:
		InputLineWidget(Hui& aHui, const string& aName, int aMaxLength)
			: Widget(aHui, aName),
			  EmbeddedWidget(aHui, aName),
			  mMaxLength( aMaxLength) {
		}
		int maxLength() const {
			return(mMaxLength);
		}

	private:
		int mMaxLength;

	}; /* class InputLineWidget */

	/* -------------------------------------------------- */

	class ButtonsGroupWidget : virtual public EmbeddedWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	public:
		virtual void text(int boxno, const string& aText) = 0;
		virtual string text(int boxno) const = 0;
		virtual void title(const string& aTitle) = 0;
		virtual string title() const = 0;
		virtual void disable_item(int boxno) = 0;
		virtual void enable_item(int boxno) = 0;
		virtual bool item_disabled(int boxno) const = 0;
		virtual int numberOfItems() const = 0;

		virtual void click(int boxno) {
			mOnClick.activate_with_arg(boxno);
		}
		virtual void onClick(Pointer<LanguageInterface::CallbackData> aCallbackData) {
			mOnClick.callbackData( aCallbackData);
		}
		virtual Pointer<LanguageInterface::CallbackData> onClick() const {
			return(mOnClick.callbackData());
		}

	protected:
		ButtonsGroupWidget(Hui& aHui, const string& aName)
			: Widget(aHui, aName),
			  EmbeddedWidget(aHui, aName),
			  mOnClick(aHui, *this, 0) {
		}

	private:
		WidgetCallbackData mOnClick;

	}; /* class ButtonsGroupWidget */

	/* -------------------------------------------------- */

	class CheckBoxesWidget : virtual public ButtonsGroupWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	public:
		virtual bool state(int boxno, int aNewState) = 0;
		virtual bool state(int boxno) const = 0;
		virtual bool toggle(int boxno) = 0;

	protected:
		CheckBoxesWidget(Hui& aHui, const string& aName)
			: Widget(aHui, aName),
			  EmbeddedWidget(aHui, aName),
			  ButtonsGroupWidget(aHui, aName) {
		}

	}; /* class CheckBoxesWidget */

	/* -------------------------------------------------- */

	class RadioMenuWidget : virtual public ButtonsGroupWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	public:
		virtual void set(int itemno) = 0;
		virtual int set() const = 0;

	protected:
		RadioMenuWidget(Hui& aHui, const string& aName)
			: Widget(aHui, aName),
			  EmbeddedWidget(aHui, aName),
			  ButtonsGroupWidget(aHui, aName) {
		}

	}; /* class RadioMenuWidget */

	/* -------------------------------------------------- */

	class ListBoxWidget : virtual public EmbeddedWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	public:
		virtual void title(const string& aTitle) = 0;
		virtual string title() const = 0;
		virtual int numberOfItems() const = 0;
		virtual string text(int aItemNo) const = 0;
		virtual void text(int aItemNo, const string& aText) = 0;
 		virtual void select(int aItemNo) = 0;
		virtual int select() const = 0;
		virtual void add(int aBefore, const string& aText) = 0;
		virtual void remove(int aItemNo) = 0;
		virtual void removeAll() = 0;
		virtual void highlighted(int itemno) {
			mOnSelect.activate_with_arg(itemno);
		}
		virtual void onSelect(Pointer<LanguageInterface::CallbackData> aCallbackData) {
			mOnSelect.callbackData( aCallbackData);
		}
		virtual Pointer<LanguageInterface::CallbackData> onSelect() const {
			return(mOnSelect.callbackData());
		}

	protected:
		ListBoxWidget(Hui& aHui, const string& aName)
			: Widget(aHui, aName),
			  EmbeddedWidget(aHui, aName),
			  mOnSelect(aHui, *this, 0) {
		}

	private:
		WidgetCallbackData mOnSelect;

	}; /* class ListBoxWidget */

	/* -------------------------------------------------- */

	class TextEditWidget : virtual public EmbeddedWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	public:
		virtual void title(const string& aTitle) = 0;
		virtual string title() const = 0;
		virtual void text(const string& aText) = 0;
		virtual string text() const = 0;

	protected:
		TextEditWidget(Hui& aHui, const string& aName)
			: Widget(aHui, aName), EmbeddedWidget(aHui, aName) {
		}

	}; /* class TextEditWidget */

	/* -------------------------------------------------- */

	class ProgressIndicatorWidget : virtual public EmbeddedWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	public:
		virtual void title(const string& aTitle) = 0;
		virtual string title() const = 0;
		virtual void set( float aValue ) = 0;
		virtual float set() const = 0;

	protected:
		ProgressIndicatorWidget(Hui& aHui, const string& aName)
			: Widget(aHui, aName),
			  EmbeddedWidget(aHui, aName) {
		}

	}; /* class ProgressIndicatorWidget */

	/* -------------------------------------------------- */

	class MenuItemWidget : virtual public EmbeddedWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	public:
		virtual void text(const string& aText) = 0;
		virtual string text() const = 0;
		virtual string key() const = 0;

		virtual void click() {
			mOnClick.activate(true);
		}
		virtual void onClick(Pointer<LanguageInterface::CallbackData> aCallbackData) {
			mOnClick.callbackData(aCallbackData);
		}
		virtual Pointer<LanguageInterface::CallbackData> onClick() const {
			return(mOnClick.callbackData());
		}

		virtual void move(const Position& /*aPosition*/) {
		}
		virtual Position position() const {
			return Position();
		}
		virtual void resize(const Size& /*aSize*/) {
		}
		virtual Size size() const {
			return Size();
		}

		virtual bool removeItem(Pointer<MenuItemWidget> /*aItem*/) {
			return(false);
		}

	protected:
		MenuItemWidget(Hui& aHui,
			       const string& aName,
			       Pointer<LanguageInterface::CallbackData> aCallbackData)
			: Widget(aHui, aName),
			  EmbeddedWidget(aHui, aName),
			  mOnClick(aHui, *this, aCallbackData) {
		}

	private:
		WidgetCallbackData mOnClick;

	}; /* class MenuItemWidget */

	// --------------------------------------------------

	class MenuSeparatorWidget : virtual public MenuItemWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	public:
		virtual void text(const string& /*aText*/) {
		}
		virtual string text() const {
			return string();
		}
		virtual string key() const {
			return string();
		}

		virtual void click() {
		}

	protected:
		MenuSeparatorWidget(Hui& aHui, const string& aName)
			: Widget(aHui, aName),
			  EmbeddedWidget(aHui, aName),
			  MenuItemWidget(aHui, aName, 0) {
		}

	}; /* class MenuSeparatorWidget */

	/* -------------------------------------------------- */

	class MenuButtonWidget : virtual public MenuItemWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	public:
		virtual void key(const string& aKey) = 0;
		virtual string key() const {
			return(mKey);
		}
		virtual string text() const {
			return(mText);
		}

	protected:
		MenuButtonWidget(Hui& aHui,
				 const string& aName,
				 const string& aText,
				 Pointer<LanguageInterface::CallbackData> aCallbackData,
				 const string& aKey,
				 const string& aComment)
			: Widget(aHui, aName),
			  EmbeddedWidget(aHui, aName),
			  MenuItemWidget(aHui, aName, aCallbackData),
			  mText(aText),
			  mKey(aKey),
			  mComment(aComment) {
		}

		virtual void keyText(const string& aKey) {
			mKey = aKey;
		}
		virtual void setText(const string& aText) {
			mText = aText;
		}

	private:
		string mText;
		string mKey;
		string mComment;

	}; /* class MenuButtonWidget */

	/* -------------------------------------------------- */

	class MenuSubmenuWidget : virtual public MenuItemWidget
	{
		CLASS_METHODS_DESCRIPTIONS_DECLARATION;

	public:
		typedef list< Pointer<MenuItemWidget> > Items;

		virtual Pointer<MenuItemWidget> addButton(const string& aName,
							  const string& aText,
							  Pointer<LanguageInterface::CallbackData> aCallbackData,
							  const string& aKey,
							  const string& aComment,
							  Pointer<MenuItemWidget> aBefore) = 0;
		virtual Pointer<MenuItemWidget> addSeparator(const string& aName,
							     Pointer<MenuItemWidget> aBefore) = 0;
		virtual Pointer<MenuItemWidget> addSubmenu(const string& aName,
							   const string& aText,
							   const string& aComment,
							   Pointer<MenuItemWidget> aBefore) = 0;

		virtual void add(Pointer<MenuItemWidget> aBefore,
				 Pointer<MenuItemWidget> aItem) = 0;
		virtual void remove(Pointer<MenuItemWidget> aItem) = 0;

		virtual Pointer<Hui::Widget> byName(const string& aName) const;
		virtual string text() const {
			return(mText);
		}
		virtual string key() const {
			return string();
		}

		virtual const Items& items() const {
			return(mItems);
		}
		virtual Items& items() {
			return(mItems);
		}

	protected:
		MenuSubmenuWidget(Hui& aHui,
				  const string& aName,
				  const string& aText,
				  const string& aComment)
			: Widget(aHui, aName),
			  EmbeddedWidget(aHui, aName),
			  MenuItemWidget(aHui, aName, 0),
			  mText(aText),
			  mComment(aComment) {
		}

		virtual void addItem(Pointer<MenuItemWidget> aBefore, Pointer<MenuItemWidget> aItem);
		virtual bool removeItem(Pointer<MenuItemWidget> aItem);
		virtual void setText(const string& aText) {
			mText = aText;
		}

	private:
		string mText;
		Items mItems;
		string mComment;

	}; /* class MenuSubmenuWidget */

	/*  -------------------------------------------------- */

	virtual Pointer<Hui::Widget> byName(const string& aName) = 0;

	virtual Pointer<Hui::WindowWidget> window(const string& aName,
						  const string& aTitle,
						  const Position& aPosition,
						  const Size& aSize,
						  const string& aAlignment,
						  bool no_menu = false) = 0;
	virtual Pointer<Hui::DialogWidget> dialog(Pointer<Hui::Widget> aParent,
						  const string& aName,
						  const string& aTitle,
						  const Position& aPosition,
						  const Size& aSize,
						  const string& aAlignment) = 0;
	virtual Pointer<Hui::MenuSubmenuWidget> menuBar() = 0;

	/* Returns an empty string if user pressed cancel button */
	virtual string fileSelectionDialog(const string& aTitle,
					   const string& aWildCard) = 0;

protected:
	Hui();
	
private:
	
}; /* class Hui */

/*====================================================================== */

#endif /* Hui_hh */
