/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/hui/HuiQt_InputLineWidget.cc,v 1.28 2001/06/20 10:24:46 alex Exp $
 */

// -*- C++ -*-   Wed Dec  2 1998
// Eugene Skepner <eu@qub.com>
// $Id: HuiQt.cc,v 1.1.1.1 2000/02/15 16:30:35 alex Exp $

#ifndef HuiQt_hh
#include "HuiQt.hh"
#endif

#ifndef QLABEL_H
#include <qlabel.h>
#endif

#ifndef QVALIDATOR_H
#include <qvalidator.h>
#endif

//======================================================================

HuiQt::InputLineWidget::InputLineWidget(Hui& aHui,
					Pointer<GroupWidget> aParent,
					const string& aName,
					const string& aInitialText,
					int aMaxLength,
					const string& aValidator,
					const string& aTitle,
					const string& aComment)
	: Hui::Widget(aHui, aName),
	  Hui::EmbeddedWidget(aHui, aName),
	  TitledWidget(aHui, aParent, aName, aTitle),
	  Hui::InputLineWidget(aHui, aName, aMaxLength),
	  mValidator( make_lineedit_validator( aValidator ) ),
	  mComment( aComment )
{
	parent(aParent);

	ContainerWidget *cp = dynamic_cast<ContainerWidget*>((Hui::Widget*) parent());
	GroupWidget *gp = dynamic_cast<GroupWidget*>((Hui::Widget*) parent());
	
	if (cp != NULL)
		mInputLine = new QLineEditEnterLeave(const_cast<QWidget*>(cp->qwidget()),
						     aName.c_str());
	else if (gp != NULL)
		mInputLine = new QLineEditEnterLeave(const_cast<QWidget*>(gp->qwidget()),
						     aName.c_str());

	if (qtitle())
		qtitle()->setBuddy(mInputLine);

	text( aInitialText );
	mInputLine->setMaxLength( maxLength() );
	mInputLine->setValidator( mValidator );

	mInputLine->resize( mInputLine->sizeHint() );
	mInputLine->setMinimumSize( mInputLine->sizeHint() );
	setsize(mInputLine->size());

	mInputLine->connect( mInputLine, SIGNAL( enter_signal() ), &huiQtEnterLeave(), SLOT( enter() ) );
	mInputLine->connect( mInputLine, SIGNAL( leave_signal() ), &huiQtEnterLeave(), SLOT( leave() ) );

} // HuiQt::InputLineWidget::InputLineWidget

//======================================================================

QValidator* HuiQt::make_lineedit_validator( const string& aValidatorType )
{
	QValidator* validator = 0;
	if ( aValidatorType == "none" || aValidatorType == "" ) {
		validator = 0;
	}
	else if ( !strncmp( aValidatorType.c_str(), "picture", 7 ) ) {
		cerr << "Warning: Picture validator is not supported by Qt yet, (\"none\" will be used)" << endl;
		validator = 0;
	}
	else if ( !strncmp( aValidatorType.c_str(), "valid_chars", 11 ) ) {
		cerr << "Warning: valid_chars validator is not supported by Qt yet, (\"none\" will be used)" << endl;
		validator = 0;
	}
	else if ( !strncmp( aValidatorType.c_str(), "integer_range", 13 ) ) {
		long min, max;
		if ( sscanf( aValidatorType.c_str(), "integer_range %ld %ld", &min, &max ) != 2 ) {
			cerr << "Warning: Invalid integer_range validator argument, -Inf +Inf will be used)" << endl;
			min = numeric_limits<long>::min();
			max = numeric_limits<long>::max();
		}
		validator = new QIntValidator( min, max, 0 );
	}
	else if ( !strncmp( aValidatorType.c_str(), "real_range", 10 ) ) {
		double min, max;
		if ( sscanf( aValidatorType.c_str(), "real_range %lf %lf", &min, &max ) != 2 ) {
			cerr << "Warning: Invalid real_range validator argument, -Inf +Inf will be used)" << endl;
			min = numeric_limits<double>::min();
			max = numeric_limits<double>::max();
		}
		validator = new QDoubleValidator( min, max, 12, 0 );
	}
	else {
		cerr << "Warning: Unknown validator `" << aValidatorType << "' for InputLine, \"none\" will be used" << endl;
		validator = 0;
	}

	return(validator);

} // HuiQt::make_lineedit_validator

//======================================================================

void HuiQt::InputLineWidget::text( const string& aText )
{
	if ( mInputLine )
		mInputLine->setText( aText.c_str() );

} // HuiQt::InputLineWidget::text

//======================================================================

string HuiQt::InputLineWidget::text() const
{
	if ( mInputLine )
		return static_cast<const char*>(mInputLine->text());
	else
		return string();

} // HuiQt::InputLineWidget::text

//======================================================================

QWidget* HuiQt::InputLineWidget::qwidget()
{
	return(mInputLine);

} // HuiQt::InputLineWidget::qwidget

//======================================================================

const QWidget* HuiQt::InputLineWidget::qwidget() const
{
	return(mInputLine);

} // HuiQt::InputLineWidget::qwidget

//======================================================================
