/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/hui/HuiTVision_InputLineWidget.cc,v 1.30 2001/06/20 10:24:50 alex Exp $
 */

// -*- C++ -*-   Wed Dec  2 1998
// Eugene Skepner <eu@qub.com>
// $Id: HuiTVision.cc,v 1.1.1.1 2000/02/15 16:30:35 alex Exp $

#ifndef HuiTVision_hh
#include "HuiTVision.hh"
#endif

#define Uses_TDialog
#define Uses_TDeskTop
#define Uses_TLabel
#define Uses_TButton
#define Uses_TInputLine
#define Uses_TRangeValidator
#define Uses_TPXPictureValidator
#define Uses_MsgBox
#define Uses_TSItem
#define Uses_TRadioButtons
#define Uses_TCheckBoxes
#define Uses_TListBox
#define Uses_TScrollBar
#define Uses_TCollection
#define Uses_TStringCollection
#define Uses_TFileDialog
#define Uses_TMenuItem
#define Uses_TMenu
#define Uses_TSubMenu
#define Uses_TMenuBar
#define Uses_TKeys
#define Uses_TScroller
#define Uses_TEditor
#define Uses_TIndicator

#include <tvision/tv.h>

#ifndef String_hh
#include "String.hh"
#endif

#include <cmath>

#define MAX(a,b) ((a) > (b) ? (a) : (b))

//======================================================================

class TIntegerRangeValidator : public TFilterValidator
{
public:
	TIntegerRangeValidator( long aMin, long aMax)
		: TFilterValidator( 0 ), min( aMin ), max( aMax )
		{
			if ( aMin < 0 )
				validChars = newStr( "-0123456789" );
			else
				validChars = newStr( "0123456789" );
		}

	virtual void error()
		{
			messageBox( mfError | mfOKButton, _("Value not in the range %ld to %ld"), min, max );
		}

	virtual Boolean isValid(const char* s)
		{
			long value;
			if ( TFilterValidator::isValid( s ) )
				if ( sscanf( s, "%ld", &value ) == 1 )
					if ( ( value >= min ) &&  ( value <= max ) )
						return(True);
			return(False);
		}

private:
	long min;
	long max;

}; // class TIntegerRangeValidator

//======================================================================

class TRealRangeValidator : public TFilterValidator
{
public:
	TRealRangeValidator( double aMin, double aMax)
		: TFilterValidator( 0 ), min( aMin ), max( aMax )
		{
			validChars = newStr( "0123456789+-eE." );
		}

	virtual void error()
		{
			messageBox( mfError | mfOKButton, _("Value not in the range %lg to %lg"), min, max );
		}

	virtual Boolean isValid(const char* s)
		{
			double value;
			if ( TFilterValidator::isValid( s ) )
				if ( sscanf( s, "%le", &value ) == 1 )
					if ( ( value >= min ) &&  ( value <= max ) )
						return(True);
			return(False);
		}

private:
	double min;
	double max;

}; // class TRealRangeValidator

//======================================================================

class HuiTInputLine : public TInputLine
{
public:
	HuiTInputLine( const string& aText, int aMaxLength, const string& aValidator )
		: TInputLine( TRect( 1, 1, 11, 2 ), aMaxLength, 0 )
		{
			setData( const_cast<char*>( aText.c_str() ) );
			setValidator( aValidator );
		}

	virtual string text() const
		{
			char s[ const_cast<HuiTInputLine*>( this )->dataSize() ];
			const_cast<HuiTInputLine*>( this )->getData( s );
			return s;
		}

	virtual void text( const string& aText )
		{
			setData( const_cast<char*>( aText.c_str() ) );
		}

	virtual void setValidator( const string& aValidator )
		{
			TValidator* validator = 0;
			if ( aValidator == "none" || aValidator == "" )
				validator = 0;
			else if ( !strncmp( aValidator.c_str(), "picture", 7 ) )
				validator = new TPXPictureValidator( aValidator.c_str() + 8, True );
			else if ( !strncmp( aValidator.c_str(), "valid_chars", 11 ) )
				validator = new TFilterValidator( aValidator.c_str() + 12 );
			else if ( !strncmp( aValidator.c_str(), "integer_range", 13 ) ) {
				long min, max;
				if ( sscanf( aValidator.c_str(), "integer_range %ld %ld", &min, &max ) != 2 ) {
					cerr << "Warning: Invalid integer_range validator argument, -Inf +Inf will be used)" << endl;
					min = numeric_limits<long>::min();
					max = numeric_limits<long>::max();
				}
				validator = new TIntegerRangeValidator( min, max );
			}
			else if ( !strncmp( aValidator.c_str(), "real_range", 10 ) ) {
				double min, max;
				if ( sscanf( aValidator.c_str(), "real_range %lf %lf", &min, &max ) != 2 ) {
					cerr << "Warning: Invalid real_range validator argument, -Inf +Inf will be used)" << endl;
					min = numeric_limits<double>::min();
					max = numeric_limits<double>::max();
				}
				validator = new TRealRangeValidator( min, max );
			}
			else {
				cerr << "Warning: Unknown validator for InputLineWidget: " << aValidator << " (\"none\" will be used)" << endl;
				validator = 0;
			}

			TInputLine::setValidator( validator );
		}

}; // class HuiTInputLine

//======================================================================

HuiTVision::InputLineWidget::InputLineWidget(Hui& aHui,
					     Pointer<Hui::Widget> aParent,
					     const string& aName,
					     const string& aInitialText,
					     int aMaxLength,
					     const string& aValidator,
					     const string& aTitle,
					     const string& aComment)
	: Hui::Widget(aHui, aName),
	  Hui::EmbeddedWidget(aHui, aName),
	  TitledWidget(aHui, aParent, aName, aTitle),
	  Hui::InputLineWidget(aHui, aName, aMaxLength),
	  mInputLine(new HuiTInputLine(aInitialText, aMaxLength, aValidator)),
	  mComment(aComment)
{
	if ( ttitle() )
		ttitle()->link( mInputLine );
	
	parent(aParent);
	resize( minSize() );

} // HuiTVision::InputLineWidget::InputLineWidget

//======================================================================

void HuiTVision::InputLineWidget::text( const string& aText )
{
	if ( mInputLine )
		mInputLine->text( aText );

} // HuiTVision::InputLineWidget::text

//======================================================================

string HuiTVision::InputLineWidget::text() const
{
	if ( mInputLine )
		return mInputLine->text();
	else
		return string();

} // HuiTVision::InputLineWidget::text

//======================================================================

TView* HuiTVision::InputLineWidget::tview()
{
	return(mInputLine);

} // HuiTVision::InputLineWidget::tview

//======================================================================

const TView* HuiTVision::InputLineWidget::tview() const
{
	return(mInputLine);

} // HuiTVision::InputLineWidget::tview

//======================================================================
