/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/sysinstall/Database.hh,v 1.2 2001/06/22 15:05:15 alex Exp $
 */

// -*- C++ -*-  Fri Aug  7 1998
// $Id: Database.hh,v 1.16 1999/02/03 14:43:05 eugene Exp $

#ifndef Database_hh
#define Database_hh

#ifndef LanguageInterface_hh
#include "LanguageInterface.hh"
#endif

#ifndef DatabaseEngine_hh
#include "DatabaseEngine.hh"
#endif

#ifndef Package_hh
#include "Package.hh"
#endif

#ifndef Feature_hh
#include "Feature.hh"
#endif

//======================================================================

class History;
typedef Container<list<Pointer<History> > > HistoryList;

//======================================================================

class Database : virtual public LanguageInterface::Object
{
     CLASS_METHODS_DESCRIPTIONS_DECLARATION;

  public:
     enum OpenFlag
     {
          ofReadonly,
          ofReadWrite, // if Configuration::no_action actually opens readonly
          ofCreate,    // Implies readwrite, if Configuration::no_action does not do anything
     };

     Database( const string& aBasedir, OpenFlag aOpenFlag, bool addSYSTEMpackage );
     virtual ~Database();

     static Pointer<Database> open( const string& aBasedir ) { return new Database( aBasedir, ofReadWrite, false ); }
     static Pointer<Database> create( const string& aBasedir ) { return new Database( aBasedir, ofCreate, true ); }
     static Pointer<Database> import_db( const string& aBasedir, const string& aFilename ) { Pointer<Database> db = new Database( aBasedir, ofCreate, false ); db->import_db( aFilename ); return db; }

       // Returns list of features matching feature_name. If there are
       // no installed features matching the specified name returns
       // an empty list.
     Pointer<FeatureList> feature( const string& aFeatureName, const string& aType );
     Pointer<FeatureList> all_features( const string& aType );

       // Returns list of features with the same name as specified
       // feature but of other versions or an empty list.
     Pointer<FeatureList> feature_other_versions( const Feature& aFeature, const string& aType );
     Pointer<FeatureList> feature_other_versions( const Pointer<Feature> aFeature, const string& aType ) { return feature_other_versions( *aFeature, aType ); }

       // Returns package providing the specified feature. If there is
       // no package providing the feature returns 0.
     Pointer<Package> package( const Feature& aFeature );
     Pointer<Package> package( const Pointer<Feature> aFeature ) { return package( *aFeature ); }

       // Returns packages with the specified name. If there is
       // no packages with this name returns 0.
     Pointer<PackageList> packages_by_name( const string& aPackageName );

       // Returns all packages installed.
     Pointer<PackageList> all_packages();

       // Returns list of packages owning the specified file.
     Pointer<PackageList> packages_owning( const string& aPathname );
     Pointer<PackageList> packages_owning( Pointer<PackageFile> aFile );

       // Returns list of packages requiring at least one of the
       // specified features. Versions of the features are ignored.
     Pointer<PackageList> packages_requiring( const FeatureList& aFeatures );

       // Returns list of packages conflicting the specified
       // feature. Version of the feature is ignored.
     Pointer<PackageList> packages_conflicting( const Feature& aFeature );
     Pointer<PackageList> packages_conflicting( const Pointer<Feature> aFeature ) { return packages_conflicting( *aFeature ); }

       // Returns undo (or NULL) for the specified package.
//       Pointer<Undo> package_undo( const Package& aPackage );

//         // Stores undo for the specified package.
//       void package_undo_store( const Package& aPackage, const Undo& aUndo );

//         // Erases undo for the specified package from the database.
//       void package_undo_erase( const Package& aPackage );

       // Returns new package handle to store in the database. If
       // Configuration::no_action returns non-zero does not do
       // anything just reports what is to be done and returns dummy
       // package handle.
     Pointer<Package> new_package();

#ifndef SYSINSTALL_FLOPPY
       // Updates records of the SYSTEM package and checks all the
       // packages for fitness into new hardware system. Returns list
       // of package having conflicts with the SYSTEM package. If
       // Configuration::no_action returns an emtpy list does not do
       // anything just reports what is to be done.
     Pointer<PackageList> hardware_upgrade();
#endif

#ifndef SYSINSTALL_FLOPPY
       // Returns list of packages having problems, e.g. requiring
       // features not provided by the system, conflicting features
       // provided by the system.
     Pointer<PackageList> check_conflicts();
#endif

#ifndef SYSINSTALL_FLOPPY
       // Returns list of features provided multiple times (different versions).
     Pointer<FeatureList> check_features_duplicates();
#endif

     virtual void updateHistory( const History& aEntry );
#ifndef SYSINSTALL_FLOPPY
     virtual Pointer<HistoryList> history();
#endif

       // Retrieves or stores configuration table record.
     Pointer<Container<list<string> > > readConfiguration( const string& aKey );
     void writeConfiguration( const string& aKey, const string& aValue );
     void writeConfiguration( const string& aKey, const list<string>& aValues );

     void saveDirectoriesAccess();
     void readDirectoriesAccess();

     virtual void sync();

     DatabaseEngine& engine() { return *mEngine; }

     enum TableId
     {
          tiPackages, tiFeatures, tiScripts, tiDescription,
          tiReferences, tiFiles, tiFileInfo, tiUndo, tiFilesUndo,
          tiConfiguration, tiHistory,
          tiNone
     };

     DatabaseEngine::Table& table( TableId aTableId ) { return *mTable[aTableId]; }

     virtual string select_string( TableId tid, DatabaseEngine::Id id, const string& aField );
     virtual void update( TableId tid, DatabaseEngine::Id id, const string& aField, const string& aValue );
     virtual time_t select_time( TableId tid, DatabaseEngine::Id id, const string& aField );
     virtual void update( TableId tid, DatabaseEngine::Id id, const string& aField, time_t aValue );

     static Pointer<FeatureList> records2features( DatabaseEngine::RecordList::const_iterator first, DatabaseEngine::RecordList::const_iterator last );

       // Returns list of all files owned (or not) by any package
     Pointer<PackageFileList> all_files();
       // Returns records in the database corresponding to the specified file
     Pointer<PackageFileList> file( const string& aPathname );

#ifndef SYSINSTALL_FLOPPY
     virtual void export_db( ostream& os );
     virtual void export_db( const string& aFilename );
     virtual void import_db( istream& is );
     virtual void import_db( const string& aFilename );
#endif

#ifndef SYSINSTALL_FLOPPY
       // Finds database inconsistencies and tries to solve them.
     virtual void repair();
#endif

  private:
     string mBaseDir;
     OpenFlag mOpenFlag;
     Pointer<DatabaseEngine> mEngine;
     DatabaseEngine::Table* mTable[tiNone];

     void create();
     void insert_SYSTEM_package();
     void insert_denied_allowed_directories();
     virtual TableId tableByName( const string& aTableName ) const;

     static const DatabaseEngine::FieldDescription sTablePackages[];
     static const DatabaseEngine::FieldDescription sTableFeatures[];
     static const DatabaseEngine::FieldDescription sTableScripts[];
     static const DatabaseEngine::FieldDescription sTableDescription[];
     static const DatabaseEngine::FieldDescription sTableReferences[];
     static const DatabaseEngine::FieldDescription sTableFiles[];
     static const DatabaseEngine::FieldDescription sTableFileInfo[];
     static const DatabaseEngine::FieldDescription sTableUndo[];
     static const DatabaseEngine::FieldDescription sTableFilesUndo[];
     static const DatabaseEngine::FieldDescription sTableConfiguration[];
     static const DatabaseEngine::FieldDescription sTableHistory[];

     static const DatabaseEngine::TableDescription sTables[];

     virtual Pointer<PackageList> unique_packages( const DatabaseEngine::RecordList& records );

#ifndef SYSINSTALL_FLOPPY
     virtual void import_table( TableId table, istream& is, size_t& lineno );
     virtual void skip_ws_comments( istream& is, size_t& lineno ) const;
#endif

#ifndef SYSINSTALL_FLOPPY
     virtual void remove_not_referenced_files();
     virtual void remove_not_referenced_features();
     virtual void remove_not_referenced_scripts();
     virtual void remove_not_referenced_description();
     virtual void remove_not_referenced_reference();
     virtual void remove_not_referenced_undo();
#endif

     virtual Pointer<FeatureList> make_architecture_features( string& architecture_feature );

}; // class Database

//======================================================================

class Record_to_PackageFile
{
  public:
     Record_to_PackageFile( Database& aDatabase ) : mDatabase( aDatabase ) {}
     Pointer<PackageFile> operator () ( Pointer<DatabaseEngine::Record> record );

  private:
     Database& mDatabase;
};

//======================================================================

#endif // Database_hh
