/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/sysinstall/Feature.cc,v 1.2 2001/06/22 15:05:15 alex Exp $
 */

// -*- C++ -*-   Fri Aug  7 1998
// $Id: Feature.cc,v 1.3 1998/12/31 13:31:21 eugene Exp $

#ifndef Feature_hh
#include "Feature.hh"
#endif

#ifndef String_hh
#include "String.hh"
#endif

#ifndef Log_hh
#include "Log.hh"
#endif

#include <cctype>
#include <cstdlib>

//======================================================================

Feature::Comparison Feature::compare( const Feature& toCompare ) const
{
     if ( name() != toCompare.name() )
          throw runtime_error( "Cannot compare features with different names" );

     const Comparison vercmp = compare_versions( toCompare );
     const Comparison sercmp = compare_serials( toCompare );
     const Comparison snapcmp = compare_snapshots( toCompare );

     switch ( vercmp ) {
       case cThisOlder:
       case cThisNewer:
          return vercmp;
       case cEqual:
       case cThisProbablyOlder:
       case cThisProbablyNewer:
       case cUnknown:
          switch ( sercmp ) {
            case cThisOlder:
            case cThisNewer:
               return sercmp;
            case cEqual:
            case cThisProbablyOlder:
            case cThisProbablyNewer:
               switch ( snapcmp ) {
                 case cThisOlder:
                 case cThisNewer:
                    return snapcmp;
                 case cEqual:
                 case cThisProbablyOlder:
                 case cThisProbablyNewer:
                 case cUnknown:
                    return sercmp;
               }
            case cUnknown:
               switch ( snapcmp ) {
                 case cThisOlder:
                 case cThisNewer:
                    return snapcmp;
                 case cEqual:
                 case cThisProbablyOlder:
                 case cThisProbablyNewer:
                 case cUnknown:
                    return vercmp == cUnknown ? snapcmp : vercmp;
               }
          }
     }

     return cUnknown;

} // Feature::compare

//======================================================================

Feature::Comparison Feature::compare_versions( const Feature& toCompare ) const
{
     const string& thisVersion = version();
     const string& aOtherFeatureVersion = toCompare.version();

     if ( thisVersion == aOtherFeatureVersion )
          return cEqual;

     const char* tv = thisVersion.c_str();
     const char* ov = aOtherFeatureVersion.c_str();

     while ( *tv && *ov ) {
          while ( *tv && !isalnum( *tv ) )
               ++tv;
          while ( *ov && !isalnum( *ov ) )
               ++ov;

          const char* tc = tv;
          const char* oc = ov;

          const bool compare_numbers = isdigit( *tc );
          if ( compare_numbers ) {
               while ( *tc && isdigit( *tc ) )
                    ++tc;
               while ( *oc && isdigit( *oc ) )
                    ++oc;
          }
          else {
               while ( *tc && isalpha( *tc ) )
                    ++tc;
               while ( *oc && isalpha( *oc ) )
                    ++oc;
          }

          if ( tc == tv && oc == ov )
               return cUnknown;
          if ( tc == tv )
               return cThisProbablyOlder;
          if ( oc == ov )
               return cThisProbablyNewer;

          if ( compare_numbers ) {
               const int tn = atoi( tv );
               const int on = atoi( ov );
               if ( tn < on )
                    return cThisOlder;
               if ( tn > on )
                    return cThisNewer;
          }
          else {
               const string thisSubver( tv, tc - tv );
               const string otherSubver( ov, oc - ov );
               if ( thisSubver < otherSubver )
                    return cThisOlder;
               if ( thisSubver > otherSubver )
                    return cThisNewer;
          }

          tv = tc;
          ov = oc;
     }

    if ( !*tv && !*ov )
         return cEqual;

    if ( !*tv )
         return cThisOlder;
    else
         return cThisNewer;

} // Feature::compare_versions

//======================================================================

Feature::Comparison Feature::compare_serials( const Feature& toCompare ) const
{
     if ( serial() == 0 || toCompare.serial() == 0 )
          return cUnknown;
     if ( serial() > toCompare.serial() )
          return cThisNewer;
     if ( serial() < toCompare.serial() )
          return cThisOlder;
     return cEqual;

} // Feature::compare_serials

//======================================================================

Feature::Comparison Feature::compare_snapshots( const Feature& toCompare ) const
{
     if ( snapshot() == 0 || toCompare.snapshot() == 0 )
          return cUnknown;
     if ( snapshot() > toCompare.snapshot() )
          return cThisNewer;
     if ( snapshot() < toCompare.snapshot() )
          return cThisOlder;
     return cEqual;

} // Feature::compare_snapshots

//======================================================================

string Feature::print() const
{
     return name() + " " + version() + " " + to_string( serial() ) + " " + to_string( snapshot() );

} // Feature::print

//======================================================================

size_t Feature::scan( const char* text )
{
     char name[1024], version[1024], serial[32], snapshot[64];
     if ( sscanf( text, "%s %s %s %s", name, version, serial, snapshot ) != 4 )
          throw runtime_error( string( "Cannot scan feature from: " ) + text );
     mName = name;
     mVersion = version;
     mSerial = from_string( serial, mSerial );
     mSnapshot = from_string( snapshot, mSnapshot );

     return strlen( name ) + strlen( version ) + strlen( serial ) + strlen( snapshot ) + 3;

} // Feature::scan

//======================================================================
