/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/sysinstall/FeaturesProvided.cc,v 1.4 2001/07/17 20:36:52 alex Exp $
 */

// -*- C++ -*-   Mon Sep 14 1998
// $Id: FeaturesProvided.cc,v 1.5 1999/01/14 21:48:34 eugene Exp $

#ifndef FeaturesProvided_hh
#include "FeaturesProvided.hh"
#endif

#ifndef PackageStream_hh
#include "PackageStream.hh"
#endif

#ifndef Archive_hh
#include "Archive.hh"
#endif

#ifndef Log_hh
#include "Log.hh"
#endif

#ifndef Callback_hh
#include "Callback.hh"
#endif

#ifndef File_hh
#include "File.hh"
#endif

#ifndef ProgressIndicator_hh
#include "ProgressIndicator.hh"
#endif

//======================================================================

Pointer<FeaturesProvided> featuresProvided;

//======================================================================

const char* const package_suffixes[] =
{
     ".pkg.zip",
     0
};

//======================================================================

const char* FeaturesProvided::sDatabaseFile = "FeaturesProvided.si2";

//======================================================================

Pointer<FeaturesProvided> FeaturesProvided::instance()
{
     if ( !featuresProvided )
          featuresProvided = new FeaturesProvided();
     return featuresProvided;

} // FeaturesProvided::instance

//======================================================================

FeaturesProvided::FeaturesProvided()
{
} // FeaturesProvided::FeaturesProvided

//======================================================================

FeaturesProvided::~FeaturesProvided()
{
} // FeaturesProvided::~FeaturesProvided

//======================================================================

void FeaturesProvided::load_dir( const string& aUrl, bool search_subdirs )
{
     if ( !load_file( aUrl + "/" + sDatabaseFile ) ) {
          File dir_access( aUrl );
          if ( !dir_access.isLocal() )
               throw runtime_error( string( "FeaturesProvided: Cannot scan remote directory " ) + aUrl );

          Pointer<Container<list<Pointer<RealFile> > > > files = File::scan( dir_access.localName(), search_subdirs, false, 0 );
          for ( list<Pointer<RealFile> >::const_iterator file = files->begin(); file != files->end(); ++file ) {
               if ( S_ISREG( (*file)->mode() ) ) {
                    for ( const char* const* suffix = package_suffixes; *suffix; ++suffix ) {
                         if ( (*file)->pathname().length() > strlen( *suffix ) && !strcmp( (*file)->pathname().c_str() + (*file)->pathname().length() - strlen( *suffix ), *suffix ) )
                              scan_archive( (*file)->pathname() );
                    }
               }
          }
     }

} // FeaturesProvided::load_dir

//======================================================================

void FeaturesProvided::scan_archive( const string& aPathname )
{
     Pointer<Archive> stream = new Archive( aPathname, false );
     Pointer<PackageStream> packageStream = new PackageStream( stream, "FeaturesProvided_", 0 );
     Pointer<FeatureList> features = packageStream->features();
     for ( FeatureList::const_iterator i = features->begin(); i != features->end(); ++i ) {
          add_feature( **i, aPathname );
     }

} // FeaturesProvided::scan_archive

//======================================================================

void FeaturesProvided::add_feature( const Feature& aFeature, const string& aPathname )
{
     list<Entry>::const_iterator existing = entry_by_feature( aFeature, true );
     if ( existing != mEntries.end() ) {
          if ( aPathname != existing->url )
               LOG( WARNING, "FeaturesProvided: same versions of feature " << aFeature.name() << " provided by packages with different URLs (former will be used): " << existing->url << " " << aPathname );
     }
     else {
          Entry e;
          e.url = aPathname;
          e.feature = aFeature;
          mEntries.push_back( e );
     }

} // FeaturesProvided::add_feature

//======================================================================

list<FeaturesProvided::Entry>::const_iterator FeaturesProvided::entry_by_feature( const Feature& aFeature, bool exact )
{
     list<Entry>::const_iterator result = mEntries.end();
     for ( list<Entry>::const_iterator e = mEntries.begin(); e != mEntries.end(); ++e ) {
          if ( exact && aFeature == e->feature ) {
               result = e;
               break;
          }
          else if ( e->feature.name() == aFeature.name() ) {
               if ( result != mEntries.end() ) {
                    const Feature::Comparison comp = result->feature.compare( e->feature );
                    if ( comp == Feature::cThisOlder || comp == Feature::cThisProbablyOlder )
                         result = e;
               }
               else {
                    const Feature::Comparison comp = aFeature.compare( e->feature );
                    if ( comp == Feature::cEqual || comp == Feature::cThisOlder || comp == Feature::cThisProbablyOlder )
                         result = e;
               }
          }
     }
     return result;

} // FeaturesProvided::entry_by_feature

//======================================================================

bool FeaturesProvided::load_file( const string& aUrl )
{
     bool result = true;
     try {
          File file_access( aUrl );
          FILE* file = file_access.open_for_reading( true );
          try {
               const int LINE_SIZE = 2048;
               char line[LINE_SIZE];
               while( ::fgets( line, LINE_SIZE, file ) ) {
                    Feature feature;
                    const size_t scanned = feature.scan( line );
                    char package[1024];
                    if ( ::sscanf( line + scanned + 1, "%s", package ) != 1 )
                         throw runtime_error( string( "Cannot scan FeaturesProvided::Entry from: " ) + line );
                    add_feature( feature, File::dirname( aUrl ) + "/" + package );
               }
          }
          catch ( ... ) {
               file_access.close();
               throw;
          }
          file_access.close();
     }
     catch ( ... ) {
          result = false;
     }
     return result;

} // FeaturesProvided::load_file

//======================================================================

void FeaturesProvided::save( const string& aDirUrl )
{
     File file_access( aDirUrl + "/" + sDatabaseFile );
     FILE* file = file_access.create();
     for ( list<Entry>::const_iterator i = mEntries.begin(); i != mEntries.end(); ++i ) {
          const string feature = i->feature.print();
          fprintf( file, "%s %s\n", feature.c_str(), File::basename( i->url, "" ).c_str() );
     }
     file_access.close();

} // FeaturesProvided::save

//======================================================================

string FeaturesProvided::package( const Feature& aFeature )
{
     list<Entry>::const_iterator existing = entry_by_feature( aFeature, false );
     if ( existing != mEntries.end() )
          return existing->url;
     else
          return string();

} // FeaturesProvided::package

//======================================================================

size_t FeaturesProvided::packages( const FeatureList& aFeatures, list<string>& aPackages )
{
     size_t features_satisfied = 0;
     for ( FeatureList::const_iterator i = aFeatures.begin(); i != aFeatures.end(); ++i ) {
          const string url = package( **i );
          if ( url.length() > 0 ) {
               list<string>::iterator p = ::find( aPackages.begin(), aPackages.end(), url );
               if ( p == aPackages.end() )
                    aPackages.push_back( url );
               ++features_satisfied;
          }
     }

     return features_satisfied;

} // FeaturesProvided::packages

//======================================================================

size_t FeaturesProvided::find_and_install_features( Pointer<Database> database, const FeatureList& features, Pointer<ProgressIndicatorStack> aProgressIndicatorStack )
{
     list<string> packs;
     packages( features, packs );
     size_t installed_packages = 0;
     for ( list<string>::const_iterator i = packs.begin(); i != packs.end(); ++i ) {
          const string action = Callback::install_upgrade( database, *i );
          if ( action == "install" || action == "upgrade" ) {
               try {
                    Pointer<Archive> stream = new Archive( *i, false );
                    Pointer<PackageStream> packageStream = new PackageStream( stream, "install_", aProgressIndicatorStack );
                    if ( action == "install" )
                         packageStream->install( database );
                    else
                         packageStream->upgrade( database );
                    ++installed_packages;
               }
               catch ( runtime_error& err ) {
                      // ignore errors
                    LOG( ERROR, "FeaturesProvided: Unable to install " << *i << ": " << err.what() );
               }
          }
          else if ( action.length() > 0 )
               throw runtime_error( action );
     }

     return installed_packages;

} // FeaturesProvided::find_and_install_features

//======================================================================
