/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/sysinstall/Package.cc,v 1.2 2001/06/22 15:05:16 alex Exp $
 */

// -*- C++ -*-   Thu Aug 13 1998
// $Id: Package.cc,v 1.17 1999/02/03 14:43:05 eugene Exp $

#ifndef Package_hh
#include "Package.hh"
#endif

#ifndef Database_hh
#include "Database.hh"
#endif

#ifndef Feature_hh
#include "Feature.hh"
#endif

#ifndef PackageDescription_hh
#include "PackageDescription.hh"
#endif

#ifndef Log_hh
#include "Log.hh"
#endif

#ifndef Syslog_hh
#include "Syslog.hh"
#endif

#ifndef Callback_hh
#include "Callback.hh"
#endif

#ifndef Configuration_hh
#include "Configuration.hh"
#endif

#ifndef History_hh
#include "History.hh"
#endif

#include <cctype>
#include <algorithm>
#include <unistd.h>

//======================================================================

Package::Package( Database& aDatabase, unsigned long aId )
     : mDatabase( aDatabase ), mId( aId ), mProgressIndicatorStack( new ProgressIndicatorStack() )
{
} // Package::Package

//======================================================================

Package::~Package()
{
} // Package::~Package

//======================================================================

string Package::name() const
{
     return database().select_string( Database::tiPackages, id(), "Name" );

} // Package::name

//======================================================================

void Package::name( const string& aName )
{
     if ( Configuration::instance()->no_action() ) {
          LOG( NO_ACTION, "Set name of package " << aName << " in the database" );
     }
     else {
          database().update( Database::tiPackages, id(), "Name", aName );
     }

} // Package::name

//======================================================================

time_t Package::installation_time() const
{
     return database().select_time( Database::tiPackages, id(), "Installation_Time" );

} // Package::installation_time

//======================================================================

void Package::installation_time( time_t aInstallation_time )
{
     if ( Configuration::instance()->no_action() ) {
          LOG( NO_ACTION, "Set installation time for package " << name() << " in the database" );
     }
     else {
          database().update( Database::tiPackages, id(), "Installation_Time", aInstallation_time );
     }

} // Package::installation_time

//======================================================================

Pointer<FeatureList> Package::features( const string& aType ) const
{
     Pointer<DatabaseEngine::RecordList> records = database().table( Database::tiFeatures ).select( "Package_Id", DatabaseEngine::FieldValue( id() ), "Type", DatabaseEngine::FieldValue( aType ) );
     return Database::records2features( records->begin(), records->end() );

} // Package::features

//======================================================================

void Package::features( const FeatureList& aFeatures, const string& aType )
{
     if ( Configuration::instance()->no_action() ) {
          LOG( NO_ACTION, "Set features of type " << aType << " for package " << name() << " in the database" );
     }
     else {
            // remove all the old entries of features the package provides
          database().table( Database::tiFeatures ).remove( "Package_Id", DatabaseEngine::FieldValue( id() ), "Type", DatabaseEngine::FieldValue( aType ) );

          for ( FeatureList::const_iterator feature = aFeatures.begin(); feature != aFeatures.end(); ++feature ) {
               Pointer<DatabaseEngine::Record> record = database().table( Database::tiFeatures ).newRecord();
               record->id( "Package_Id", id() );
               record->str( "Type", aType );
               record->str( "Name", (*feature)->name() );
               record->str( "Version", (*feature)->version() );
               record->integer( "Serial", (*feature)->serial() );
               record->time( "Snapshot", (*feature)->snapshot() );
               database().table( Database::tiFeatures ).insert( *record );
          }
     }
} // Package::features

//======================================================================

Pointer<FeatureList> Package::features() const
{
     return features( "provide" );

} // Package::features

//======================================================================

void Package::features( const FeatureList& aFeatures )
{
       // Check if either of these features are already provided by
       // the system. The same feature cannot be provided by different
       // packages.
     for ( FeatureList::const_iterator feature = aFeatures.begin(); feature != aFeatures.end(); ++feature ) {
          Pointer<FeatureList> provided_features = database().feature( (*feature)->name(), "provide" );
          const FeatureList::const_iterator duplicate = find( provided_features->begin(), provided_features->end(), *feature );
          if ( duplicate != provided_features->end() ) // if duplicates were present
               throw runtime_error( string( "Feature " ) + (*duplicate)->name() + " " + (*duplicate)->version() + " is already provided by the system" );
     }
     features( aFeatures, "provide" );

} // Package::features

//======================================================================

Pointer<FeatureList> Package::requires() const
{
     return features( "require" );

} // Package::requires

//======================================================================

void Package::requires( const FeatureList& aFeatures )
{
     features( aFeatures, "require" );

} // Package::requires

//======================================================================

Pointer<FeatureList> Package::conflicts() const
{
     return features( "conflict" );

} // Package::conflicts

//======================================================================

void Package::conflicts( const FeatureList& aFeatures )
{
     features( aFeatures, "conflict" );

} // Package::conflicts

//======================================================================

bool Package::could_be_uninstalled( const FeatureList& aFeatures ) const
{
     Pointer<TclLanguageInterface::SlaveInterpreter> slave = tclLanguageInterface->make_slave_interpreter( string( "could_be_uninstalled_" ) + name_space(), true );
       /*const char* eval_result =*/ slave->eval( proc( "could_be_uninstalled" ) );
     Call_could_be_uninstalled call( new TclLanguageInterface::CallbackDataSlaveInterpreter( slave, "could_be_uninstalled" ), &database(), const_cast<Package*>( this ), aFeatures );
     call.activate();
     return call.result();

} // Package::could_be_uninstalled

//======================================================================

void Package::could_be_uninstalled_proc( const string& aProcCode )
{
     proc( aProcCode, "could_be_uninstalled" );

} // Package::could_be_uninstalled_proc

//======================================================================

bool Package::could_be_installed( const Feature& aFeature ) const
{
     Pointer<TclLanguageInterface::SlaveInterpreter> slave = tclLanguageInterface->make_slave_interpreter( string( "could_be_installed_" ) + name_space(), true );
       /*const char* eval_result =*/ slave->eval( proc( "could_be_installed" ) );
     Call_could_be_installed call( new TclLanguageInterface::CallbackDataSlaveInterpreter( slave, "could_be_installed" ), &database(), const_cast<Package*>( this ), &const_cast<Feature&>( aFeature ) );
     call.activate();
     return call.result();

} // Package::could_be_installed

//======================================================================

bool Package::could_be_installed( Pointer<Feature> aFeature ) const
{
     return could_be_installed( *aFeature );

} // Package::could_be_installed

//======================================================================

void Package::could_be_installed_proc( const string& aProcCode )
{
     proc( aProcCode, "could_be_installed" );

} // Package::could_be_installed_proc

//======================================================================

void Package::uninstall_proc( const string& aProcCode )
{
     proc( aProcCode, "uninstall" );

} // Package::uninstall_proc

//======================================================================

static inline Pointer<PackageStreamReference> record_to_PackageStreamReference( Pointer<DatabaseEngine::Record> record )
{
     return new PackageStreamReference( record->str( "Reference" ) );
}

Pointer<PackageDescription> Package::description() const
{
     Pointer<DatabaseEngine::RecordList> records = database().table( Database::tiDescription ).select( "Package_Id", DatabaseEngine::FieldValue( id() ) );
     Pointer<PackageDescription> description;
     if ( records->size() > 0 ) {
          Pointer<DatabaseEngine::RecordList> records_ref = database().table( Database::tiReferences ).select( "Package_Id", DatabaseEngine::FieldValue( id() ) );
          Pointer<PackageStreamReferenceList> references = new PackageStreamReferenceList;
          transform( records_ref->begin(), records_ref->end(), back_inserter( *references ), &record_to_PackageStreamReference );
          description = new PackageDescription( (*records->begin())->str( "Summary" ), (*records->begin())->str( "Description" ), (*records->begin())->str( "Author" ), (*records->begin())->str( "Canonical_Name" ), references );
     }

     return description;

} // Package::description

//======================================================================

void Package::description( Pointer<PackageDescription> aDescription ) const
{
     description( *aDescription );

} // Package::description

//======================================================================

void Package::description( const PackageDescription& aDescription ) const
{
     if ( Configuration::instance()->no_action() ) {
          LOG( NO_ACTION, "Insert Description record in the database for package " << name() );
     }
     else {
            // remove all the old entries of description and references the package provides
          database().table( Database::tiDescription ).remove( "Package_Id", DatabaseEngine::FieldValue( id() ) );
          database().table( Database::tiReferences ).remove( "Package_Id", DatabaseEngine::FieldValue( id() ) );

          Pointer<DatabaseEngine::Record> record_desc = database().table( Database::tiDescription ).newRecord();
          record_desc->id( "Package_Id", id() );
          record_desc->str( "Summary", aDescription.summary() );
          record_desc->str( "Description", aDescription.description() );
          record_desc->str( "Author", aDescription.author() );
          record_desc->str( "Canonical_Name", aDescription.canonical_name() );
          database().table( Database::tiDescription ).insert( *record_desc );

          if ( aDescription.references() ) {
               for ( PackageStreamReferenceList::const_iterator ref = aDescription.references()->begin(); ref != aDescription.references()->end(); ++ref ) {
                    Pointer<DatabaseEngine::Record> record_ref = database().table( Database::tiReferences ).newRecord();
                    record_ref->id( "Package_Id", id() );
                    record_ref->str( "Reference", (*ref)->url() );
                    database().table( Database::tiReferences ).insert( *record_ref );
               }
          }
     }

} // Package::description

//======================================================================

Pointer<PackageFileList> Package::files() const
{
     Pointer<DatabaseEngine::RecordList> records = database().table( Database::tiFiles ).select( "Package_Id", DatabaseEngine::FieldValue( id() ) );
     Pointer<PackageFileList> files = new PackageFileList;
     transform( records->begin(), records->end(), back_inserter( *files ), Record_to_PackageFile( database() ) );
     return files;

} // Package::files

//======================================================================

size_t Package::number_of_files() const
{
     Pointer<DatabaseEngine::RecordList> records = database().table( Database::tiFiles ).select( "Package_Id", DatabaseEngine::FieldValue( id() ) );
     return records->size();

} // Package::number_of_files

//======================================================================

void Package::disown( const string& aPathname )
{
     Pointer<DatabaseEngine::RecordList> records_fileinfo = database().table( Database::tiFileInfo ).select( "Pathname", DatabaseEngine::FieldValue( aPathname ) );
     if ( records_fileinfo->empty() ) {
          LOG( WARNING, "File " << aPathname << " is not owned by any package" );
          return;
     }

       // Search among returned files for file with the same pathname
       // as requested. This is done to overcome querying corrupted
       // database which may have pathname index entries pointed to
       // nowhere.
     DatabaseEngine::RecordList::const_iterator rec = records_fileinfo->begin();
     for ( ; rec != records_fileinfo->end() && (*rec)->str( "Pathname" ) != aPathname; ++rec );
     if ( rec == records_fileinfo->end() ) { // All data returned by pathname search is invalid.
          LOG( WARNING, "File " << aPathname << " is not owned by any package" );
          return;
     }
     const DatabaseEngine::Id fileinfo_id = (*rec)->id();
     Pointer<DatabaseEngine::RecordList> records = database().table( Database::tiFiles ).select( "Package_Id", DatabaseEngine::FieldValue( id() ), "FileInfo_Id", DatabaseEngine::FieldValue( fileinfo_id ) );
     if ( records->empty() ) {
          LOG( WARNING, "File " << aPathname << " is not owned by package " << name() );
          return;
     }

     if ( Configuration::instance()->no_action() ) {
          LOG( NO_ACTION, "Remove Files entry from the database for file: " << aPathname );
     }
     else {
          database().table( Database::tiFiles ).remove( (*records)[0]->id() );

            // Check if FileInfo entry is referenced by other File or
            // FilesUndo entries and if not referenced remove FileInfo
            // entry
          Pointer<DatabaseEngine::RecordList> records_other = database().table( Database::tiFiles ).select( "FileInfo_Id", DatabaseEngine::FieldValue( fileinfo_id ), "FileInfo_Id" );
          Pointer<DatabaseEngine::RecordList> records_undo = database().table( Database::tiFilesUndo ).select( "FileInfo_Id", DatabaseEngine::FieldValue( fileinfo_id ), "FileInfo_Id" );
          if ( records_other->empty() && records_undo->empty() ) {
               database().table( Database::tiFileInfo ).remove( (*records->begin())->id( "FileInfo_Id" ) );
          }
     }

} // Package::disown

//======================================================================

void Package::own( const string& aPathname )
{
     Pointer<DatabaseEngine::RecordList> records_fileinfo = database().table( Database::tiFileInfo ).select( "Pathname", DatabaseEngine::FieldValue( aPathname ) );

       // Search for entry for this pathname skipping corrupted entries
     DatabaseEngine::RecordList::iterator record_fileinfo = records_fileinfo->begin();
     for ( ; record_fileinfo != records_fileinfo->end() && (*record_fileinfo)->str( "Pathname" ) != aPathname; ++record_fileinfo );

     DatabaseEngine::Id fileinfo_id;
     if ( record_fileinfo != records_fileinfo->end() ) {
            // Entry for this file already exists
          DatabaseEngine::Record& record_f = **record_fileinfo;
          fileinfo_id = record_f.id();

            // Check if file is already owned by the package
          if ( database().table( Database::tiFiles ).select( "Package_Id", DatabaseEngine::FieldValue( id() ), "FileInfo_Id", DatabaseEngine::FieldValue( fileinfo_id ) )->size() > 0 )
               throw runtime_error( string( "File " ) + aPathname + " is already owned by package " + name() );

            // Update entry for this file (it could be patched etc.)
          RealFile aFile( aPathname );
          record_f.longlong( "Length", aFile.length() );
          record_f.str( "MD5", aFile.md5() );
          record_f.time( "MTime", aFile.mtime() );
          record_f.str( "User", aFile.user() );
          record_f.str( "Group", aFile.group() );
          record_f.integer( "Permissions", aFile.mode() );
          record_f.str( "Symlink", aFile.symlink() );
          record_f.longlong( "Device", aFile.dev() );
          database().table( Database::tiFileInfo ).update( record_f );
     }
     else {
          if ( Configuration::instance()->no_action() ) {
               LOG( NO_ACTION, "Create FileInfo entry in the database for file: " << aPathname );
          }
          else {
                 // Create new FileInfo entry
               RealFile aFile( aPathname );
               Pointer<DatabaseEngine::Record> record_fileinfo = database().table( Database::tiFileInfo ).newRecord();
               record_fileinfo->str( "Pathname", aFile.pathname() );
               record_fileinfo->str( "Filesystem", aFile.filesystem() );
               record_fileinfo->str( "Inode", aFile.inode() );
               record_fileinfo->longlong( "Length", aFile.length() );
               record_fileinfo->str( "MD5", aFile.md5() );
               record_fileinfo->time( "MTime", aFile.mtime() );
               record_fileinfo->str( "User", aFile.user() );
               record_fileinfo->str( "Group", aFile.group() );
               record_fileinfo->integer( "Permissions", aFile.mode() );
               record_fileinfo->str( "Symlink", aFile.symlink() );
               record_fileinfo->longlong( "Device", aFile.dev() );
               database().table( Database::tiFileInfo ).insert( *record_fileinfo );
               fileinfo_id = record_fileinfo->id();
          }
     }

     if ( Configuration::instance()->no_action() ) {
          LOG( NO_ACTION, "Create Files entry in the database for file: " << aPathname );
     }
     else {
            // Insert reference into the Files table
          Pointer<DatabaseEngine::Record> record_file = database().table( Database::tiFiles ).newRecord();
          record_file->id( "Package_Id", id() );
          record_file->id( "FileInfo_Id", fileinfo_id );
          database().table( Database::tiFiles ).insert( *record_file );
     }

} // Package::own

//======================================================================

Pointer<PackageFile> Package::owns( const string& aPathname ) const
{
     DatabaseEngine::Id fileinfo_id;
     Pointer<FileInfo> basefile = file( aPathname, fileinfo_id );
     Pointer<PackageFile> file;
     if ( basefile ) {
          Pointer<DatabaseEngine::RecordList> records = database().table( Database::tiFiles ).select( "Package_Id", DatabaseEngine::FieldValue( id() ), "FileInfo_Id", DatabaseEngine::FieldValue( fileinfo_id ) );
          if ( records->size() > 0 ) // owned by this package
               file = new PackageFile( *basefile );
     }

     return file;

} // Package::owns

//======================================================================

bool Package::has_undo_info()
{
     return database().table( Database::tiUndo ).select( "Package_Id", DatabaseEngine::FieldValue( id() ) )->size() > 0;

} // Package::has_undo_info

//======================================================================

void Package::undo()
{
     const string package_name = name();

     if ( !has_undo_info() )
          throw runtime_error( string( "Package " ) + package_name + " has no undo info saved" );

     Syslog syslog;
     syslog.message( string( "Package " ) + package_name + " undone" );

     LOG( INFO, "Package " << package_name << " undone" );

} // Package::undo

//======================================================================

bool Package::has_uninstall_proc()
{
     return proc( "uninstall" ).length() > 0;

} // Package::has_uninstall_proc

//======================================================================

void Package::uninstall( bool no_log )
{
     const string package_name = name();
     if ( package_name == "SYSTEM" )
          throw runtime_error( string( "Package " ) + package_name + " cannot be uninstalled" );

     ProgressIndicatorPusher progressIndicatorPusher( mProgressIndicatorStack, string( "Uninstall " ) + package_name, ProgressIndicator::ProgressInfo( number_of_files(), 0, 0, 4 ) );

     const string ns = string( "uninstall_" ) + name_space();
     Pointer<TclLanguageInterface::SlaveInterpreter> slave = tclLanguageInterface->make_slave_interpreter( ns, true );

     progress().progress( ProgressIndicator::itAdditional );

     try {
          const string proc_uninstall = proc( "uninstall" );
          if ( proc_uninstall.length() == 0 )
               throw runtime_error( string( "Uninstallation of " ) + package_name + " failed: no proc uninstall recorded in the database for the package" );
          slave->eval( proc_uninstall ); // source proc uninstall into slave interpreter
          progress().progress( ProgressIndicator::itAdditional );

          Call_uninstall call( new TclLanguageInterface::CallbackDataSlaveInterpreter( slave, "uninstall" ), &database(), this );
          call.activate();
          if ( !call.result().empty() )
               throw runtime_error( string( "Uninstallation of " ) + package_name + " failed: " + call.result() );
     }
     catch ( runtime_error& err ) {
          throw runtime_error( string( "Cannot uninstall " ) + package_name + ": " + err.what() );
     }

     progress().progress( ProgressIndicator::itAdditional );

     progress().progress( ProgressIndicator::itAdditional );
     progress().progress( 1.0 );

     database().updateHistory( History( package_name, "uninstalled" ) );
     if ( !no_log ) {
          Syslog syslog;
          syslog.message( string( "Package " ) + package_name + " uninstalled" );

          LOG( INFO, "Package " << package_name << " uninstalled" );
     }

} // Package::uninstall

//======================================================================

void Package::uninstall_file( Pointer<PackageFile> aFile )
{
     if ( aFile->pathname().empty() || aFile->pathname()[0] != '/' )
          throw runtime_error( string( "Internal error, invalid pathname to uninstall: " ) + aFile->pathname() );

     const bool exists = File::exist( aFile->pathname() );
     if ( !exists ) {
          const string result = Callback::owned_file_does_not_exist_for_uninstall( this, aFile );
          if ( result.length() > 0 )
               throw runtime_error( result );
     }

     directory_write_access( aFile );

     disown( aFile->pathname() );
     bool erase = database().packages_owning( aFile->pathname() )->size() == 0;
     if ( erase && exists ) {
            // Check MD5
          RealFile rf( aFile->pathname() );
          if ( aFile->md5().length() > 0 && rf.md5() != aFile->md5() ) {
                 // ????

                 // If MD5 of the file is recorded in the database and
                 // differs from MD5 of the real file user is asked to
                 // choose: view the file and then return to this
                 // menu, remove the file, leave the file and probably
                 // have problem of unreferenced file later, move the
                 // file to another location (preferably into
                 // directory which is not handled by the package
                 // system).
          }
     }

     if ( erase ) {
          if ( Configuration::instance()->no_action() ) {
               LOG( NO_ACTION, "Unlink " << aFile->pathname() );
          }
          else {
               LOG( INFO, "About to unlink " << aFile->pathname() );
               ::unlink( aFile->pathname().c_str() );
          }
     }

     progress().progress( ProgressIndicator::itFile );

} // Package::uninstall_file

//======================================================================

Pointer<PackageFile::ConsistencyList> Package::consistency_check() const
{
     Pointer<PackageFile::ConsistencyList> result = new PackageFile::ConsistencyList;
     Pointer<PackageFileList> package_files = files();
     for ( PackageFileList::const_iterator file = package_files->begin(); file != package_files->end(); ++file ) {
          Pointer<PackageFile::Consistency> consistency = (*file)->check();
          if ( consistency->differs() )
               result->push_back( consistency );
     }
     return result;

} // Package::consistency_check

//======================================================================

void Package::proc( const string& aProcCode, const string& aType )
{
     if ( Configuration::instance()->no_action() ) {
          LOG( NO_ACTION, "Insert proc " << aType << " for package " << name() << " into the database" );
     }
     else {
            // remove all the old entries of proc of aType the package provides
          database().table( Database::tiScripts ).remove( "Package_Id", DatabaseEngine::FieldValue( id() ), "Type", DatabaseEngine::FieldValue( aType ) );

          if ( aProcCode.length() > 0 ) {
               Pointer<DatabaseEngine::Record> record = database().table( Database::tiScripts ).newRecord();
               record->id( "Package_Id", id() );
               record->str( "Type", aType );
               record->str( "Data", aProcCode );
               database().table( Database::tiScripts ).insert( *record );
          }
     }

} // Package::proc

//======================================================================

string Package::proc( const string& aType ) const
{
     Pointer<DatabaseEngine::RecordList> records = database().table( Database::tiScripts ).select( "Package_Id", DatabaseEngine::FieldValue( id() ), "Type", DatabaseEngine::FieldValue( aType ) );
     return records->size() > 0 ? (*records->begin())->str( "Data" ) : string();

} // Package::proc

//======================================================================

Pointer<FileInfo> Package::file( const string& aPathname, DatabaseEngine::Id& fileinfo_id ) const
{
     Pointer<DatabaseEngine::RecordList> records_fileinfo = database().table( Database::tiFileInfo ).select( "Pathname", DatabaseEngine::FieldValue( aPathname ) );
     Pointer<FileInfo> file;
     if ( records_fileinfo->size() > 0 ) { // owned by any package
          const DatabaseEngine::Record& record_f = **records_fileinfo->begin();
          fileinfo_id = record_f.id();
          string symlink = record_f.null( "Symlink" ) ? string() : record_f.str( "Symlink" );
          unsigned long long device = record_f.null( "Device" ) ? 0 : record_f.longlong( "Device" );
          file = new FileInfo( record_f.str( "Pathname" ), record_f.longlong( "Length" ), record_f.time( "MTime" ), record_f.str( "User" ), record_f.str( "Group" ), record_f.integer( "Permissions" ), symlink, device );
          file->md5( record_f.str( "MD5" ) );
     }

     return file;

} // Package::file

//======================================================================

void Package::remove()
{
     if ( name() == "SYSTEM" )
          throw runtime_error( string( "Cannot remove SYSTEM package" ) );

     if ( Configuration::instance()->no_action() ) {
          LOG( NO_ACTION, "Remove all records corresponding to package " << name() << " from the database" );
     }
     else {
            // Remove information about package dependant features
          features( FeatureList() );
          requires( FeatureList() );
          conflicts( FeatureList() );

            // Remove package description and references
          database().table( Database::tiDescription ).remove( "Package_Id", DatabaseEngine::FieldValue( id() ) );
          database().table( Database::tiReferences ).remove( "Package_Id", DatabaseEngine::FieldValue( id() ) );

            // Disown files
          Pointer<PackageFileList> fs = files();
          for ( PackageFileList::const_iterator f = fs->begin(); f != fs->end(); ++f )
               disown( (*f)->pathname() );

            // Remove undo and uninstall info
          database().table( Database::tiUndo ).remove( "Package_Id", DatabaseEngine::FieldValue( id() ) );
          database().table( Database::tiFilesUndo ).remove( "Package_Id", DatabaseEngine::FieldValue( id() ) );
            // Remove information about scripts
          database().table( Database::tiScripts ).remove( "Package_Id", DatabaseEngine::FieldValue( id() ) );

            // Remove Packages table entry
          database().table( Database::tiPackages ).remove( id() );
     }

} // Package::remove

//======================================================================

bool Package::exists()
{
     return ! database().table( Database::tiPackages ).select( id() )->null( "Name" );

} // Package::exists

//======================================================================

static inline char replace_no_alnum( char src )
{
     return isalnum( src ) ? src : '_';
}

string Package::name_space() const
{
     string ns = name();
     transform( ns.begin(), ns.end(), ns.begin(), &replace_no_alnum );
     return ns;

} // Package::name_space

//======================================================================

void Package::directory_write_access( Pointer<PackageFile> aFile )
{
     Pointer<Container<list<string> > > result = Callback::directory_write_access( &mDatabase, 0, File::dirname( aFile->pathname() ) );
     string err;
     if ( result && result->size() > 0 )
          err = *result->rbegin();
     if ( err.length() > 0 )
          throw runtime_error( err );

} // Package::directory_write_access

//======================================================================

Package::Call_could_be_uninstalled::Call_could_be_uninstalled( Pointer<TclLanguageInterface::CallbackDataSlaveInterpreter> aCallbackData, Pointer<Database> aDatabase, Pointer<Package> aPackage, const FeatureList& aFeatures )
     : Call( aCallbackData ), mDatabase( aDatabase ), mPackage( aPackage ), mFeatures( aFeatures )
{
}

//======================================================================

Pointer<Database> Package::Call_could_be_uninstalled::database()
{
     return mDatabase;

} // Package::Call_could_be_uninstalled::database

//======================================================================

Package::Call_could_be_installed::Call_could_be_installed( Pointer<TclLanguageInterface::CallbackDataSlaveInterpreter> aCallbackData, Pointer<Database> aDatabase, Pointer<Package> aPackage, Pointer<Feature> aFeature )
     : Call( aCallbackData ), mDatabase( aDatabase ), mPackage( aPackage ), mFeature( aFeature )
{
}

//======================================================================

Pointer<Database> Package::Call_could_be_installed::database()
{
     return mDatabase;

} // Package::Call_could_be_installed::database

//======================================================================

Package::Call_uninstall::Call_uninstall( Pointer<TclLanguageInterface::CallbackDataSlaveInterpreter> aCallbackData, Pointer<Database> aDatabase, Pointer<Package> aPackage )
     : Call( aCallbackData ), mDatabase( aDatabase ), mPackage( aPackage )
{
} // Package::Call_uninstall::Call_uninstall

//======================================================================

Pointer<Database> Package::Call_uninstall::database()
{
     return mDatabase;

} // Package::Call_uninstall::database

//======================================================================
