/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/sysinstall/Package.hh,v 1.2 2001/06/22 15:05:16 alex Exp $
 */

// -*- C++ -*-  Thu Aug 13 1998
// $Id: Package.hh,v 1.10 1999/01/20 10:35:22 eugene Exp $

#ifndef Package_hh
#define Package_hh

#ifndef PackageFile_hh
#include "PackageFile.hh"
#endif

#ifndef Feature_hh
#include "Feature.hh"
#endif

#ifndef ProgressIndicator_hh
#include "ProgressIndicator.hh"
#endif

#ifndef TclLanguageInterface_hh
#include "TclLanguageInterface.hh"
#endif

//======================================================================

class PackageDescription;
class FileInfo;
class Database;

//======================================================================

class Package : virtual public LanguageInterface::Object
{
     CLASS_METHODS_DESCRIPTIONS_DECLARATION;

  public:
     Package( Database& aDatabase, unsigned long aId );
     virtual ~Package();

     virtual string name() const;
     virtual void name( const string& aName );

     virtual time_t installation_time() const;
     virtual void installation_time( time_t aInstallation_time );

     virtual Pointer<FeatureList> features() const;
     virtual void features( const FeatureList& aFeatures );

     virtual Pointer<FeatureList> requires() const;
     virtual void requires( const FeatureList& aFeatures );

     virtual Pointer<FeatureList> conflicts() const;
     virtual void conflicts( const FeatureList& aFeatures );

     virtual bool could_be_uninstalled( const FeatureList& aFeatures ) const;
     virtual void could_be_uninstalled_proc( const string& aProcCode );

     virtual bool could_be_installed( const Feature& aFeature ) const;
     virtual bool could_be_installed( Pointer<Feature> aFeature ) const;
     virtual void could_be_installed_proc( const string& aProcCode );

     virtual Pointer<PackageDescription> description() const;
     virtual void description( const PackageDescription& aDescription ) const;
     virtual void description( Pointer<PackageDescription> aDescription ) const;

     virtual size_t number_of_files() const;
     virtual Pointer<PackageFileList> files() const;
     virtual void disown( const string& aPathname );
     virtual void own( const string& aPathname );
     virtual Pointer<PackageFile> owns( const string& aPathname ) const;

       // Returns if undo info is saved for the package.
     virtual bool has_undo_info();
       // Undoes package installation (throws an exception if undo info
       // was not saved.
     virtual void undo();

     virtual void uninstall_proc( const string& aProcCode );
       // Returns if proc uninstall is saved for the package.
     virtual bool has_uninstall_proc();
     virtual void uninstall( bool no_log = false );
     virtual void uninstall_file( Pointer<PackageFile> aFile );

     virtual Pointer<PackageFile::ConsistencyList> consistency_check() const;

       // Removes from the database all information regarding
       // package. If Configuration::no_action does not do anything
       // just reports what is to be done.
     virtual void remove();

       // Returns if this package records are still exist in the database.
     virtual bool exists();

     bool operator == ( const Package& another ) const { return id() == another.id(); }

     unsigned long id() const { return mId; }

     Pointer<ProgressIndicatorStack> progress_indicator_stack() const { return mProgressIndicatorStack; }
     void progress_indicator_stack( Pointer<ProgressIndicatorStack> aProgressIndicatorStack ) { mProgressIndicatorStack = aProgressIndicatorStack; }

     static inline bool lessById( Pointer<Package> a, Pointer<Package> b ) { return a->id() < b->id(); }

  protected:
     Database& database() const { return mDatabase; }
     ProgressIndicatorStack& progress() { return *mProgressIndicatorStack; }

  private:
     mutable Database& mDatabase;
     unsigned long mId;
     Pointer<ProgressIndicatorStack> mProgressIndicatorStack;

     virtual Pointer<FeatureList> features( const string& aType ) const;
     virtual void features( const FeatureList& aFeatures, const string& aType );

     virtual void proc( const string& aProcCode, const string& aType );
     virtual string proc( const string& aType ) const;

     virtual Pointer<FileInfo> file( const string& aPathname, unsigned long& fileinfo_id ) const;
     virtual string name_space() const;

     virtual void directory_write_access( Pointer<PackageFile> aFile );

  public:
     class Call : public LanguageInterface::SpecificCallbackData
     {
       public:
          Call( Pointer<TclLanguageInterface::CallbackDataSlaveInterpreter> aCallbackData ) : LanguageInterface::SpecificCallbackData( aCallbackData ) {}
          virtual const char* typeid_name() const { return typeid( Package ).name(); }
          virtual void activate() { if ( callbackData() ) callbackData()->activate( *this ); }

     }; // class Call

     class Call_could_be_uninstalled : public Call
     {
       public:
          Call_could_be_uninstalled( Pointer<TclLanguageInterface::CallbackDataSlaveInterpreter> aCallbackData, Pointer<Database> aDatabase, Pointer<Package> aPackage, const FeatureList& aFeatures );

          Pointer<Database> database();
          Pointer<Package> package() { return mPackage; }
          const FeatureList& features() { return mFeatures; }
          bool& result() { return mResult; }
          bool result() const { return mResult; }

       private:
          bool mResult;
          Pointer<Database> mDatabase;
          Pointer<Package> mPackage;
          const FeatureList& mFeatures;

     }; // class Call_could_be_uninstalled

     class Call_could_be_installed : public Call
     {
       public:
          Call_could_be_installed( Pointer<TclLanguageInterface::CallbackDataSlaveInterpreter> aCallbackData, Pointer<Database> aDatabase, Pointer<Package> aPackage, Pointer<Feature> aFeature );

          Pointer<Database> database();
          Pointer<Package> package() { return mPackage; }
          Pointer<Feature> feature() { return mFeature; }
          bool& result() { return mResult; }
          bool result() const { return mResult; }

       private:
          bool mResult;
          Pointer<Database> mDatabase;
          Pointer<Package> mPackage;
          Pointer<Feature> mFeature;

     }; // class Call_could_be_installed

     class Call_uninstall : public Call
     {
       public:
          Call_uninstall( Pointer<TclLanguageInterface::CallbackDataSlaveInterpreter> aCallbackData, Pointer<Database> aDatabase, Pointer<Package> aPackage );

          Pointer<Database> database();
          Pointer<Package> package() { return mPackage; }
          string& result() { return mResult; }
          const string& result() const { return mResult; }

       private:
          string mResult;
          Pointer<Database> mDatabase;
          Pointer<Package> mPackage;

     }; // class Call_uninstall

}; // class Package

//======================================================================

typedef Container<list<Pointer<Package> > > PackageList;

//======================================================================

#endif // Package_hh
