/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/sysinstall/PackageCreator.cc,v 1.4 2001/07/17 18:29:20 alex Exp $
 */

// -*- C++ -*-   Mon Aug  3 1998
// $Id: PackageCreator.cc,v 1.15 1999/01/20 10:35:22 eugene Exp $

#ifndef PackageCreator_hh
#include "PackageCreator.hh"
#endif

#ifndef PackageStream_hh
#include "PackageStream.hh"
#endif

#ifndef PackageStreamStructure_hh
#include "PackageStreamStructure.hh"
#endif

#ifndef PackageDescription_hh
#include "PackageDescription.hh"
#endif

#ifndef Feature_hh
#include "Feature.hh"
#endif

#ifndef PatchSuffixes_hh
#include "PatchSuffixes.hh"
#endif

#ifndef Archive_hh
#include "Archive.hh"
#endif

#ifndef Log_hh
#include "Log.hh"
#endif

#ifndef RealFile_hh
#include "RealFile.hh"
#endif

#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/param.h>

//======================================================================

const char* PackageCreator::SYSINSTALL_DIR = "SYSINSTALL";

const char* description_procs[] = { 0 };
const char* description_vars[] = { "description_name_internal", "description_summary_internal", "description_description_internal", "description_author_internal", 0 };
const char* features_procs[] = { "provides", "requires", "check_requirements", "conflicts", "check_conflicts", "could_be_installed_proc", "could_be_uninstalled_proc", "check_users_requirements", "check_groups_requirements", "check_suggestions", 0 };
const char* features_vars[] = { 0 };
const char* structure_procs[] = { "structure", 0 };
const char* structure_vars[] = { 0 };
const char* install_procs[] = { "install", "upgrade", "uninstall_proc", 0 };
const char* install_vars[] = { 0 };

const PackageCreator::SysinstallScript PackageCreator::sScripts[] =
{
     PackageCreator::SysinstallScript( "description.tcl", description_procs, description_vars, PackageStream::description_helper_procs, &generate_description, &check_description ),
     PackageCreator::SysinstallScript( "features.tcl", features_procs, features_vars, PackageStream::features_helper_procs, &generate_features, &check_features ),
     PackageCreator::SysinstallScript( "structure.tcl", structure_procs, structure_vars, PackageStream::structure_helper_procs, &generate_structure, &check_structure ),
     PackageCreator::SysinstallScript( "install.tcl", install_procs, install_vars, PackageStream::install_helper_procs, &generate_install, &check_install ),
     PackageCreator::SysinstallScript( 0, 0, 0, 0, 0, 0 )
};

//======================================================================

PackageCreator::PackageCreator( const string& aPackageName, const string& aPackageDir, const string& aPackageFilesDir, Pointer<Container<list<string> > > aIgnore )
     : mPackageName( aPackageName ), mPackageDir( aPackageDir ),
       mPackageFilesDir( aPackageFilesDir ), mIgnore( aIgnore )
{
     if ( !mIgnore )
          mIgnore = new Container<list<string> >;
     mIgnore->push_back( SYSINSTALL_DIR );

} // PackageCreator::PackageCreator

//======================================================================

void PackageCreator::generate()
{
     mSlaveInterp = tclLanguageInterface->make_slave_interpreter( "slave_PackageCreator", false );
     scan_files( mPackageFilesDir, "" );
     scan_scripts();

} // PackageCreator::generate

//======================================================================

void PackageCreator::scan_files( const string& name, const string& basename )
{
     Pointer<Container<list<Pointer<RealFile> > > > files = File::scan( name, true, false, mIgnore );
     for ( list<Pointer<RealFile> >::const_iterator file = files->begin(); file != files->end(); ++file )
          add_file( *file, name, basename );

} // PackageCreator::scan

//======================================================================

void PackageCreator::scan_scripts()
{
     const string scriptsDir = mPackageFilesDir + "/" + SYSINSTALL_DIR;
     LOG( INFO, "Scanning scripts in " << scriptsDir );

     const string namespaceForProcs = string( "::" ) + name() + "::" + SYSINSTALL_DIR;

     Pointer<SysinstallRealScriptList> scripts;
     scripts = sourceExistingScripts( namespaceForProcs, scriptsDir );
     generateAbsentProcs( namespaceForProcs, *scripts );
     checkProcs( namespaceForProcs );
     writeAbsentProcs( scriptsDir, *scripts );

} // PackageCreator::scan_scripts

//======================================================================

void PackageCreator::write_package_file()
{
     string prefix;
     if ( packagedir()[0] == '/' )
          prefix = "";
     else {
          char cwdbuf[MAXPATHLEN] = "";
          prefix = string( getcwd( cwdbuf, MAXPATHLEN ) ) + "/";
     }
     const string zipfile = prefix + packagedir() + "/" + name() + ".pkg.zip";

     LOG( INFO, "Writing package into " << zipfile );

     Archive stream( zipfile, true );

     for ( const SysinstallScript* script = sScripts; script->filename; ++script )
          stream.add( mPackageFilesDir + "/" + SYSINSTALL_DIR + "/" + script->filename, string( SYSINSTALL_DIR ) + "/" + script->filename );

     for ( PackageStreamFileList::const_iterator file = files().begin(); file != files().end(); ++file ) {
          switch ( (*file)->attachment() ) {
            case PackageStreamFile::aPatch:
            {
                 string pathname_to_record = (*file)->pathname().c_str();
                 for ( const char* const* suffix = PatchSuffixes; *suffix; ++suffix ) {
                      if ( pathname_to_record.length() > strlen( *suffix ) && !strcmp( pathname_to_record.c_str() + pathname_to_record.length() - strlen( *suffix ), *suffix ) ) {
                           pathname_to_record = pathname_to_record.substr( 0, pathname_to_record.length() - strlen( *suffix ) );
                           break;
                      }
                 }
                 stream.add( mPackageFilesDir + "/" + (*file)->pathname(), pathname_to_record );
            }
                 break;
            case PackageStreamFile::aAppend:
            {
                 string pathname_to_record = (*file)->pathname().c_str();
                 for ( const char* const* suffix = AppendSuffixes; *suffix; ++suffix ) {
                      if ( pathname_to_record.length() > strlen( *suffix ) && !strcmp( pathname_to_record.c_str() + pathname_to_record.length() - strlen( *suffix ), *suffix ) ) {
                           pathname_to_record = pathname_to_record.substr( 0, pathname_to_record.length() - strlen( *suffix ) );
                           break;
                      }
                 }
                 stream.add( mPackageFilesDir + "/" + (*file)->pathname(), pathname_to_record );
            }
                 break;
            case PackageStreamFile::aFile:
                 stream.add( mPackageFilesDir + "/" + (*file)->pathname(), (*file)->pathname() );
                 break;
            case PackageStreamFile::aGenerated:
            case PackageStreamFile::aUnknown:
                 break;
     }
     }

} // PackageCreator::write_package_file

//======================================================================

void PackageCreator::add_file( Pointer<RealFile> aFile, const string& name, const string& basename )
{
     string filename = aFile->pathname();
     if ( strncmp( filename.c_str(), name.c_str(), name.length() ) )
          throw runtime_error( string( "Internal error in PackageCreator::add_file(): pathname: " ) + filename + " name: " + name + " basename: " + basename );
     filename.replace( 0, name.length(), basename );
     if ( filename[0] == '/' )
          filename.erase( 0, 1 );

     PackageStreamFile::Attachment attachment = S_ISREG( aFile->mode() ) ? PackageStreamFile::aFile : PackageStreamFile::aGenerated;
     if ( attachment == PackageStreamFile::aFile ) {
          for ( const char* const* suffix = PatchSuffixes; *suffix; ++suffix ) {
               if ( filename.length() > strlen( *suffix ) && !strcmp( filename.c_str() + filename.length() - strlen( *suffix ), *suffix ) ) {
                    attachment = PackageStreamFile::aPatch;
                    break;
               }
          }
     }
     if ( attachment == PackageStreamFile::aFile ) {
          for ( const char* const* suffix = AppendSuffixes; *suffix; ++suffix ) {
               if ( filename.length() > strlen( *suffix ) && !strcmp( filename.c_str() + name.length() - strlen( *suffix ), *suffix ) ) {
                    attachment = PackageStreamFile::aAppend;
                    break;
               }
          }
     }

     LOG( INFO, "Adding " << filename );
     mFiles.push_back( new PackageStreamFile( FileInfo( filename, *aFile ), attachment, PackageStreamFile::rDefault, string() ) );

} // PackageCreator::add_file

//======================================================================

Pointer<PackageCreator::SysinstallRealScriptList> PackageCreator::sourceExistingScripts( const string& aNamespace, const string& aDir )
{
     int script_files = 0;
     for ( const SysinstallScript* script = sScripts; script->filename; ++script ) {
          interpreter()->eval( script->helper_code, strlen( script->helper_code ), aNamespace );
          string file = aDir + "/" + script->filename;
          if ( File::exist( file ) ) {
               LOG( INFO, "Evaluating " << file );
               try {
                    interpreter()->eval( string( "source " ) + file, aNamespace );
               }
               catch ( exception& err ) {
                    throw runtime_error( string( "Error sourcing " ) + file + ": " + err.what() );
               }
          }
          ++script_files;
     }

     Pointer<SysinstallRealScriptList> scripts = new SysinstallRealScriptList;
     for ( int i = 0; i < script_files; ++i ) {
          scripts->push_back( new SysinstallRealScript( sScripts[i] ) );
     }

     return scripts;

} // PackageCreator::sourceingExistingScripts

//======================================================================

void PackageCreator::generateAbsentProcs( const string& aNamespace, SysinstallRealScriptList& scripts )
{
     for ( SysinstallRealScriptList::iterator script = scripts.begin(); script != scripts.end(); ++script ) {
          for ( const char** proc = (*script)->procs; *proc; ++proc ) {
               if ( interpreter()->command_exists( string( "::" ) + aNamespace + "::" + *proc ) ) {
                    LOG( INFO, "\t" << *proc << " present" );
               }
               else {
                    LOG( INFO, "\t" << *proc << " absent" );
                    (*script)->absentProcs.push_back( (*script)->generate( *proc, *this ) );
                    try {
                         interpreter()->eval( *(*script)->absentProcs.rbegin(), aNamespace );
                    }
                    catch ( exception& err ) {
                         throw runtime_error( string( "Evalutaing of " ) + (*script)->filename + " failed: " + err.what() );
                    }
               }
          }
          for ( const char** var = (*script)->vars; *var; ++var ) {
               if ( interpreter()->variable_exists( string( "::" ) + aNamespace + "::" + *var ) ) {
                    LOG( INFO, "\t" << *var << " present" );
               }
               else {
                    LOG( INFO, "\t" << *var << " absent" );
                    (*script)->absentProcs.push_back( (*script)->generate( *var, *this ) );
                    try {
                         interpreter()->eval( *(*script)->absentProcs.rbegin(), aNamespace );
                    }
                    catch ( exception& err ) {
                         throw runtime_error( string( "Evalutaing of " ) + (*script)->filename + " failed: " + err.what() );
                    }
               }
          }
     }

} // PackageCreator::generateAbsentProcs

//======================================================================

string PackageCreator::generate_description( const string& aName, PackageCreator& packageCreator )
{
     if ( aName == "description_name_internal" ) {
          return string( "name \"" ) + packageCreator.name() + "\"\n";
     }
     else if ( aName == "description_summary_internal" ) {
          return string( "summary \"" ) + packageCreator.name() + " summary\"\n";
     }
     else if ( aName == "description_description_internal" ) {
          return string( "description \"" ) + packageCreator.name() + " description\"\n";
     }
     else if ( aName == "description_author_internal" ) {
          return string( "author \"" ) + packageCreator.name() + " author\"\n";
     }
     else
          throw runtime_error( "generate_description: unknown proc name to generate: " + aName );

} // PackageCreator::generate_description

//======================================================================

string PackageCreator::generate_features( const string& aProcName, PackageCreator& packageCreator )
{
     if ( aProcName == "provides" ) {
          return string( "proc provides {} {\n\treturn [list [Feature \"" ) + packageCreator.name() + "\" \"1.0\"]]\n}\n\n";
     }
     else if ( aProcName == "requires" ) {
          return "proc requires {} {\n\treturn [list [Feature \"i386\" \"3.0\"]]\n}\n\n";
     }
     else if ( aProcName == "check_requirements" ) {
          return "\
proc check_requirements {database_handle} {\n\
        set result {}\n\
        foreach fr [requires] {\n\
                set apnd 1\n\
		set fdb [$database_handle feature [$fr name] provide]\n\
		foreach f $fdb {\n\
			if {[$f compare $fr] >= 0} {\n\
				set apnd 0; # same or newer feature is provided by the system\n\
				break\n\
			}\n\
                }\n\
                if {$apnd} {\n\
                	lappend result $fr\n\
		}\n\
	}\n\
	return $result\n\
}\n\n";
     }
     else if ( aProcName == "conflicts" ) {
          return "proc conflicts {} {\n\treturn {}\n}\n\n";
     }
     else if ( aProcName == "check_conflicts" ) {
          return "\
proc check_conflicts {database_handle} {\n\
        set result {}\n\
        foreach fr [conflicts] {\n\
		set fdb [$database_handle feature [$fr name] provide]\n\
		if {[llength $fdb] > 0} {\n\
                	lappend result $fr\n\
		}\n\
	}\n\
	return $result\n\
}\n\n";
     }
     else if ( aProcName == "could_be_installed_proc" ) {
          return "proc could_be_installed_proc {} {\n\treturn {\n\
	proc could_be_installed {database_handle package feature} {\n\
		return 1\n\
	}\n\
}\n}\n\n";
     }
     else if ( aProcName == "could_be_uninstalled_proc" ) {
          return "proc could_be_uninstalled_proc {} {\n\
\treturn {\n\
	proc could_be_uninstalled {database_handle package features} {\n\
	        set result 1\n\
        	foreach feature_required [$package requires] {\n\
			foreach feature_to_uninstall $features {\n\
	                        if {[string compare [$feature_to_uninstall name] [$feature_required name]] == 0 && [$feature_to_uninstall compare $feature_required] >= 0} {\n\
	                                # feature to be uninstalled is required by this package\n\
					set result 0\n\
	                                # check if system provides another suitable feature with this name\n\
					set features_in_db [$database_handle feature_other_versions $feature_to_uninstall provide]\n\
					foreach feature_in_db $features_in_db {\n\
						if {[$feature_in_db compare $feature_required] >= 0} {\n\
							# system will provide suitable feature
							set result 1\n\
							break\n\
						}\n\
					}\n\
	                        }\n\
	                        if {$result == 0} {\n\
	                                break\n\
	                        }\n\
			}\n\
                        if {$result == 0} {\n\
                                break\n\
                        }\n\
		}\n\
		return $result\n\
	}\n\
}\n}\n\n";
     }
     else if ( aProcName == "check_users_requirements" ) {
          return "proc check_users_requirements {} {\n\treturn {}\n}\n\n";
     }
     else if ( aProcName == "check_groups_requirements" ) {
          return "proc check_groups_requirements {} {\n\treturn {}\n}\n\n";
     }
     else if ( aProcName == "check_suggestions" ) {
          return "proc check_suggestions {database_handle} {\n\treturn {}\n}\n\n";
     }
     else
          throw runtime_error( "generate_features: unknown proc name to generate: " + aProcName );

} // PackageCreator::generate_features

//======================================================================

string PackageCreator::generate_structure( const string& aProcName, PackageCreator& packageCreator )
{
     static const list<string> empty;
     static list<string> patch_suffixes;
     if ( patch_suffixes.size() == 0 ) {
          for ( const char* const* suffix = PatchSuffixes; *suffix; ++suffix )
               patch_suffixes.push_back( *suffix );
     }
     static list<string> append_suffixes;
     if ( append_suffixes.size() == 0 ) {
          for ( const char* const* suffix = AppendSuffixes; *suffix; ++suffix )
               append_suffixes.push_back( *suffix );
     }

     if ( aProcName == "structure" ) {
          string subpackage_list;
          string package_stream_file_list;
          for ( PackageStreamFileList::const_iterator file = packageCreator.files().begin(); file != packageCreator.files().end(); ++file ) {
               if ( strncmp( (*file)->pathname().c_str(), "man/", 4 ) && strncmp( (*file)->pathname().c_str(), "doc/", 4 ) && strncmp( (*file)->pathname().c_str(), "share/doc/", 10 ) && strncmp( (*file)->pathname().c_str(), "share/man/", 10 ) ) {
                    package_stream_file_list += string( "\t\t\t[" ) + (*file)->tclForCreation( (*file)->attachment() == PackageStreamFile::aPatch ? patch_suffixes : ( (*file)->attachment() == PackageStreamFile::aAppend ? append_suffixes : empty ) ) + "] \\\n";
               }
               else
                    subpackage_list += string( "\t\t\t\t[" ) + (*file)->tclForCreation( (*file)->attachment() == PackageStreamFile::aPatch ? patch_suffixes : ( (*file)->attachment() == PackageStreamFile::aAppend ? append_suffixes : empty ) ) + "] \\\n";
          }
          if ( subpackage_list.length() > 0 )
               subpackage_list = string( "\t\t\t[PackageStreamSubpackage {documentation} {documentation} [list \\\n" ) + subpackage_list + "\t\t\t]] \\\n";

          return string( "proc structure {} {\n\tPackageStreamStructure \"/usr/local\" \\\n\t\t[list \\\n" ) + package_stream_file_list + "\t\t] \\\n\t\t[list \\\n" + subpackage_list + "\t\t] \\\n\t\t{}\n}\n\n";
     }
     else
          throw runtime_error( "generate_structure: unknown proc name to generate: " + aProcName );

} // PackageCreator::generate_structure

//======================================================================

string PackageCreator::generate_install( const string& aProcName, PackageCreator& /*packageCreator*/ )
{
     if ( aProcName == "install" ) {
          return "\
proc install {database_handle package_stream package} {\n\
\n\
	# Insert uninstall proc into database, it will be called in\n\
	# case of package partial installation due to unsolvable problems\n\
	$package uninstall_proc [uninstall_proc]\n\
	$package name [package_name]\n\
\n\
	set err [check_features $database_handle $package_stream]\n\
	if {[string length $err] > 0} {return $err}
\n\
	set err [check_installation_over_installed $database_handle $package_stream]\n\
	if {[string length $err] > 0} {return $err}
\n\
	set err [check_reverse_conflicts $database_handle $package_stream]\n\
	if {[string length $err] > 0} {return $err}
\n\
	set package_structure [structure]\n\
	set err [check_files_installation $database_handle $package_stream [$package_structure files]]\n\
	if {[llength $err] > 0} {\n\
		return $err\n\
	}\n\
\n\
	# Install files supplied with the package\n\
	set err [install_files $database_handle $package_stream $package [$package_structure files]]\n\
	if {[string length $err] > 0} {\n\
		return $err\n\
	}\n\
\n\
	# Update database
	$package description [package_description]\n\
	$package features [provides]\n\
	$package requires [requires]\n\
	$package conflicts [conflicts]\n\
	$package could_be_uninstalled_proc [could_be_uninstalled_proc]\n\
	$package could_be_installed_proc [could_be_installed_proc]\n\
\n\
	# Install subpackages\n\
	if [$package_structure has_subpackages] {\n\
		foreach subpackage [Callback::install_subpackages $package_structure] {\n\
	                set err [check_files_installation $database_handle $package_stream [$subpackage files]]\n\
	                if {[llength $err] == 0} {\n\
				set err [install_files $database_handle $package_stream $package [$subpackage files]]
				if {[string length $err] > 0} {\n\
					break\n\
				}\n\
	                } else {\n\
				break\n\
			}\n\
		}\n\
	}\n\
\n\
	if {[string length $err] > 0} {\n\
	        # Install references\n\
	        if [$package_structure has_references] {\n\
	        }\n\
	}\n\
\n\
	return $err\n\
}\n\
\n\
proc check_features {database_handle package_stream} {\n\
\n\
	# Check if there are requirements, conflicts, users problems.\n\
	# If promlems exist and not solved by callback returns error message,\n\
	# otherwise returns an empty string\n\
\n\
	while 1 {\n\
		set requirements_problems [check_requirements $database_handle]\n\
		set conflicts_problems [check_conflicts $database_handle]\n\
		set users_problems [check_users_requirements]\n\
		set groups_problems [check_groups_requirements]\n\
		set suggestions_dont_met [check_suggestions $database_handle]\n\
		if {[llength $requirements_problems] == 0 && [llength $conflicts_problems] == 0 && [llength $users_problems] == 0 && [llength $groups_problems] == 0 && [llength $suggestions_dont_met] == 0} {\n\
			break\n\
		}\n\
	        set callback_result [Callback::package_requirements_not_met $database_handle $package_stream $requirements_problems $conflicts_problems $users_problems $groups_problems $suggestions_dont_met]\n\
	        switch $callback_result {\n\
	                \"\" { continue }\n\
	                \"ignore\" { break }\n\
	                default { return $callback_result }\n\
	        }\n\
	}\n\
	return \"\"\n\
}\n\
\n\
proc check_installation_over_installed {database_handle package_stream} {\n\
\n\
	# Check if features provided by the package are already provided by the system.\n\
	# returns error message or an empty string if there are no problems.\n\
\n\
	while 1 {\n\
		set features_of_package [provides]\n\
		set duplicated_features {}\n\
		foreach feature_provided $features_of_package {\n\
		        foreach sys_feature [$database_handle feature [$feature_provided name] provide] {\n\
		                if {[$sys_feature equal $feature_provided]} {\n\
		                        lappend duplicated_features $sys_feature\n\
		                }\n\
		        }\n\
		}\n\
	        set callback_called 0\n\
		if {[llength $duplicated_features] > 0} {\n\
		        set callback_result [Callback::installation_over_installed $database_handle $package_stream $duplicated_features]\n\
		        set callback_called 1\n\
		}\n\
		if $callback_called {\n\
			if {[string length $callback_result] > 0} {\n\
				return $callback_result\n\
			}\n\
		} else {\n\
			break\n\
		}\n\
	}\n\
	return \"\"\n\
}\n\
\n\
proc check_reverse_conflicts {database_handle package_stream} {\n\
\n\
	# Check if packages installed have features provided by the package\n\
        # being installed in their conflicting lists.\n\
	# returns error message or an empty string if there are no problems.\n\
\n\
	while 1 {\n\
		set conflicting_packages {}\n\
		set features_of_package [provides]\n\
		foreach feature_provided $features_of_package {\n\
			set conflicting_packages [concat $conflicting_packages [$database_handle packages_conflicting $feature_provided]]\n\
		}\n\
	        set callback_called 0\n\
		if {[llength $conflicting_packages] >0} {\n\
		        set callback_result [Callback::reverse_conflict $database_handle $package_stream $conflicting_packages]\n\
		        set callback_called 1\n\
		}\n\
		if $callback_called {\n\
			if {[string length $callback_result] > 0} {\n\
				return $callback_result\n\
			}\n\
		} else {\n\
			break\n\
		}\n\
	}\n\
	return ""\n\
}\n\
\n\
proc check_files_installation {database_handle package_stream files} {\n\
\n\
	# Check installation problems for the files supplied with the package\n\
\n\
	set err {}\n\
	foreach file $files {\n\
		if [$package_stream file_exists $file] {\n\
			set owners [$database_handle packages_owning [$package_stream file_pathname_for_installation $file]]\n\
			if {[$file attached] == \"file\"} {\n\
				switch [$file replacement_scheme] {\n\
					\"always_replace\" {}\n\
					\"never_replace\" {}\n\
					default {\n\
						if {[llength $owners] > 0} {\n\
							set err_ret [Callback::file_to_be_installed_exists_and_owned_by_other_packages $database_handle $package_stream $file $owners]\n\
							if {[string length $err_ret] > 0} {\n\
								lappend err $err_ret\n\
							}\n\
						}\n\
					}\n\
				}\n\
			}\n\
		} else {\n\
			if {[$file attached] == \"patch\"} {\n\
				lappend err [format \"File %s is to be patched but does not exist\" [$file pathname]]\n\
				break\n\
			}\n\
		}\n\
	}\n\
	return $err\n\
}\n\
\n\
proc install_files {database_handle package_stream package files} {\n\
	set error {}\n\
	foreach file $files {\n\
		if [$package_stream file_exists $file] {\n\
			set owners [$database_handle packages_owning [$package_stream file_pathname_for_installation $file]]\n\
			if {[$file attached] == \"file\"} {\n\
				switch [$file replacement_scheme] {\n\
					\"always_replace\" {set installed_file [$package_stream install_file $file]}\n\
					\"never_replace\" {set installed_file [$package_stream file_pathname_for_installation $file]}\n\
					default {set installed_file [$package_stream install_file $file]}\n\
				}\n\
			} elseif {[$file attached] == \"patch\"} {\n\
				set installed_file [$package_stream patch_file $file]\n\
			} elseif {[$file attached] == \"append\"} {\n\
				set installed_file [$package_stream append_file $file]\n\
			} else {\n\
				if {[string length [$file symlink]] > 0} {\n\
					set installed_file [$package_stream create_symlink $file]\n\
				} elseif {[llength [$file device]] > 0} {\n\
					set installed_file [$package_stream create_device $file]\n\
				}\n\
			}\n\
		} else {\n\
			if {[$file attached] == \"patch\"} {\n\
				set error [format \"File %s is to be patched but does not exist\" [$file pathname]]\n\
			} elseif {[$file attached] == \"file\"} {\n\
				set installed_file [$package_stream install_file $file]\n\
			} elseif {[$file attached] == \"append\"} {\n\
				set installed_file [$package_stream append_file $file create]\n\
			} else {\n\
				if {[string length [$file symlink]] > 0} {\n\
					set installed_file [$package_stream create_symlink $file]\n\
				} elseif {[llength [$file device]] > 0} {\n\
					set installed_file [$package_stream create_device $file]\n\
				}\n\
			}\n\
		}\n\
		if {[string length $error] == 0} {\n\
			$package own $installed_file\n\
		}\n\
		if {[string length $error] > 0} {\n\
			break\n\
		}\n\
	}\n\
				return $error\n\
}\n\
\n\
\n";
     }
     else if ( aProcName == "upgrade" ) {
          return "\
proc upgrade {database_handle package_stream package} {\n\
\n\
	install $database_handle $package_stream $package\n\
}\n\
\n";
     }
     else if ( aProcName == "uninstall_proc" ) {
          return "proc uninstall_proc {} { return {\n\
proc uninstall {database_handle package} {\n\
\n\
	# Check if other packages require presence of this package\n\
	set packages_requiring [$database_handle packages_requiring [$package features]]\n\
	set packages_really_requiring {}\n\
	foreach package_requiring $packages_requiring {\n\
		if {[$package_requiring could_be_uninstalled [$package features]] == 0} {\n\
			lappend packages_really_requiring $package_requiring\n\
		}\n\
	}\n\
	if {[llength $packages_really_requiring] > 0} {\n\
		set callback_result [Callback::uninstall_required_features $database_handle $package $packages_really_requiring]\n\
		if {[string length $callback_result] > 0} {\n\
			return $callback_result\n\
		}\n\
	}\n\
\n\
	# Uninstall owned files\n\
	foreach file [$package files] {\n\
		$package uninstall_file $file\n\
	}\n\
\n\
	# Erase package database record\n\
	$package remove\n\
\n\
	return \"\"\n\
}\n\
}\n\
}\n";
     }
     else
          throw runtime_error( "generate_install: unknown proc name to generate: " + aProcName );

} // PackageCreator::generate_install

//======================================================================

void PackageCreator::writeAbsentProcs( const string& aDir, SysinstallRealScriptList& scripts )
{
     File::make_path( aDir );

     for ( SysinstallRealScriptList::iterator script = scripts.begin(); script != scripts.end(); ++script ) {
          string file = aDir + "/" + (*script)->filename;
          int fd = ::open( file.c_str(), O_WRONLY | O_CREAT /*| O_APPEND*/, 0644 );
          if ( fd < 0 )
               throw runtime_error( string( "Unable to open " ) + file + " for appending: " + ::strerror( errno ) );
          try {
               if ( (*script)->absentProcs.size() > 0 && ::lseek( fd, 0, SEEK_END ) != 0 )
                    ::write( fd, "\n\n\n", 3 );
               for ( list<string>::const_iterator absent = (*script)->absentProcs.begin(); absent != (*script)->absentProcs.end(); ++absent ) {
                    if ( ::write( fd, absent->c_str(), absent->length() ) != int( absent->length() ) )
                         throw runtime_error( string( "Unable to write " ) + file + ": " + ::strerror( errno ) );
               }
          }
          catch ( ... ) {
               close( fd );
               throw;
          }

          close( fd );
     }

} // PackageCreator::writeAbsentProcs

//======================================================================

void PackageCreator::checkProcs( const string& aNamespace )
{
     for ( const SysinstallScript* script = sScripts; script->filename; ++script ) {
          LOG( INFO, "Checking " << script->filename );
          for ( const char** proc = script->procs; *proc; ++proc ) {
               LOG( INFO, "\tChecking proc " << *proc );
               try {
                    script->check( *proc, aNamespace, *this );
               }
               catch ( exception& err ) {
                    LOG( ERROR, "Invalid " << script->filename << " proc " << *proc << ": " << err.what() );
               }
          }
     }

} // PackageCreator::checkProcs

//======================================================================

void PackageCreator::check_description( const string& aProcName, const string& aNamespace, PackageCreator& packageCreator )
{
     if ( aProcName == "package_name" ) {
          packageCreator.interpreter()->eval( aNamespace + "::" + aProcName );
     }
     else if ( aProcName == "package_description" ) {
          PackageStream_CallPackageSuppliedProc<Pointer<PackageDescription> > call( new TclLanguageInterface::CallbackDataSlaveInterpreter( packageCreator.interpreter(), string( "::" ) + aNamespace + "::" + aProcName ) );
          call.activate();
          if ( call.result()->check_integrity() )
               throw runtime_error( "Inavlid integrity" );
     }
     else
          throw runtime_error( "check_description: unknown proc name to check: " + aProcName );

} // PackageCreator::check_description

//======================================================================

void PackageCreator::check_features( const string& aProcName, const string& aNamespace, PackageCreator& packageCreator )
{
     if ( aProcName == "provides" ) {
          PackageStream_CallPackageSuppliedProc<Pointer<FeatureList> > call( new TclLanguageInterface::CallbackDataSlaveInterpreter( packageCreator.interpreter(), string( "::" ) + aNamespace + "::" + aProcName ) );
          call.activate();
          if ( call.result()->size() == 0 )
               LOG( WARNING, "Package provides no features" );
     }
     else if ( aProcName == "requires" || aProcName == "conflicts" ) {
          PackageStream_CallPackageSuppliedProc<Pointer<FeatureList> > call( new TclLanguageInterface::CallbackDataSlaveInterpreter( packageCreator.interpreter(), string( "::" ) + aNamespace + "::" + aProcName ) );
          call.activate();
     }
     else if ( aProcName == "check_requirements" ) {
     }
     else if ( aProcName == "check_conflicts" ) {
     }
     else if ( aProcName == "could_be_installed_proc" ) {
     }
     else if ( aProcName == "could_be_uninstalled_proc" ) {
     }
     else if ( aProcName == "check_users_requirements" ) {
     }
     else if ( aProcName == "check_groups_requirements" ) {
     }
     else if ( aProcName == "check_suggestions" ) {
     }
     else
          throw runtime_error( "check_features: unknown proc name to check: " + aProcName );

} // PackageCreator::check_features

//======================================================================

void PackageCreator::check_structure( const string& aProcName, const string& aNamespace, PackageCreator& packageCreator )
{
     if ( aProcName == "structure" ) {
          PackageStream_CallPackageSuppliedProc<Pointer<PackageStreamStructure> > call( new TclLanguageInterface::CallbackDataSlaveInterpreter( packageCreator.interpreter(), string( "::" ) + aNamespace + "::" + aProcName ) );
          call.activate();
          if ( call.result()->check_integrity() )
               throw runtime_error( "Inavlid integrity" );
          if ( call.result()->checkPackageCreationData( packageCreator.files() ) )
               throw runtime_error( "Inavlid package creation data" );
     }
     else
          throw runtime_error( "check_structure: unknown proc name to check: " + aProcName );

} // PackageCreator::check_structure

//======================================================================

void PackageCreator::check_install( const string& aProcName, const string& aNamespace, PackageCreator& packageCreator )
{
     if ( aProcName == "install" ) {
            // ????
     }
     else if ( aProcName == "upgrade" ) {
     }
     else if ( aProcName == "uninstall_proc" ) {
          PackageStream_CallPackageSuppliedProc<string> call( new TclLanguageInterface::CallbackDataSlaveInterpreter( packageCreator.interpreter(), string( "::" ) + aNamespace + "::" + aProcName ) );
          call.activate();
     }
     else
          throw runtime_error( "check_install: unknown proc name to check: " + aProcName );

} // PackageCreator::check_install

//======================================================================
