/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/sysinstall/PackageStream.hh,v 1.2 2001/06/22 15:05:17 alex Exp $
 */

// -*- C++ -*-  Wed Aug 19 1998
// $Id: PackageStream.hh,v 1.7 1999/01/04 11:13:25 eugene Exp $

#ifndef PackageStream_hh
#define PackageStream_hh

#ifndef Feature_hh
#include "Feature.hh"
#endif

#ifndef TclLanguageInterface_hh
#include "TclLanguageInterface.hh"
#endif

#ifndef Archive_hh
#include "Archive.hh"
#endif

#ifndef Package_hh
#include "Package.hh"
#endif

#ifndef Database_hh
#include "Database.hh"
#endif

#ifndef ProgressIndicator_hh
#include "ProgressIndicator.hh"
#endif

//======================================================================

class PackageStreamFile;
class PackageStreamStructure;

//======================================================================

class PackageStream : virtual public LanguageInterface::Object
{
     CLASS_METHODS_DESCRIPTIONS_DECLARATION;

  public:
     PackageStream( Pointer<Archive> aStream, const string& aSlaveInterpreterPrefix, Pointer<ProgressIndicatorStack> aProgressIndicatorStack );
     virtual ~PackageStream();

     static Pointer<PackageStream> open( Pointer<Archive> aStream, const string& aSlaveInterpreterPrefix, Pointer<ProgressIndicatorStack> aProgressIndicatorStack );

       // Returns name of namespace where all procs supplied with the
       // package file are to be sourced.
     virtual const string& name_space() const;

       // Returns package_name. Sources the SYSINSTALL/description.tcl
       // script if not sourced yet and then calls
       // package_namespace::name
     virtual string name();

       // Returns package_description. Sources the
       // SYSINSTALL/description.tcl script if not sourced yet and
       // then calls package_namespace::description
     virtual Pointer<PackageDescription> description();

       // Returns list of features provided by the package. Sources
       // the SYSINSTALL/features.tcl script if not sourced yet and
       // then calls package_namespace::provides
     virtual Pointer<FeatureList> features();

       // Returns list of required features for the package. Sources
       // the SYSINSTALL/features.tcl script if not sourced yet and
       // then calls package_namespace::requires
     virtual Pointer<FeatureList> requires();

       // Returns list of conflicting features for the
       // package. Sources the SYSINSTALL/features.tcl script if not
       // sourced yet and then calls package_namespace::conflicts
     virtual Pointer<FeatureList> conflicts();

       // Returns string with proc code. Sources the
       // SYSINSTALL/features.tcl script if not sourced yet.
     virtual string could_be_uninstalled_proc();

       // Returns string with proc code. Sources the
       // SYSINSTALL/features.tcl script if not sourced yet.
     virtual string could_be_installed_proc();

       // Returns package_structure. Sources the
       // SYSINSTALL/structure.tcl script if not sourced yet and then
       // calls package_namespace::structure
     virtual Pointer<PackageStreamStructure> structure();

       // Sources all not yet sourced scripts supplied with the
       // package file.
     virtual void source();

       // Calls proc install supplied with package in the safe interpreter
     virtual void install( Pointer<Database> database );

       // Calls proc upgrade supplied with package in the safe interpreter
     virtual void upgrade( Pointer<Database> database );

       // Returns installation prefix based on configuration
     virtual string prefix();

       // Returns information about existing file which will be
       // overwritten or patched by aFile installation or 0 if file
       // does not exist
       //virtual PackageFile* existing_file( Pointer<PackageStreamFile> aFile );

     virtual string create_symlink( Pointer<PackageStreamFile> aFile );
     virtual string create_device( Pointer<PackageStreamFile> aFile );
     virtual string create_hardlink( Pointer<PackageStreamFile> aFile );

       // Installs the file attached to the package. Returns file
       // pathname. If file already exists it is replaced. (Uses
       // PackageStream::File::write) [If Configuration::no_action
       // returns non-zero does not install file and does not change
       // file ownership for other packages just reports what is to be
       // done. If file is to be installed under the path for which
       // access is forbidden by the configuration (see the Security
       // section) exception is raised.]
     virtual string install_file( Pointer<PackageStreamFile> aFile );

       // Patches original file using the patch attached to the
       // package. Returns file pathname. If original file does not
       // exist raises an exception. (Uses PackageStream::File::patch)
       // [If Configuration::no_action returns non-zero does not patch
       // file and does not change file ownership for other packages
       // just reports what is to be done. If file to be patched is
       // under the path for which access is forbidden by the
       // configuration (see the Security section) exception is
       // raised.]
     virtual string patch_file( Pointer<PackageStreamFile> aFile );

       // Append to original file using the data attached to the
       // package. Returns file pathname. If original file does not
       // exist and !create raises an exception. (Uses
       // PackageStream::File::patch) [If Configuration::no_action
       // returns non-zero does not patch file and does not change
       // file ownership for other packages just reports what is to be
       // done. If file to be patched is under the path for which
       // access is forbidden by the configuration (see the Security
       // section) exception is raised.]
     virtual string append_file( Pointer<PackageStreamFile> aFile, bool create );

       // Installs each reference containing in the package using a
       // separate stream.
     virtual void install_references();

     virtual bool file_exists( Pointer<PackageStreamFile> aFile );
     virtual string file_pathname_for_installation( Pointer<PackageStreamFile> aFile );

     static const char* description_helper_procs;
     static const char* features_helper_procs;
     static const char* structure_helper_procs;
     static const char* install_helper_procs;

     virtual Pointer<ProgressIndicatorStack> progress_indicator_stack() { return mProgressIndicatorStack; }

  protected:
     Archive& stream() { return *mStream; }
     Pointer<TclLanguageInterface::SlaveInterpreter> slave() { return mSlaveInterp; }
     ProgressIndicatorStack& progress() { return *mProgressIndicatorStack; }

  private:
     Pointer<Archive> mStream;
     string mNameSpace;
     Pointer<TclLanguageInterface::SlaveInterpreter> mSlaveInterp;
     string mPrefix;
     Pointer<Database> mDatabase;
     string mSlaveInterpreterPrefix;
     Pointer<ProgressIndicatorStack> mProgressIndicatorStack;

     void source_description();
     void source_features();
     void source_structure();
     void source_install();

       // Check if the specified command is present in slave
       // interpreter in the specified namespace.
     bool command_exists( const char* cmdName, const string& aNamespace );

       // Source the specified file into slave interpreter into
       // name_space()
     void source( const string& file );

       // Calls callback and throws an exception if abort is returned
     void directory_write_access( Pointer<PackageStreamFile> aFile );

       // If necessary creates directories leading to this file
     virtual void create_leading_dirs( Pointer<PackageStreamFile> aFile );

     void update_features_provided_database();

}; // class PackageStream

//======================================================================

template <typename Result> class PackageStream_CallPackageSuppliedProc : public LanguageInterface::SpecificCallbackData
{
  public:
     PackageStream_CallPackageSuppliedProc( Pointer<TclLanguageInterface::CallbackDataSlaveInterpreter> aCallbackData )
          : LanguageInterface::SpecificCallbackData( aCallbackData ) {}

     virtual const char* typeid_name() const { return typeid( PackageStream ).name(); }
     virtual void activate() { if ( callbackData() ) callbackData()->activate( *this ); }

     Result& result() { return mResult; }
     const Result& result() const { return mResult; }

  private:
     Result mResult;

}; // class PackageStream_CallPackageSuppliedProc

//======================================================================

class PackageStream_CallPackageInstall : public LanguageInterface::SpecificCallbackData
{
  public:
     PackageStream_CallPackageInstall( Pointer<TclLanguageInterface::CallbackDataSlaveInterpreter> aCallbackData, Pointer<Database> aDatabase, Pointer<PackageStream> aPackageStream, Pointer<Package> aPackage )
          : LanguageInterface::SpecificCallbackData( aCallbackData ), mDatabase( aDatabase ), mPackageStream( aPackageStream ), mPackage( aPackage ) {}

     Pointer<Database> database() { return mDatabase; }
     Pointer<PackageStream> packageStream() { return mPackageStream; }
     Pointer<Package> package() { return mPackage; }

     virtual const char* typeid_name() const { return typeid( PackageStream ).name(); }
     virtual void activate() { if ( callbackData() ) callbackData()->activate( *this ); }

     string& result() { return mResult; }
     const string& result() const { return mResult; }

  private:
     string mResult;
     Pointer<Database> mDatabase;
     Pointer<PackageStream> mPackageStream;
     Pointer<Package> mPackage;

}; // class PackageStream_CallPackageInstall

//======================================================================

#endif // PackageStream_hh
