/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/tcl/TclLanguageInterface.hh,v 1.3 2001/06/20 10:30:00 alex Exp $
 */

// -*- C++ -*-  Thu Nov  5 1998
// Eugene Skepner <eu@qub.com>
// $Id: TclLanguageInterface.hh,v 1.36 1999/01/04 23:15:58 eugene Exp $

#ifndef TclLanguageInterface_hh
#define TclLanguageInterface_hh

#ifndef LanguageInterface_hh
#include "LanguageInterface.hh"
#endif

#ifndef String_hh
#include "String.hh"
#endif

#ifndef Type_hh
#include "Type.hh"
#endif

#ifndef Container_hh
#include "Container.hh"
#endif

#include <tcl.h>

//======================================================================

class TclLanguageInterface : public LanguageInterface
{
  public:
     TclLanguageInterface( Tcl_Interp * aInterp, bool enableGC );
     virtual ~TclLanguageInterface();

     Tcl_Interp* interpreter() { return mInterp; }

     static const char* to_string( Tcl_Obj* obj )
          {
               if ( obj )
                    return Tcl_GetStringFromObj( obj, 0 );
               else
                    return (char*)"<null Tcl_Obj*>";
          }

     string object_type( Tcl_Obj* obj );

     static int tcl_obj2int( Tcl_Interp* interp, Tcl_Obj* obj, const char* method = 0, int argno = 0 )
          {
               int value;
               if ( Tcl_GetIntFromObj( interp, obj, &value ) != TCL_OK )
                    throw Error( string( "Invalid method " ) + ( method ? method : "" ) + " argument " + ::to_string( argno ) + ": \"" + to_string( obj ) + "\" (int expected)" );
               return value;
          }

     static long tcl_obj2long( Tcl_Interp* interp, Tcl_Obj* obj, const char* method = 0, int argno = 0 )
          {
               long value;
               if ( Tcl_GetLongFromObj( interp, obj, &value ) != TCL_OK )
                    throw Error( string( "Invalid method " ) + ( method ? method : "" ) + " argument " + ::to_string( argno ) + ": \"" + to_string( obj ) + "\" (long expected)" );
               return value;
          }

     static bool tcl_obj2bool( Tcl_Interp* interp, Tcl_Obj* obj, const char* method = 0, int argno = 0 )
          {
               int value;
               if ( Tcl_GetBooleanFromObj( interp, obj, &value ) != TCL_OK )
                    throw Error( string( "Invalid method " ) + ( method ? method : "" ) + " argument " + ::to_string( argno ) + ": \"" + to_string( obj ) + "\" (int expected)" );
               return value;
          }

     static double tcl_obj2double( Tcl_Interp* interp, Tcl_Obj* obj, const char* method = 0, int argno = 0 )
          {
               double value;
               if ( Tcl_GetDoubleFromObj( interp, obj, &value ) != TCL_OK )
                    throw Error( string( "Invalid method " ) + ( method ? method : "" ) + " argument " + ::to_string( argno ) + ": \"" + to_string( obj ) + "\" (double expected)" );
               return value;
          }

     static string tcl_obj2string( Tcl_Interp* /*interp*/, Tcl_Obj* obj, const char* method = 0, int argno = 0 )
          {
               const char* s = Tcl_GetStringFromObj( obj, 0 );
               if ( s )
                    return s;
               else
                    throw Error( string( "Invalid method " ) + ( method ? method : "" ) + " argument " + ::to_string( argno ) + ": \"" + to_string( obj ) + "\" (string expected)" );
          }

     template <class O> Pointer<O> tcl2object( Tcl_Obj* obj, O* )
          {
               if ( obj == 0 )
                    return Pointer<O>();

               Object* o = 0;
               if ( gc() ) {
                    if ( obj->typePtr != &sHtcl )
                         throw Error( string( "Invalid Tcl_Obj type (expected htcl, was " ) + ( obj->typePtr? obj->typePtr->name : "none" ) + ")" );
                    o = reinterpret_cast<Object*>( obj->internalRep.otherValuePtr );
               }
               else {
                    const char* s = Tcl_GetStringFromObj( obj, 0 );
                    if ( s != 0 ) {
                         void* p;
                         char type[256];
                         if ( sscanf( s, ObjectStringFormat, &p, &type ) == 2 ) {
                              o = reinterpret_cast<Object*>( p );
                              if ( o && !referenced( o ) )
                                   throw Error( string( "Not a reference to existing Object" ) );
                         }
                         else if ( !strncmp( s, "(nil)_", 6 ) )
                              o = 0;
                         else
                              throw Error( string( "Not a reference to Object" ) );
                    }
               }

               if ( o == 0 )
                    return Pointer<O>();
               O* ptr = dynamic_cast<O*>( o );
               if ( ptr == 0 )
                    throw Error( string( "Cannot cast object of type " ) + type_to_name( o->objecttypename() ) + " to " + type_to_name( typeid( O ).name() ) );
               return ptr;
          }

     template <class O> Pointer<O> tcl2object( Tcl_Obj* obj, O* pO, const char* method = 0, int argno = 0 )
          {
               try {
                    return this->tcl2object( obj, pO ); // have to use this due to 2.8.1 internal compiler error
               }
               catch ( exception& err ) {
                    throw Error( string( "Invalid method " ) + ( method ? method : "" ) + " argument " + ::to_string( argno ) + ": " + err.what() + " (reference to " + type_to_name( typeid( pO ).name() ) + " or inherited from it expected)" );
               }
          }

     void tcl2cxx( Tcl_Interp* interp, Tcl_Obj* obj, short& value ) { value = tcl_obj2int( interp, obj ); }
     void tcl2cxx( Tcl_Interp* interp, Tcl_Obj* obj, unsigned short& value ) { value = tcl_obj2int( interp, obj ); }
     void tcl2cxx( Tcl_Interp* interp, Tcl_Obj* obj, int& value ) { value = tcl_obj2int( interp, obj ); }
     void tcl2cxx( Tcl_Interp* interp, Tcl_Obj* obj, unsigned int& value ) { value = tcl_obj2int( interp, obj ); }
     void tcl2cxx( Tcl_Interp* interp, Tcl_Obj* obj, long& value ) { value = tcl_obj2long( interp, obj ); }
     void tcl2cxx( Tcl_Interp* interp, Tcl_Obj* obj, unsigned long& value ) { value = tcl_obj2long( interp, obj ); }
     void tcl2cxx( Tcl_Interp* interp, Tcl_Obj* obj, bool& value ) { value = tcl_obj2bool( interp, obj ); }
     void tcl2cxx( Tcl_Interp* interp, Tcl_Obj* obj, float& value ) { value = tcl_obj2double( interp, obj ); }
     void tcl2cxx( Tcl_Interp* interp, Tcl_Obj* obj, double& value ) { value = tcl_obj2double( interp, obj ); }
     void tcl2cxx( Tcl_Interp* interp, Tcl_Obj* obj, string& value ) { value = tcl_obj2string( interp, obj ); }
     void tcl2cxx( Tcl_Interp* interp, Tcl_Obj* obj, char& value ) { value = tcl_obj2string( interp, obj )[0]; }

     template <class O> void tcl2cxx( Tcl_Interp* /*interp*/, Tcl_Obj* obj, Pointer<O>& value )
          {
               value = this->tcl2object( obj, (O*)0 ); // have to use this due to 2.8.1 internal compiler error
          }

     template <class C> void append2list( C& result, Tcl_Interp* interp, Tcl_Obj* obj )
          {
               typename C::value_type value;
               tcl2cxx( interp, obj, value );
               insert( result, value );
          }

     template <class C> Pointer< Container<C> > tcl2list( Tcl_Interp* interp, Tcl_Obj* aList, C* pO )
          {
               Pointer< Container<C> > result = new Container<C>();
               if ( aList->typePtr == &sHtcl ) { // list of one element
                    this->append2list( *result, interp, aList ); // have to use this due to 2.8.1 internal compiler error
               }
               else {
                    int size;
                    if ( Tcl_ListObjLength( interp, aList, &size ) != TCL_OK )
                         throw Error( "Cannot get list length" );
                    for ( int i = 0; i < size; ++i ) {
                         Tcl_Obj* obj;
                         if ( Tcl_ListObjIndex( interp, aList, i, &obj ) != TCL_OK )
                              throw Error( "Cannot get list element" );
                         append2list( *result, interp, obj );
                    }
               }
               return result;
          }

     template <class C> void tcl2cxx( Tcl_Interp* interp, Tcl_Obj* aList, Pointer< Container<C> >& value )
          {
               value = this->tcl2list( interp, aList, (C*)0 );
          }

     template <class C> Pointer< Container<C> > tcl2list( Tcl_Interp* interp, Tcl_Obj* aList, C* pO, const char* method, int argno )
          {
               try {
                    return this->tcl2list( interp, aList, pO ); // have to use this due to 2.8.1 internal compiler error
               }
               catch ( exception& err ) {
                    throw Error( string( "Invalid method " ) + method + " argument " + ::to_string( argno ) + " type (list of references to " + type_to_name( typeid( C::value_type ).name() ) + " expected)" );
               }
          }

     template <class C> void tcl2list( C& result, Tcl_Interp* interp, Tcl_Obj* aList )
          {
               if ( aList->typePtr == &sHtcl ) { // list of one element
                    this->append2list( result, interp, aList ); // have to use this due to 2.8.1 internal compiler error
               }
               else {
                    int size;
                    if ( Tcl_ListObjLength( interp, aList, &size ) != TCL_OK )
                         throw Error( "Cannot get list length" );
                    for ( int i = 0; i < size; ++i ) {
                         Tcl_Obj* obj;
                         if ( Tcl_ListObjIndex( interp, aList, i, &obj ) != TCL_OK )
                              throw Error( "Cannot get list element" );
                         this->append2list( result, interp, obj );
                    }
               }
          }

     template <class C> void tcl2list( C& result, Tcl_Interp* interp, Tcl_Obj* aList, const char* method, int argno )
          {
               try {
                    this->tcl2list( result, interp, aList );
               }
               catch ( exception& err ) {
                    throw Error( string( "Invalid method " ) + method + " argument " + ::to_string( argno ) + " type (list of references to " + type_to_name( typeid( C::value_type ).name() ) + " expected): " + err.what() );
               }
          }

     template <class C> void tcl2list( Pointer<C>& result, Tcl_Interp* interp, Tcl_Obj* aList, const char* method, int argno )
          {
               result = new C;
               this->tcl2list( *result, interp, aList, method, argno );
          }

     Pointer<LanguageInterface::CallbackData> tcl2callbackData( Tcl_Interp* interp, Tcl_Obj* aData );

       // --------------------------------------------------

     virtual Tcl_Obj* object2tcl( Pointer<Object> ptr );

     Tcl_Obj* cxx2tcl( int value ) { return Tcl_NewIntObj( value ); }
     Tcl_Obj* cxx2tcl( short value ) { return Tcl_NewIntObj( value ); }
     Tcl_Obj* cxx2tcl( unsigned short value ) { return Tcl_NewIntObj( value ); }
     Tcl_Obj* cxx2tcl( long value ) { return Tcl_NewLongObj( value ); }
     Tcl_Obj* cxx2tcl( unsigned int value ) { return Tcl_NewLongObj( value ); }
     Tcl_Obj* cxx2tcl( unsigned long value ) { return Tcl_NewLongObj( value ); }
     Tcl_Obj* cxx2tcl( long long value ) { return Tcl_NewLongObj( value ); }
     Tcl_Obj* cxx2tcl( unsigned long long value ) { return Tcl_NewLongObj( value ); }
     Tcl_Obj* cxx2tcl( float value ) { return Tcl_NewDoubleObj( value ); }
     Tcl_Obj* cxx2tcl( double value ) { return Tcl_NewDoubleObj( value ); }
     Tcl_Obj* cxx2tcl( char value ) { return Tcl_NewStringObj( &value, 1 ); }
     Tcl_Obj* cxx2tcl( const string& value ) { return Tcl_NewStringObj( const_cast<char*>( value.c_str() ), -1 ); }
     Tcl_Obj* cxx2tcl( bool value ) { return Tcl_NewBooleanObj( value ); }
     template <class O> Tcl_Obj* cxx2tcl( Pointer<O> ptr ) { return object2tcl( ptr ); }

     template <class C> Tcl_Obj* list2tcl( Tcl_Interp* interp, const Pointer<Container<C> >& aList )
          {
             Tcl_Obj* result = Tcl_NewObj();
               if ( aList ) {
                    for ( typename C::const_iterator element = aList->begin(); element != aList->end(); ++element ) {
                         Tcl_Obj* el = this->cxx2tcl( *element );
                         if ( Tcl_ListObjAppendElement( interp, result, el ) != TCL_OK )
                              throw Error( string( "Error creating result list: " ) + Tcl_GetStringResult( interp ) );
                    }
               }
             return result;
          }

     template <class C> Tcl_Obj* cxx2tcl( const Pointer<Container<C> >& aList )
          {
               return list2tcl( interpreter(), aList );
          }

     template <class List> Tcl_Obj* list2tcl( Tcl_Interp* interp, const List& aList )
          {
               Tcl_Obj* result = Tcl_NewObj();
               for ( typename List::const_iterator element = aList.begin(); element != aList.end(); ++element ) {
                    Tcl_Obj* el = this->cxx2tcl( *element );
                    if ( Tcl_ListObjAppendElement( interp, result, el ) != TCL_OK )
                         throw Error( string( "Error creating result list: " ) + Tcl_GetStringResult( interp ) );
               }
               return result;
          }

     template <class C> Tcl_Obj* cxx2tcl( const Container<C>& aList )
          {
               return list2tcl( interpreter(), aList );
          }

     template <class C> Tcl_Obj* cxx2tcl( const list<C>& aList )
          {
               return list2tcl( interpreter(), aList );
          }

     Tcl_Obj* callbackData2tcl( Tcl_Interp* interp, Pointer<LanguageInterface::CallbackData> aCallbackData );

       // --------------------------------------------------

     class CommandAccess
     {
       public:
          CommandAccess() {}
          virtual ~CommandAccess() {}

          virtual void create( TclLanguageInterface& aTclLanguageInterface, const Object* aObject );
          virtual Tcl_ObjCmdProc* methodCmd() const = 0;

          virtual void activate_callback( Tcl_Interp* aInterp, Tcl_Obj* aCommand, Tcl_Obj* aUserData, Tcl_Obj* aArg, SpecificCallbackData& aSpecificCallbackData );

          static int error( Tcl_Interp* aInterp, const char* message ... );
     };

     virtual void registerCommandAccess( const string& aName, CommandAccess* aAccess );
     virtual CommandAccess* commandAccess( const string& aName );

     virtual void registerObject( const Object* aObject );
     virtual void unregisterObject( const Object* aObject );

       // --------------------------------------------------

     class CallbackData : public LanguageInterface::CallbackData
     {
       public:
          CallbackData( TclLanguageInterface* aLanguageInterface, Tcl_Obj* aCommand, Tcl_Obj* aUserData );
          CallbackData( TclLanguageInterface* aLanguageInterface, const string& aCommand );
          virtual ~CallbackData();

          virtual void activate( SpecificCallbackData& aSpecificCallbackData );
          virtual void activate( SpecificCallbackData& aSpecificCallbackData, Tcl_Obj* arg );

          virtual void activate( SpecificCallbackData& aSpecificCallbackData, int arg )
               {
                    activate( aSpecificCallbackData, Tcl_NewIntObj( arg ) );
               }

          Tcl_Obj* command() { return mCommand; }
          Tcl_Obj* userData() { return mUserData; }

       private:
          Tcl_Obj* mCommand;
          Tcl_Obj* mUserData;

     }; // class CallbackData

       // --------------------------------------------------

     class SlaveInterpreter : virtual public Shared
     {
       public:
          virtual ~SlaveInterpreter();

          virtual void alias_command( const string& aCommand );
          virtual void delete_command( const string& aCommand );

          virtual bool command_exists( const string& aCommand );
          virtual bool variable_exists( const string& aVariable );
          virtual const char* eval( const string& data, const string& aNamespace );
          virtual const char* eval( const string& data );
          virtual const char* eval( const void* data, unsigned long length, const string& aNamespace )
               { return eval( string( (const char*)data, length ), aNamespace ); }

          Tcl_Interp* interpreter() { return mInterp; }
          TclLanguageInterface& tclLanguageInterface() { return mParent; }

       private:
          string mInterpName;
          Tcl_Interp* mInterp;
          TclLanguageInterface& mParent;

          friend class TclLanguageInterface;
          SlaveInterpreter( const string& aInterpName, Tcl_Interp* aInterp, TclLanguageInterface& aParent );

          static unsigned sInterpreterSerial;

     }; // class SlaveInterpreter

       // --------------------------------------------------

     class CallbackDataSlaveInterpreter : public CallbackData
     {
       public:
          CallbackDataSlaveInterpreter( Pointer<SlaveInterpreter> aSlaveInterpreter, const string& aCommand );
          CallbackDataSlaveInterpreter( Pointer<SlaveInterpreter> aSlaveInterpreter, const string& aCommand, const string& aUserData );
          CallbackDataSlaveInterpreter( Pointer<SlaveInterpreter> aSlaveInterpreter, Tcl_Obj* aCommand, Tcl_Obj* aUserData );
          virtual void activate( SpecificCallbackData& aSpecificCallbackData );

       private:
          Pointer<SlaveInterpreter> mSlaveInterpreter;

     }; // class CallbackDataSlaveInterpreter

       // --------------------------------------------------

     Pointer<SlaveInterpreter> make_slave_interpreter( const string& aNamePrefix, bool aSafe );

     virtual void create_command( const string& aName, Tcl_ObjCmdProc* aProc, void* aData );


  private:
     Tcl_Interp* mInterp;
     list<SlaveInterpreter*> mSlaveInterpreters;
     list<string> mStaticMethods;

     typedef map<string,CommandAccess*> CommandsAccess;
     CommandsAccess mCommandsAccess;

     friend class SlaveInterpreter;
     friend class CommandAccess;
     void slave_interpreter_destroyed( SlaveInterpreter* aSlaveInterpreter );

     void registerObjType();

     static Tcl_ObjType sHtcl;
     static void HTclFreeInternalRepProc( Tcl_Obj* obj );
     static void HTclDupInternalRepProc( Tcl_Obj* src, Tcl_Obj* dest );
     static void HTclUpdateStringProc( Tcl_Obj* obj );
     static int HTclSetFromAnyProc( Tcl_Interp* aInterp, Tcl_Obj* obj );

     static string master_command_name( const string& aCommandName ) { return aCommandName + "_master"; }

}; // class TclLanguageInterface

//======================================================================

extern TclLanguageInterface* tclLanguageInterface;

//======================================================================

#endif // TclLanguageInterface_hh
