/*-
 * Copyright (c) 1998-2001 The FreeBSD Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation
 * are those of the authors and should not be interpreted as representing
 * official policies, either expressed or implied, of the FreeBSD Project.
 *
 * $FreeBSD: libh/lib/tcl/tcl_interface_gen.cc,v 1.6 2001/06/20 10:30:00 alex Exp $
 */

// Eugene Skepner <eu@qub.com>
// $Id: tcl_interface_gen.cc,v 1.45 1999/01/06 15:17:48 eugene Exp $

#ifndef HSystem_hh
#include "HSystem.hh"
#endif

#ifndef String_hh
#include "String.hh"
#endif

#include <fstream>
#include <stdexcept>
#include <cerrno>
#include <ctime>
#include <cstdio>

int main( int argc, char** argv )
{
	if ( argc != 3 ) {
		cerr << "Usage: " << argv[0] << " <system-name> <file-to-generate>" << endl;
		return 1;
	}

	try {
		Systems systems;
		build_systems( systems );

		Systems::iterator system = systems.find( HSystem( argv[1] ) );
		if ( system == systems.end() )
			throw runtime_error( string( "System " ) + argv[1] + " is unknown" );

		ofstream out( argv[2] );
		if ( !out )
			throw runtime_error( string( "Unable to create " ) + argv[2] + ": " + ::strerror( errno ) );

		HSystem& s = const_cast<HSystem&>( *system );
		s.gen_header( out );
		s.gen_classes_interfaces( out );
		s.gen_register_command_access( out );
		s.gen_footer( out );
	}
	catch ( exception& err ) {
		cerr << "Error: " << err.what() << endl;
		return 2;
	}

	return 0;

} /* main */

void HSystem::gen_header( ostream& out )
{
	time_t t = time( 0 );
	out << "/*" << endl;
	out << " * This file is automatically generated by tcl_interface_gen, do not edit" << endl;
	out << " */" << endl;
	out << endl;

	out << "#ifndef TclLanguageInterface_hh" << endl;
	out << "#include \"TclLanguageInterface.hh\"" << endl;
	out << "#endif" << endl;
	out << endl;

	out << "#include <tcl.h>" << endl;
	out << "#include <iostream>" << endl;
	out << "#include <typeinfo>" << endl;
	out << endl;
	out << "#define H_LIB_TCL_INTERFACE 1" << endl;
	out << endl;
	out << endl;

	for ( Includes::const_iterator ii = includes.begin(); ii != includes.end(); ++ii ) {
		out << "#include \"" << *ii << "\"" << endl;
	}
	out << endl;
	out << endl;

} /* HSystem::gen_header */

void HSystem::gen_class_header( ostream& out, const LanguageInterface::Object::ClassDescription* aClassDescription )
{
	out << "class " << replace_chars( aClassDescription->name, ':', '_' ) << "Access : ";
	if ( aClassDescription->parents[0] ) {
		for ( const LanguageInterface::Object::ClassDescription* const* pd = aClassDescription->parents; *pd; ++pd ) {
			if ( pd != aClassDescription->parents )
				out << ", ";
			out << "public " << replace_chars( (*pd)->name, ':', '_' ) << "Access";
		}
	}
	else {
		out << "virtual public TclLanguageInterface::CommandAccess";
	}
	out << "\n{\n  public:\n"
	<< "	virtual Tcl_ObjCmdProc* methodCmd() const\n		{\n			return(&" << replace_chars( aClassDescription->name, ':', '_' ) << "Access::methodCommand);\n		}\n\n"
	<< "  protected:\n";

} /* HSystem::gen_class_header */

void HSystem::gen_callback_interfaces( ostream& out, const LanguageInterface::Object::ClassDescription* aClassDescription )
{
	if ( aClassDescription->has_callbacks() ) {
		out << "	virtual void activate_callback( Tcl_Interp* interp, Tcl_Obj* aCommand, Tcl_Obj* aUserData, Tcl_Obj* aArg, LanguageInterface::SpecificCallbackData& aSpecificCallbackData )\n"
		<< "		{\n"
		<< "			/* cerr << \"" << aClassDescription->name << " activate_callback\" << endl; */\n"
		<< "			";
		for ( const LanguageInterface::Object::CallbackDescription* cd = aClassDescription->callbacks; cd->type; ++cd ) {
			out << "if ( dynamic_cast<" << cd->type << "*>( &aSpecificCallbackData ) != 0 ) {\n"
			<< "				" << cd->type << "& scd = dynamic_cast<" << cd->type << "&>( aSpecificCallbackData );\n"
			<< "				Tcl_Obj* toEval = Tcl_NewObj();\n"
			<< "				if ( Tcl_ListObjAppendElement( interp, toEval, aCommand ) != TCL_OK )\n"
			<< "					throw LanguageInterface::Error( string( \"Error creating Tcl command invokation list: \" ) + Tcl_GetStringResult( interp ) );\n";

			for ( const char* const* arg = cd->args; *arg; ++arg ) {
				out << "				if ( Tcl_ListObjAppendElement( interp, toEval, tclLanguageInterface->cxx2tcl( scd." << *arg << " ) ) != TCL_OK )\n"
				<< "					throw LanguageInterface::Error( string( \"Error creating Tcl command invokation list: \" ) + Tcl_GetStringResult( interp ) );\n";
			}
			out << "				if ( aArg && Tcl_ListObjAppendElement( interp, toEval, aArg ) != TCL_OK )\n"
			<< "					throw LanguageInterface::Error( string( \"Error creating Tcl command invokation list: \" ) + Tcl_GetStringResult( interp ) );\n";
			out << "				if ( aUserData && Tcl_ListObjAppendElement( interp, toEval, aUserData ) != TCL_OK )\n"
			<< "					throw LanguageInterface::Error( string( \"Error creating Tcl command invokation list: \" ) + Tcl_GetStringResult( interp ) );\n";
			out << "				Tcl_IncrRefCount( toEval );\n"
			<< "				if ( Tcl_EvalObj( interp, toEval ) != TCL_OK )\n"
			<< "					throw LanguageInterface::Error( string( \"Error evaluating Tcl command: \" ) + Tcl_GetStringResult( interp ) );\n"
			<< "				Tcl_DecrRefCount( toEval );\n"
			<< "				tclLanguageInterface->tcl2cxx( interp, Tcl_GetObjResult( interp ), scd.result() );\n";

			out << "			}\n"
			<< "			else ";
		}

		out << "{\n"
		<< "				throw LanguageInterface::Error( string( \"Callback invocation for \" ) + typeid( aSpecificCallbackData ).name() + \" is not supported\" );\n"
		<< "			}\n";

		out << "		}\n\n";
	}

} /* HSystem::gen_callback_interfaces */

void HSystem::gen_dynamic_methods_interfaces( ostream& out, const LanguageInterface::Object::ClassDescription* aClassDescription )
{
	out << "	static int methodCommand( ClientData obj, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[] )\n"
	<< "		{\n"
	<< "			try {\n"
	<< "				Pointer<LanguageInterface::Object> object = static_cast<LanguageInterface::Object*>( obj );\n"
	<< "				if ( !object )\n"
	<< "					return error( interp, \"Internal error in methodCommand: Invalid object \", TclLanguageInterface::to_string( objv[0] ), 0 );\n"
	<< "				const char* interpreter_name = TclLanguageInterface::to_string( objv[1] );\n"
	<< "				const bool master_interpreter = !strncmp( interpreter_name, \"master\", 6 );\n"
	<< "				/* const bool safe_interpreter = !strncmp( interpreter_name, \"safe\", 4 ); */\n"
	<< "				const char* method = TclLanguageInterface::to_string( objv[2] );\n"
	<< "				return methodCommand_object( object, method, master_interpreter, interp, objc, objv );\n"
	<< "			}\n"
	<< "			catch ( exception& err ) {\n"
	<< "				return error( interp, err.what(), 0 );\n"
	<< "			}\n"
	<< "		}\n\n";

	out << "	static int methodCommand_object( LanguageInterface::Object* object, const char* method, bool master_interpreter, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[] )\n"
	<< "		{\n"
	<< "			" << aClassDescription->cxx_name << "* object_t = dynamic_cast<" << aClassDescription->cxx_name << "*>( object );\n"
	<< "			if ( object_t == 0 )\n"
	<< "				return error( interp, \"Internal error in methodCommand: Invalid object type (" << aClassDescription->name << "expected): \", object_t->objecttypename(), 0 );\n"
	<< "			if ( !strcmp( method, \"help\" ) ) {\n"
	<< "				Tcl_SetObjResult( interp, Tcl_NewStringObj( const_cast<char*>( object_t->class_documentation().c_str() ), -1 ) );\n"
	<< "			}\n"
	<< "			else ";

	for ( const LanguageInterface::Object::MethodDescription* method = aClassDescription->methods; method->name; ++method ) {
		if ( method->dynamic() ) {
			cout << '\t' << method->name << endl;
			out << "if ( !strcmp( method, \"" << method->name << "\" ) ) {\n";
			if ( method->safeExecution() ) {
				out << "				if ( !master_interpreter ) \n"
				<< "					return error( interp, \"" << aClassDescription->name << "::" << method->name << " cannot be called in slave interpreter\", 0 );\n";
			}

			int min_args, max_args;
			method->numberOfArgs( min_args, max_args );
			out << "				if ( objc - 3 < " << min_args << " || objc - 3 > " << max_args << " ) \n"
			<< "					return error( interp, \"Wrong number of arguments: \", TclLanguageInterface::to_string( objv[0] ), \" \", method, \"\\n\", object_t->method_documentation( object_t->classDescription(), method ).c_str(), 0 );\n";
			string prefix = "				";
			for ( int argno = 0; argno < max_args; ++argno ) {
				if ( !method->getOrSet() || argno != max_args - 1 ) {
					out << prefix << LanguageInterface::method_argument_type( method->args[argno], false ) << " a" << to_string( argno );
					if ( method->args[argno] == LanguageInterface::Object::MethodDescription::vtList || method->args[argno] == LanguageInterface::Object::MethodDescription::vtListRef || method->args[argno] == LanguageInterface::Object::MethodDescription::vtListPtr )
						out << ";\n" << prefix << gen_method_argument( method->args[argno], argno + 3, argno + 1, false ) << ";\n";
					else
						out << " = " << gen_method_argument( method->args[argno], argno + 3, argno + 1, false ) << ";\n";
				}
			}

			const bool result_present = method->result != LanguageInterface::Object::MethodDescription::vtVoid;
			if ( method->getOrSet() ) {
				if ( method->setReturnsOld() ) {
					out << prefix << "const " << LanguageInterface::method_result_type( method->result, false ) << " result_object = object_t->" << method->name_to_call << "( ";
					for ( int argno = 0; argno < max_args - 1; ++argno ) {
						out << "a" << to_string( argno );
						if ( argno + 2 < max_args )
							out << ", ";
					}
					out << " );\n";

					out << prefix << "if ( objc - 3 == " << max_args << " ) { /* set requested */\n";
					out << prefix << "	" << LanguageInterface::method_argument_type( method->args[max_args - 1], false ) << " a" << to_string( max_args - 1 );
					if ( method->args[max_args - 1] == LanguageInterface::Object::MethodDescription::vtList || method->args[max_args - 1] == LanguageInterface::Object::MethodDescription::vtListRef || method->args[max_args - 1] == LanguageInterface::Object::MethodDescription::vtListPtr )
						out << ";\n" << prefix << "	" << gen_method_argument( method->args[max_args - 1], max_args + 2, max_args, true ) << ";\n";
					else
						out << " = " << gen_method_argument( method->args[max_args - 1], max_args + 2, max_args, true ) << ";\n";
					out << prefix << "	object_t->" << method->name_to_call << "( ";
					for ( int argno = 0; argno < max_args; ++argno ) {
						out << "a" << to_string( argno );
						if ( argno + 1 < max_args )
							out << ", ";
					}
					out << " );\n"
					<< prefix << "}\n";
				}
				else if ( method->setReturnsNew() ) {
					out << prefix << "if ( objc - 3 == " << max_args << " ) { /* set requested */\n";
					out << prefix << "	" << LanguageInterface::method_argument_type( method->args[max_args - 1], false ) << " a" << to_string( max_args - 1 );
					if ( method->args[max_args - 1] == LanguageInterface::Object::MethodDescription::vtList || method->args[max_args - 1] == LanguageInterface::Object::MethodDescription::vtListRef || method->args[max_args - 1] == LanguageInterface::Object::MethodDescription::vtListPtr )
						out << ";\n" << prefix << "	" << gen_method_argument( method->args[max_args - 1], max_args + 2, max_args, true ) << ";\n";
					else
						out << " = " << gen_method_argument( method->args[max_args - 1], max_args + 2, max_args, true ) << ";\n";
					out << prefix << "	object_t->" << method->name_to_call << "( ";
					for ( int argno = 0; argno < max_args; ++argno ) {
						out << "a" << to_string( argno );
						if ( argno + 1 < max_args )
							out << ", ";
					}
					out << " );\n";
					out << prefix << "}\n";
					out << prefix << "const " << LanguageInterface::method_result_type( method->result, false ) << " result_object = object_t->" << method->name_to_call << "( ";
					for ( int argno = 0; argno < max_args - 1; ++argno ) {
						out << "a" << to_string( argno );
						if ( argno + 2 < max_args )
							out << ", ";
					}
					out << " );\n";
				}
				else {
					out << prefix << "if ( objc - 3 == " << max_args << " ) { /* set requested */\n";
					out << prefix << "	" << LanguageInterface::method_argument_type( method->args[max_args - 1], false ) << " a" << to_string( max_args - 1 );
					if ( method->args[max_args - 1] == LanguageInterface::Object::MethodDescription::vtList || method->args[max_args - 1] == LanguageInterface::Object::MethodDescription::vtListRef || method->args[max_args - 1] == LanguageInterface::Object::MethodDescription::vtListPtr )
						out << ";\n" << prefix << "	" << gen_method_argument( method->args[max_args - 1], max_args + 2, max_args, true ) << ";\n";
					else
						out << " = " << gen_method_argument( method->args[max_args - 1], max_args + 2, max_args, true ) << ";\n";
					out << prefix << "	object_t->" << method->name_to_call << "( ";
					for ( int argno = 0; argno < max_args; ++argno ) {
						out << "a" << to_string( argno );
						if ( argno + 1 < max_args )
							out << ", ";
					}
					out << " );\n";
					out << prefix << "	Tcl_ResetResult( interp );\n";
					out << prefix << "}\n";
					out << prefix << "else { /* get requested */\n"
					<< prefix << "	const " << LanguageInterface::method_result_type( method->result, false ) << " result_object = object_t->" << method->name_to_call << "( ";
					for ( int argno = 0; argno < max_args - 1; ++argno ) {
						out << "a" << to_string( argno );
						if ( argno + 2 < max_args )
							out << ", ";
					}
					out << " );\n";
					out << prefix << "	Tcl_SetObjResult( interp, " << gen_result_to_obj( method->result ) << " );\n";
					out << prefix << "}\n";
				}
			}
			else {
				out << prefix;
				if ( result_present )
					out << "const " << LanguageInterface::method_result_type( method->result, false ) << " result_object = ";
				out << "object_t->" << method->name_to_call << "( ";
				for ( int argno = 0; argno < max_args; ++argno ) {
					out << "a" << to_string( argno );
					if ( argno + 1 < max_args )
						out << ", ";
				}
				out << " );\n";
			}

			if ( result_present ) {
				if ( !method->getOrSet() || method->setReturnsNew() || method->setReturnsOld() )
					out << "				Tcl_SetObjResult( interp, " << gen_result_to_obj( method->result ) << " );\n";
			}
			else {
				out << "				Tcl_ResetResult( interp );\n";
			}
			out << "			}\n"
			<< "			else ";
		}
	}

	if ( aClassDescription->parents[0] ) {
		out << "if ( ";
		for ( const LanguageInterface::Object::ClassDescription* const* pd = aClassDescription->parents; *pd; ++pd ) {
			if ( pd != aClassDescription->parents )
				out << " && ";
			out << replace_chars( (*pd)->name, ':', '_' ) << "Access::methodCommand_object( object, method, master_interpreter, interp, objc, objv ) == TCL_ERROR";
		}
		out << " ) {\n"
		<< "				 return(TCL_ERROR);\n"
		<< "			}\n";
	}
	else {
		out << "{\n"
		<< "				 /* cerr << \"methodCommand: \" << (void*)objv[0] << \" \" << TclLanguageInterface::to_string( objv[0] ) << ' ' << method << endl; */\n"
		<< "				 return error( interp, \"Unknown method \", method, \" for object of type \", object_t->objecttypename(), 0 );\n"
		<< "			}\n";
	}
	out << "			return(TCL_OK);\n"
	<< "		}\n\n";

} /* HSystem::gen_dynamic_methods_interfaces */

void HSystem::gen_static_method_interfaces( ostream& out, const LanguageInterface::Object::ClassDescription* aClassDescription )
{
	if ( aClassDescription->staticMethods() > 0 ) {
		out << "  public:\n";
		for ( const LanguageInterface::Object::MethodDescription* method = aClassDescription->methods; method->name; ++method ) {
			if ( !method->dynamic() ) {
				cout << '\t' << method->name << endl;

				out << "	static int staticMethod_" << method->name << "( ClientData /*obj*/, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[] )\n"
				<< "		{\n"
				<< "			try {\n";

				if ( method->safeExecution() ) {
					out << "				const char* interpreter_name = TclLanguageInterface::to_string( objv[1] );\n"
					<< "				const bool master_interpreter = !strncmp( interpreter_name, \"master\", 6 );\n"
					<< "				/* const bool safe_interpreter = !strncmp( interpreter_name, \"safe\", 4 ); */\n";
					out << "				if ( !master_interpreter ) \n"
					<< "					return error( interp, \"" << aClassDescription->name << "::" << method->name << " cannot be called in slave interpreter\", 0 );\n";
				}

				int min_args, max_args;
				method->numberOfArgs( min_args, max_args );
				const int max_args_in_call = method->args[max_args-1] == LanguageInterface::Object::MethodDescription::vtLanguageInterface ? max_args - 1 : max_args;
				out << "				if ( objc - 2 < " << min_args << " || objc - 2 > " << max_args_in_call << " ) \n"
				<< "					return error( interp, \"Wrong number of arguments: \", \"" << aClassDescription->name << "::" << method->name << "\", \" (was \", to_string( objc - 2 ).c_str(), \", expected " << ( min_args < max_args_in_call ? to_string( min_args ) + " to " + to_string( max_args_in_call ) : to_string( max_args_in_call ) ) << ")\", 0 );\n";
				out << "				const char* method = \"" << method->name << "\";\n";
				string prefix = "				";
				for ( int argno = 0; argno < max_args; ++argno ) {
					out << prefix << LanguageInterface::method_argument_type( method->args[argno], false ) << " a" << to_string( argno );
					if ( method->args[argno] == LanguageInterface::Object::MethodDescription::vtList || method->args[argno] == LanguageInterface::Object::MethodDescription::vtListRef || method->args[argno] == LanguageInterface::Object::MethodDescription::vtListPtr )
						out << ";\n" << prefix << gen_method_argument( method->args[argno], argno + 2, argno + 1, false ) << ";\n";
					else
						out << " = " << gen_method_argument( method->args[argno], argno + 2, argno + 1, false ) << ";\n";
				}
				out << prefix;
				if ( method->result != LanguageInterface::Object::MethodDescription::vtVoid )
					out << LanguageInterface::method_result_type( method->result, false ) << " result_object = ";
				out << aClassDescription->cxx_name << "::" << method->name_to_call << "( ";
				for ( int argno = 0; argno < max_args; ++argno ) {
					out << "a" << to_string( argno );
					if ( argno + 1 < max_args )
						out << ", ";
				}
				out << " );\n";
				if ( method->result != LanguageInterface::Object::MethodDescription::vtVoid ) {
					out << "				Tcl_SetObjResult( interp, " << gen_result_to_obj( method->result ) << " );\n";
				}
				else {
					out << "				Tcl_ResetResult( interp );\n";
				}

				out << "				return(TCL_OK);\n"
				<< "			}\n"
				<< "			catch ( exception& err ) {\n"
				<< "				return error( interp, err.what(), 0 );\n"
				<< "			}\n"
				<< "		}\n\n";

				staticMethods.push_back( make_pair( aClassDescription->name, method->name ) );
			}
		}
	}

} /* HSystem::gen_static_method_interfaces */

void HSystem::gen_classes_interfaces( ostream& out )
{
	for ( Descriptions::const_iterator di = descriptions.begin(); di != descriptions.end(); ++di ) {
		cout << (*di)->name << endl;
		gen_class_header( out, *di );
		gen_dynamic_methods_interfaces( out, *di );
		gen_static_method_interfaces( out, *di );
		gen_callback_interfaces( out, *di );
		out << "}; /* class " << replace_chars( (*di)->name, ':', '_' ) << "Access */\n";

		out << endl;
		out << endl;

		commandAccessToRegister.push_back( (*di)->name );
	}

} /* HSystem::gen_dynamic_methods_interfaces */


void HSystem::gen_register_command_access( ostream& out )
{
	out << "void register_command_access_tcl()\n"
	<< "{\n";

	for ( CommandAccessToRegister::const_iterator ci = commandAccessToRegister.begin(); ci != commandAccessToRegister.end(); ++ci ) {
		out << "	tclLanguageInterface->registerCommandAccess( typeid( " << *ci << " ).name(), new " << replace_chars( *ci, ':', '_' ) << "Access() );\n";
	}
	out << "\n";

	/* No const_iterator below due to problems with as on FreeBSD 2.2.6 */
	/* XXX can this be changed for RELENG_4 or CURRENT (5) (alex) */
	for ( StaticMethods::const_iterator smi = staticMethods.begin(); smi != staticMethods.end(); ++smi ) {
		out << "	tclLanguageInterface->create_command( \"" << smi->first << "::" << smi->second << "\", &" << smi->first << "Access::staticMethod_" << smi->second << ", 0 );\n";
		if ( smi->first == smi->second )
			out << "	tclLanguageInterface->create_command( \"" << smi->first << "\", &" << smi->first << "Access::staticMethod_" << smi->second << ", 0 );\n";
	}

	out << "}\n\n";
	out << endl;

} /* HSystem::gen_register_command_access */


void HSystem::gen_footer( ostream& out )
{
	out << "/* end */" << endl;
	out << endl;

} /* HSystem::gen_footer */


string HSystem::gen_result_to_obj( LanguageInterface::Object::MethodDescription::Value t )
{
	switch ( (LanguageInterface::Object::MethodDescription::ValueType)t ) {
	case LanguageInterface::Object::MethodDescription::vtVoid:
		return("Tcl_NewObj()");
	case LanguageInterface::Object::MethodDescription::vtChar:
	case LanguageInterface::Object::MethodDescription::vtUChar:
	case LanguageInterface::Object::MethodDescription::vtShort:
	case LanguageInterface::Object::MethodDescription::vtUShort:
	case LanguageInterface::Object::MethodDescription::vtInt:
	case LanguageInterface::Object::MethodDescription::vtUInt:
	case LanguageInterface::Object::MethodDescription::vtLong:
	case LanguageInterface::Object::MethodDescription::vtULong:
	case LanguageInterface::Object::MethodDescription::vtLongLong:
	case LanguageInterface::Object::MethodDescription::vtULongLong:
	case LanguageInterface::Object::MethodDescription::vtBool:
	case LanguageInterface::Object::MethodDescription::vtFloat:
	case LanguageInterface::Object::MethodDescription::vtDouble:
	case LanguageInterface::Object::MethodDescription::vtLongDouble:
	case LanguageInterface::Object::MethodDescription::vtString:
	case LanguageInterface::Object::MethodDescription::vtObject:
		return("tclLanguageInterface->cxx2tcl( result_object )");
	case LanguageInterface::Object::MethodDescription::vtList:
	case LanguageInterface::Object::MethodDescription::vtListRef:
	case LanguageInterface::Object::MethodDescription::vtListPtr:
		return("tclLanguageInterface->list2tcl( interp, result_object )");
	case LanguageInterface::Object::MethodDescription::vtLanguageInterface:
		throw LanguageInterface::Error( string( "vtLanguageInterface cannot be used as method return type for method " ) + (const char*)t );
	case LanguageInterface::Object::MethodDescription::vtCallbackData:
		return("tclLanguageInterface->callbackData2tcl( interp, result_object )");
	}

	return("unknown");

} /* HSystem::gen_result_to_obj */


string HSystem::gen_method_argument( LanguageInterface::Object::MethodDescription::Value t, int argno, int method_argno, bool last_arg_in_set_method )
{
	string result;
	if ( t.optional() && t != LanguageInterface::Object::MethodDescription::vtList && t != LanguageInterface::Object::MethodDescription::vtListRef && t != LanguageInterface::Object::MethodDescription::vtListPtr && !last_arg_in_set_method ) {
		result += string( "( objc - 1 < " ) + to_string( argno ) + " ) ? " + t.defaultValue() + " : ";
	}

	switch ( (LanguageInterface::Object::MethodDescription::ValueType)t ) {
	case LanguageInterface::Object::MethodDescription::vtVoid:
		return("void");
	case LanguageInterface::Object::MethodDescription::vtChar:
	case LanguageInterface::Object::MethodDescription::vtUChar:
		result += string( "*Tcl_GetStringFromObj( objv[" ) + to_string( argno ) + "], 0 )";
		break;
	case LanguageInterface::Object::MethodDescription::vtShort:
	case LanguageInterface::Object::MethodDescription::vtUShort:
	case LanguageInterface::Object::MethodDescription::vtInt:
	case LanguageInterface::Object::MethodDescription::vtUInt:
		result += string( "TclLanguageInterface::tcl_obj2int( interp, objv[" ) + to_string( argno ) + "], method, " + to_string( method_argno ) + " )";
		break;
	case LanguageInterface::Object::MethodDescription::vtLong:
	case LanguageInterface::Object::MethodDescription::vtULong:
	case LanguageInterface::Object::MethodDescription::vtLongLong:
	case LanguageInterface::Object::MethodDescription::vtULongLong:
		result += string( "TclLanguageInterface::tcl_obj2long( interp, objv[" ) + to_string( argno ) + "], method, " + to_string( method_argno ) + " )";
		break;
	case LanguageInterface::Object::MethodDescription::vtBool:
		result += string( "TclLanguageInterface::tcl_obj2bool( interp, objv[" ) + to_string( argno ) + "], method, " + to_string( method_argno ) + " )";
		break;
	case LanguageInterface::Object::MethodDescription::vtFloat:
	case LanguageInterface::Object::MethodDescription::vtDouble:
	case LanguageInterface::Object::MethodDescription::vtLongDouble:
		result += string( "TclLanguageInterface::tcl_obj2double( interp, objv[" ) + to_string( argno ) + "], method, " + to_string( method_argno ) + " )";
		break;
	case LanguageInterface::Object::MethodDescription::vtString:
		result += string( "TclLanguageInterface::to_string( objv[" ) + to_string( argno ) + "] )";
		break;
	case LanguageInterface::Object::MethodDescription::vtObject:
		result += string( "tclLanguageInterface->tcl2object( objv[" ) + to_string( argno ) + "], (" + (const char*)t + "*)0, method, " + to_string( method_argno ) + " )";
		break;
	case LanguageInterface::Object::MethodDescription::vtList:
	case LanguageInterface::Object::MethodDescription::vtListRef:
	case LanguageInterface::Object::MethodDescription::vtListPtr:
		if ( t.optional() && !last_arg_in_set_method )
			result += string( "( objc - 1 < " ) + to_string( argno ) + " ) ? ( a" + to_string( method_argno - 1 ) + " = " + t.defaultValue() + " ) : ";
		result += string( "tclLanguageInterface->tcl2list( a" ) + to_string( method_argno - 1 ) + ", interp, objv[" + to_string( argno ) + "], method, " + to_string( method_argno ) + " )";
		break;
	case LanguageInterface::Object::MethodDescription::vtLanguageInterface:
		return("tclLanguageInterface");
	case LanguageInterface::Object::MethodDescription::vtCallbackData:
		result += string( "tclLanguageInterface->tcl2callbackData( interp, objv[" ) + to_string( argno ) + "] )";
		break;
	default:
		return("unknown");
	}

	return(result);

} /* HSystem::gen_method_argument */
